/* $Id: dfs.c,v 1.1 1998/04/23 18:43:20 ashaikh Exp $ */

/*
*
*   Routines that implement a depth-first search on a arbitrary graph
*   given in the adjacency matrix format.  These routines
*   are primarily for determining whether the graph is connected.  Note we only
*   care about the connected component containing the source so we only search as
*   far as possible from the source.
* 
*/

#include "dfs.h"
#include "rsdefs.h"
#include "rstrace.h"
#include <stdlib.h>

enum color {white, grey, black};

static int		*d;	    /* discovery time in depth-first search */
static int		*f;	    /* finish time in depth-first search */
static int		*parents;	    /* predecessor of each node in depth-first tree */
static int		*cc;	    /* label of the connected component to which node belongs */
static int		ccount;	    /* label for connected components */
static enum color	*colors;    /* color field to aid in depth-first search */
static int		timestamp;

static int dfs (link_t ***, int);
static void dfsVisit (int, link_t ***, int );
static void dfsDestroy (void);
static int dfsTranspose (link_t ***, int *, int);
static int getNextFNode (int *, int);

/*******************************************************************************/
static int
dfs (link_t ***netmatrix, int numnodes)
{
  int	i;

  d = (int *) malloc (numnodes*sizeof(int));
  f = (int *) malloc (numnodes*sizeof(int));
  parents = (int *) malloc (numnodes*sizeof(int));
  cc = (int *) malloc (numnodes*sizeof(int));
  colors = (enum color *) malloc (numnodes*sizeof(enum color));
  if (d == (int *) 0 ||
      f == (int *) 0 ||
      parents == (int *) 0 ||
      cc == (int *) 0 ||
      colors == (enum color *) 0)
    return (-1);
      
  for (i = 0; i < numnodes; i++) {
    colors[i] = white;
    parents[i] = -1;
    cc[i] = -1;
  }
  timestamp = 0;
  ccount = 0;
  for (i = 0; i < numnodes; i++) {
    if (colors[i] == white) {
      dfsVisit(i, netmatrix, numnodes);   /* start a depth-first search from the source */
      ccount++;
    }
  }

  return (0);

}
/*******************************************************************************/
static void
dfsVisit (int node, link_t ***netmatrix, int numnodes)
{
  int	j;
  int	neighbor;

  colors[node] = grey;
  cc[node] = ccount;
  d[node] = timestamp++;

  for (j = 0; j < numnodes; j++)
    if (netmatrix[node][j] != (link_t *) 0) {	/* check for adjacency */
      neighbor = j;
      if (colors[neighbor] == white) {
	parents[neighbor] = node;
	dfsVisit(neighbor, netmatrix, numnodes);
      }
    }
  colors[node] = black;
  f[node] = timestamp++;

}
/********************************************************************************/
int
isConnected (link_t ***netmatrix, int numnodes, int src)
/*  Uses a depth-first search algorithm to label the connected component
 *  that is reachable from the source and then check that all nodes are
 *  reachable from the source.
 */
{
  int	i;
  int	sourcelabel;
  int	retval;

  if (dfs (netmatrix, numnodes)) {
    RS_PRINT_ERROR ("isConnected:  dfs() failed\n");
    return (0);
  }
  sourcelabel = cc[src];
  retval = 1;
  for (i = 0; i < numnodes; i++)
    if (cc[i] != sourcelabel) {
      retval = 0;
      continue;
    }
  dfsDestroy();
  return (retval);
}
/********************************************************************************/
int
isStronglyConnected (link_t ***graph, int numnodes)
/* Use dfs twice to compute the strongly connected components of the graph.
 * This follows the procedure laid out in Cormen, Leiserson, and Rivest,
 * "Introduction to Algorithms"
 */
{
  link_t    ***transpgraph;
  int	    *origf;
  int	    i, j;
  int	    label, retval;

  /* run dfs on original graph to compute each node's finish time */
  if (dfs (graph, numnodes)) {
    RS_PRINT_ERROR ("isStronglyConnected:  dfs() failed\n");
    return (0);
  }

  /* save the finish time matrix, f[] */
  origf = (int *) malloc (numnodes*sizeof(int));
  for (i = 0; i < numnodes; i++)
    origf[i] = f[i];

  dfsDestroy();	/* free memory from first computation */
  
  /* compute the transposed graph */
  transpgraph = (link_t ***) malloc (sizeof(link_t *)*numnodes);
  if (transpgraph == (link_t ***) 0) {
    RS_PRINT_ERROR ("isStronglyConnected: Couldn't allocate transposed graph\n");
    return (0);
  }
  for (i = 0; i < numnodes; i++) {
    transpgraph[i] = (link_t **) malloc (sizeof(link_t *)*numnodes);
    if (transpgraph[i] == (link_t **) 0) {
      RS_PRINT_ERROR ("isStronglyConnected: Couldn't allocate transposed graph entry\n");      
      return (-1);
    }
    for (j = 0; j < numnodes; j++) {
      transpgraph[i][j] = graph[j][i];
    }
  }

  /* run dfs on the transposed graph */
  if (dfsTranspose (transpgraph, origf, numnodes)) {
    RS_PRINT_ERROR ("isStronglyConnected:  dfsTranspose() failed\n");
    return (0);
  }
  for (i = 0; i < numnodes; i++) {
    free (transpgraph[i]);
  }
  free (transpgraph);
  free (origf);

  /* check if all nodes have the same label */
  label = cc[0];
  retval = 1;
  for (i = 0; i < numnodes; i++) {
    if (cc[i] != label) {
      retval = 0;
      continue;
    }
  }
  
  dfsDestroy();

  return (retval);

}
/********************************************************************************/
static int
dfsTranspose (link_t ***graph, int *origf, int numnodes)
/* Like dfs() but runs on the transposed graph and considers nodes in order of
 * decreasing finish time as computed for the original graph.
 */
{
  int	i;
  int	next;

  d = (int *) malloc (numnodes*sizeof(int));
  f = (int *) malloc (numnodes*sizeof(int));
  parents = (int *) malloc (numnodes*sizeof(int));
  cc = (int *) malloc (numnodes*sizeof(int));
  colors = (enum color *) malloc (numnodes*sizeof(enum color));
  if (d == (int *) 0 ||
      f == (int *) 0 ||
      parents == (int *) 0 ||
      cc == (int *) 0 ||
      colors == (enum color *) 0)
    return (-1);
      
  for (i = 0; i < numnodes; i++) {
    colors[i] = white;
    parents[i] = -1;
    cc[i] = -1;
  }
  ccount = 0;
  timestamp = 0;

  next = getNextFNode (origf, numnodes);
  while (next >= 0) { /* getNextFNode return -1 when it can't find an unvisited node */
    cc[next] = ccount;
    dfsVisit(next, graph, numnodes);   /* start a depth-first search from the source */
    ccount++;
    next = getNextFNode (origf, numnodes);
  }

  return (0);

}
/********************************************************************************/
static int
getNextFNode (int *farray, int numnodes)
/* Find the next unvisited node with highest f-value -- brute force
 */
{
  int	i;
  int	maxf, maxfindex;

  maxf = 0;
  maxfindex = -1;
  for (i = 0; i < numnodes; i++) {
    if (( farray[i] > maxf) && (colors[i] == white)) {
      maxf = farray[i];
      maxfindex = i;
    }
  }
  return (maxfindex);
}
/********************************************************************************/
static void
dfsDestroy (void)
{
  free (d);
  free (f);
  free (parents);
  free (cc);
  free (colors);
}
