/* $Id: rsroute.c,v 2.6 1999/06/10 17:32:54 ashaikh Exp $ */

static const char rcsid[] = "$Id: rsroute.c,v 2.6 1999/06/10 17:32:54 ashaikh Exp $";

#include "rsroute.h"
#include "rsstat.h"
#include "spt.h"
#include "bfalg.h"
#include "hsfalg.h"
#include "rsdefs.h"
#include "rstrace.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>


routepolicy_t routepolicy = (routepolicy_t) -1;	/* on-demand or precomputed routes */
staleness_t accuracy = (staleness_t) -1;	/* link-state information to use */

/* SPT policy options - for SINGLE or dynamic routing in DUAL */
prunepolicy_t prunep = (prunepolicy_t) -1;	/* policy used for pruning links */
distfunction_t distfunc = (distfunction_t) -1;	/* shortest path calc parameter */
alternate_t altpolicy = (alternate_t) -1;	/* how alternate routing should work */
multipolicy_t multipolicy = (multipolicy_t) -1;	/* how to use multiple routes in extraction */
uniqpolicy_t uniqpolicy = (uniqpolicy_t) -1;	/* how to pick uniquely among multiple routes */

boolean feascheck = (boolean) FALSE;		/* whether to check feasibility on extraction */
boolean trigrecompsig = (boolean) FALSE;	/* trigger route recomputations on sig failures */
boolean trigrecomproute = (boolean) FALSE;	/* trigger route recomputations on route comp fail */
boolean prunerecomp = (boolean) FALSE;		/* removed failed links from top db on recomp */
boolean recompafterfail = (boolean) FALSE;	/* recompute even when crankbacks are exhausted */
int	hopthresh = -1;				/* threshold to control non-minimal routes */
int	maxhopcount = 0;			/* max width of bellman-ford table */

double	rscostalpha = 0.0;
double	rscostmin = 0.0;
double	bwclasssize = 0.0;
int	rscostlevels = MAX_COST_LEVELS;
double	rscostconst = 0.0;

/* routing stats */
extern simpleratio_t rcompinterval;
extern int extractfailures;

static int routeCallOnDemand (call_t *, double, int *);
static int routeCallPrecomp (call_t *, double, int *);
static int computeSPTFallback (call_t *, double, int *);
static int isNodeConnected (int, int, topology_t *);
static void resetCallRoute (call_t *);
static int createInitRoutes (topology_t *);
static int extractSPTRoute (call_t *, int *);
static int extractBFRoute (call_t *, int *);
static int extractHSFRoute (call_t *, int *);

static double calcExpLinkCost (double);
static int validRouteExists (int, topology_t *);
/********************************************************************************/
int
initRouting (routealg_t routealg, topology_t *network)
{
  int	    i, numnodes;
  spt_t	    **sstrees;	/* source-specific shortest-path trees */
  bftab_t   **bftables;	/* source-specific bellman-ford tables */

  
  if (rscostlevels <= 0)
    rscostlevels = MAX_COST_LEVELS;

  numnodes = getNumNodes(network);
  network->routingalg = routealg;

  /* init source-specific data structures */
  switch (routealg) {
  case SPT:
    sstrees = (spt_t **) malloc (sizeof(spt_t *) * numnodes);
    for (i = 0; i < numnodes; i++) {
      if (sptInit (&(sstrees[i]), network)) {
	RS_PRINT_ERROR ("initRouting: SPT allocation failed\n");
	return (-1);
      }
    }
    network->routetable = (void **) sstrees;
    break;
  case WIDE_SHORT:
    bftables = (bftab_t **) malloc (sizeof(bftab_t *) * numnodes);
    for (i = 0; i < numnodes; i++) {
      if (bfInit (&(bftables[i]), i, maxhopcount, hopthresh, network)) {
	RS_PRINT_ERROR ("initRouting: BF Table allocation failed\n");
	return (-1);
      }
    }
    network->routetable = (void **) bftables;
    break;
    
    /* Dont need any initializing at this point.
     * Ganga
     */    

  case HSF:
    break;

  default:
    RS_PRINT_ALWAYS ("initRouting:  Unknown routing algorithm!\n");
    return (-1);
  }    

  if (routepolicy == PRECOMP) {	/* set up initial routes trees for precomputed */
    if (createInitRoutes(network)) {
      RS_PRINT_ERROR ("createInitTrees returned error\n");
      return (-1);
    }
  }

  /* if hopthresh has not been specified, we set it to be very large */
  if (hopthresh < 0) {
    hopthresh = numnodes - 1;
  }

  /* this should get filled in later if necessary */
  network->routepolicies = (void *) 0;

  return (0);
}
/********************************************************************************/
int
setRoutePolicy (topology_t *network, void *policy)
/* Sets the routing policies for the network.  If the policy pointer is NULL, policies
 * are filled in from global vars.  Otherwise, policies are assigned to those pointed to
 * by the policy pointer.  Currently, policies are only defined for the SPT routing alg.
 */
{
  sptpolicy_t	*sptpolicy;
  
  if (policy == (void *) 0) {
    switch (network->routingalg) {
    case SPT:
      sptpolicy = (sptpolicy_t *) malloc (sizeof(sptpolicy_t));
      if (sptpolicy == (sptpolicy_t *) 0) {
	RS_PRINT_ERROR ("setRoutePolicy:  unable to allocate sptpolicy\n");
	return (-1);
      }
      sptpolicy->prunep =  prunep;
      sptpolicy->distfunc = distfunc;
      sptpolicy->altpolicy = altpolicy;
      sptpolicy->multipolicy = multipolicy;
      sptpolicy->uniqpolicy = uniqpolicy;
      network->routepolicies = (void *) sptpolicy;
      break;
    case WIDE_SHORT:
      RS_PRINT_DETAIL ("setRoutePolicy:  No routing policies defined for WIDE_SHORT algorithm\n");
      network->routepolicies = (void *) 0;
      break;

      /* No routing policies defined for HSF
       * Ganga
       */
    case HSF:
      RS_PRINT_DETAIL ("setRoutePolicy: No routing policies defined for HSF algorithm\n");
      network->routepolicies = (void *) 0;
      break;


    default:
      RS_PRINT_ALWAYS ("setRoutePolicy:  Unknown routing algorithm!\n");
      network->routepolicies = (void *) 0;
      return (-1);
    }
  }
  else {
    network->routepolicies = policy;
  }

  return (0);
}
/********************************************************************************/
  
static int
createInitRoutes (topology_t *network)
{
  int	    i, numnodes;
  int	    calcresult;
  spt_t	    **sstrees;
  sptpolicy_t	*sptpolicy;
  bftab_t   **bftables;

  numnodes = getNumNodes(network);
  switch (network->routingalg) {
  case SPT:
    sstrees = (spt_t **) network->routetable;
    sptpolicy = (sptpolicy_t *) network->routepolicies;
    for (i = 0; i < numnodes; i++) {
      sptReset (sstrees[i]);
      /* compute trees with 0.0 bandwidth requirement and at 0.0 time */
      if (sptExec (sstrees[i], i, 0.0, 0.0,
		   sptpolicy->prunep, sptpolicy->distfunc, accuracy, &calcresult)) {
	RS_PRINT_ERROR ("createInitRoutes: sptExec failed\n");
	return (-1);
      }
      if (calcresult != ROUTE_CALC_FAIL) {
	sptMultiPathExtractInit (sstrees[i]);    /* reset head pointers */
      }
      else {  /* route calculation failed for some reason -- shouldn't happen */
	RS_PRINT_ERROR ("createInitRoutes: SPT route calc failed\n");
	return (-1);
      }
    }
    break;
  case WIDE_SHORT:
    bftables = (bftab_t **) network->routetable;
    for (i = 0; i < numnodes; i++) {
      bfReset (bftables[i]);
      /* compute BF tables at 0 time */
      if (bfExec (bftables[i], 0.0)) {
	RS_PRINT_ERROR ("createInitRoutes:  BF route calc returned error\n");
	return (-1);
      }
    }
    break;

    /* Nothing to initialize at this point
     * Ganga
     */
    
  case HSF:
    break;


  default:
    RS_PRINT_ALWAYS ("createInitRoutes:  Unknown routing algorithm!\n");
    return (-1);
  }
      
  return (0);
}

/********************************************************************************/
int
routeCall (call_t *call, double time, int *routecode)
{
  int	code;
  
  switch (routepolicy) {
  case ON_DEMAND:
    if (routeCallOnDemand (call, time, &code)) {
      RS_PRINT_ERROR ("newCallRouteHandler: routeCallOnDemand failed\n");
      return (-1);
    }
    break;
  case PRECOMP:
    if (routeCallPrecomp (call, time, &code)) {
      RS_PRINT_ERROR ("newCallRouteHandler: routeCallPrecomp failed\n");
      return (-1);
    }
    break;
  default:
    RS_PRINT_ERROR ("Unknown routing policy\n");
    return (-1);
  }
  *routecode = code;

  return (0);
}
/********************************************************************************/
static int
routeCallOnDemand (call_t *newcall, double time, int *result)
/* Computes a new route tree and tries only to extract from the a route.
 */
{
  int		calcresult, extractresult;
  int		source, dest;
  rsstat_t	*stats;

  source = newcall->source;
  dest = newcall->dest;
  stats = newcall->network->stats;
  

  /* calculate path */

  /* Added "newcall" as an argument
   * Ganga - 03/12/2000
   */
  if (reComputeRoute (newcall, source, time, newcall->bandw, newcall->network, &calcresult)) {
    RS_PRINT_ERROR ("newCallEvHandler:  reComputeRoute returned failure\n");
    return (-1);
  }  
  newcall->routecomps++;
  if (updateSimpleRatioStat ((time - getLastRouteCompTime(source, newcall->network)), &stats->rcompinterval)) {
    RS_PRINT_ERROR ("signalCall: Error updating rcompinterval\n");
  }    

  if ( (calcresult == ROUTE_CALC_SUCCESS) ||
       ( (calcresult == ROUTE_CALC_DISCONN) &&
	 (isNodeConnected (source, dest, newcall->network)) ) ) {
    /* if we get ROUTE_CALC_DISCONN,  we still may be okay if the
     * dest of interest is connected.
     */
    /* now try to extract */
    if (extractRoute (newcall, time, &extractresult)) {
      RS_PRINT_ERROR ("routeCallOnDemand:  extractRoute returned failure\n");
      return (-1);
    }
    if (extractresult == ROUTE_EXT_SUCCESS) {
      *result = ROUTING_SUCCESS_R;
    }
    else {  /* could not extract route --> routing failure */
      RS_PRINT_EVENT ("routeCallOnDemand: Route extraction failed\n");
      *result = ROUTING_FAILURE;
    }
  } /* if calcresult == ROUTE_CALC_SUCCESS */
  else if ( (newcall->network->blkpolicy == NON_BLOCKING) ) {
    /* fall back to minimal path in this case. not clear what to do
     * with precomputed routes -- recomputing to get minimal may not
     * be what we want.  just do this for on-demand for now...
     */
    if (computeSPTFallback (newcall, time, &calcresult)) {
      RS_PRINT_ERROR ("reComputeRoute:  computeSPTFallback returned error\n");
      return (-1);
    }
    newcall->routecomps++;
    if (calcresult != ROUTE_CALC_SUCCESS) {
      *result = ROUTING_FAILURE;
    }
    else {
      /* extract route after fallback */
      if (extractRoute (newcall, time, &extractresult)) {
	RS_PRINT_ERROR ("routeCallOnDemand:  extractRoute returned failure\n");
	return (-1);
      }
      if (extractresult == ROUTE_EXT_SUCCESS) {
	*result = ROUTING_SUCCESS_R;
      }
      else {  /* could not extract route --> routing failure */
	RS_PRINT_EVENT ("routeCallOnDemand: Fallback route extraction failed\n");
	*result = ROUTING_FAILURE;
      }
    }
  } /* if blkpolicy == NON_BLOCKING */
  else {
    /* route computation failed -- and policy is BLOCKING */
    RS_PRINT_EVENT1 ("routeCallOnDemand: Route computation for call %d failed (on-demand)\n",
		     newcall->callid);
    *result = ROUTING_FAILURE;
  }

  return (0);
  
}
/********************************************************************************/
int
routeCallPrecomp (call_t *newcall, double time, int *result)
{
  int		source;
  int		extractresult, calcresult;

  source = newcall->source;

  /* first make sure that we have some valid routes lying around */
  if (!validRouteExists(source, newcall->network)) {
    /* this would be unexpected */
    RS_PRINT_SOFT_ERROR ("routeCallPrecomp:  No valid routes available... recomputing\n");
    /* Added "newcall" as an argument.
     * Ganga - 03/12/2000
     */

    if (reComputeRoute (newcall, source, time, newcall->bandw, newcall->network, &calcresult)) {
      RS_PRINT_ERROR ("routeCallPrecomp:  reComputeRoute returned failure\n");
      return (-1);
    }
    if (calcresult == ROUTE_CALC_FAIL) {    /* computation failed */
      *result = ROUTING_FAILURE;
      return (0);
    }
  }
  
  if (extractRoute(newcall, time, &extractresult)) {
    RS_PRINT_ERROR ("routeCallPrecomp:  extractRoute returned failure\n");
    return (-1);
  }
  if (extractresult == ROUTE_EXT_SUCCESS) {
    *result = ROUTING_SUCCESS_NR;
  }
  else if (trigrecomproute == TRUE) {  /* extraction failed -- all routes infeasible or dest is disconn. */
    /* recompute and try extraction again */
    /* Added newcall as an argument
     * Ganga - 03/12/2000
     */

    if (reComputeRoute (newcall, source, time, newcall->bandw, newcall->network, &calcresult)) {
      RS_PRINT_ERROR ("routeCallPrecomp:  reComputeRoute returned failure\n");
      return (-1);
    }
    if (calcresult == ROUTE_CALC_FAIL) {    /* recomputation failed */
      *result = ROUTING_FAILURE;
    }
    else {  /* try to extract from new tree */
      if (extractRoute (newcall, time, &extractresult)) {
	RS_PRINT_ERROR ("routeCallPrecomp:  extractRoute returned failure\n");
	return (-1);
      }
      if (extractresult == ROUTE_EXT_SUCCESS) {
	*result = ROUTING_SUCCESS_R;      
      }
      else {  /* still could not extract route -- declare failure */
	RS_PRINT_EVENT ("routeCallPrecomp: Can't find a route\n");
	*result = ROUTING_FAILURE;	
      }
    }
  } /* else had to recompute tree */
  else {    /* recomps are disabled */
    *result = ROUTING_FAILURE;
    RS_PRINT_EVENT ("routeCallPrecomp: Can't find a route (no recomp)\n");
  }

  return (0);
}

/********************************************************************************/
/* Added call_t* as an argument to the function
 * Ganga - 03/12/2000
 */

int
reComputeRoute (call_t *newcall, int source, double time, double bw, topology_t *network, int *result)
{
  int		calcresult;
  spt_t		**sstrees;
  sptpolicy_t	*sptpolicy;
  bftab_t	**bftables;
  routealg_t	routealg;

  routealg = network->routingalg;
  /* recompute routes for this source */
  switch (routealg) {    
    /* Adding code for HSF route computation.
     * Ganga 03/11/2000
     */
  case HSF:
    /* Here call hsfExec()
     * Return success or failure
     * Ganga 03/12/2000
     */
    
    if(hsfExec(newcall, time, bw, network, &calcresult)) {
      RS_PRINT_ERROR("reComputeRoute: hsfExec failed\n");
      return(-1);
    }
    *result = calcresult;
    break;

  case SPT:
    sstrees = (spt_t **) network->routetable;
    sptpolicy = (sptpolicy_t *) network->routepolicies;
    sptReset (sstrees[source]);
    if (sptExec (sstrees[source], source, bw, time, sptpolicy->prunep,
		 sptpolicy->distfunc, accuracy, &calcresult)) {
      RS_PRINT_ERROR ("reComputeRoute: sptExec failed\n");
      return (-1);
    }
    if (calcresult != ROUTE_CALC_FAIL) {
      sptMultiPathExtractInit (sstrees[source]);	/* reset head pointers */
      *result = calcresult; /* may want to know ROUTE_CALC_DISCONN */
    }
    else {
      *result = ROUTE_CALC_FAIL;
    }
    break;
  case WIDE_SHORT:
    bftables = (bftab_t **) network->routetable;
    bfReset (bftables[source]);
    if (bfExec (bftables[source], time)) {
      RS_PRINT_ERROR ("reComputeRoute:  BF route calc returned error\n");
      return (-1);
    }
    /* bfExec never returns a route calc error -- infeasible path will be */
    /* discovered upon extraction */
    *result = ROUTE_CALC_SUCCESS;
    break;
  default:
    RS_PRINT_ALWAYS ("reComputeRoute:  Unknown routing algorithm!\n");
    return (-1);
  }
    
  return (0);
}

/********************************************************************************/
int
extractRoute (call_t *newcall, double time, int *result)
/* Extract route from existing tree
 */
{
  int		extresult, calcresult;
  routealg_t	routealg;

  resetCallRoute (newcall);
  routealg = newcall->network->routingalg;
  
  switch (routealg) {
  case SPT:
    if (extractSPTRoute (newcall, &extresult)) {
      RS_PRINT_ERROR ("extractRoute:  extractSPTRoute returned error\n");
      return (-1);
    }
    if ( (extresult == ROUTE_EXT_FAIL_LEN) &&
	 (newcall->network->blkpolicy == NON_BLOCKING) &&
	 (routepolicy == ON_DEMAND) ) {
      /* fall back to minimal path in this case. not clear what to do
       * with precomputed routes -- recomputing to get minimal may not
       * be what we want.  just do this for on-demand for now...
       */
      if (computeSPTFallback (newcall, time, &calcresult)) {
	RS_PRINT_ERROR ("extractRoute: computeSPTFallback returned error\n");
	return (-1);
      }
      newcall->routecomps++;      
      if (calcresult != ROUTE_CALC_FAIL) {
	/* reextract a minimal route from the SPT */
	if (extractSPTRoute (newcall, &extresult)) {
	  RS_PRINT_ERROR ("extractRoute:  extractSPTRoute returned error\n");
	  return (-1);
	}
      }
    }
    *result = extresult;
    break;
  case WIDE_SHORT:
    if (extractBFRoute (newcall, &extresult)) {
      RS_PRINT_ERROR ("extractRoute:  extractBFRoute returned error\n");
      return (-1);
    }
    *result = extresult;
    break;

    /* Code to extract an HSF path 
     * Ganga
     */
  case HSF:
    if (extractHSFRoute (newcall, &extresult)) {
      RS_PRINT_ERROR ("extractRoute: extractHSFRoute returned error\n");
      return (-1);
    }
    *result = extresult;
    break;
  default:
    RS_PRINT_ALWAYS ("extractRoute: Uknown routing algorithm\n");
    return (-1);
  }

  return (0);
    
}

/********************************************************************************/
static int
isNodeConnected (int src, int dest, topology_t *network)
{

  routealg_t	routealg;
  int		retval = 0;
  spt_t		**sstrees;
  
  routealg = network->routingalg;
  switch (routealg) {
  case SPT:
    sstrees = (spt_t **) network->routetable;
    retval = sptIsNodeConnected (sstrees[src], dest);
    break;
  case WIDE_SHORT:
    /* in the BF table, nodes are always connected -- but not necessarily
     * via a feasible path */
    retval = 1;
    break;

    /* same as for WIDE_SHORT for HSF too
     * Ganga
     */
  case HSF:
    retval = 1;
    break;

  default:
    RS_PRINT_ALWAYS ("extractRoute: Uknown routing algorithm\n");
    return (-1);
  }

  return (retval);
}
/********************************************************************************/
static int
extractSPTRoute (call_t *newcall, int *result)
{
  int		source;
  int		extractresult;
  boolean	localfeas;
  spt_t		**sstrees;
  sptpolicy_t	*sptpolicy;
  rsstat_t	*stats;

  source = newcall->source;
  sstrees = (spt_t **) newcall->network->routetable;
  sptpolicy = (sptpolicy_t *) newcall->network->routepolicies;
  stats =  newcall->network->stats;

  /* can't have feasibility checking with a non-blocking network */
  (sstrees[source]->network->blkpolicy == NON_BLOCKING) ? (localfeas = FALSE) : (localfeas = feascheck);  
  switch (sptpolicy->multipolicy) {
  case MULTI:
    if (sptExtractMultiPath(sstrees[source], newcall, sptpolicy->altpolicy, localfeas,
			    &newcall->route, &extractresult)) {
      RS_PRINT_ERROR1 ("extractSPTRoute: sptExtractMultiPath returned failure for call %d\n", newcall->callid);
      return (-1);
    }
    break;
  case FIRST: /* just extract the first path */
    if (sptExtractSinglePath(sstrees[source], newcall, localfeas,
			     &newcall->route, &extractresult)) {
      RS_PRINT_ERROR1 ("extractSPTRoute: sptExtractSinglePath failed for call %d\n", newcall->callid);
      return (-1);
    }
    break;
  case UNIQ:
    if (sptExtractUniqPath (sstrees[source], newcall, -1, -1, sptpolicy->uniqpolicy, localfeas,
			    &newcall->route, &extractresult)) {
      RS_PRINT_ERROR1 ("extractSPTRoute: sptExtractUniqPath failed for call %d\n", newcall->callid);
      return (-1);
    }
    break;
  default:
    RS_PRINT_ALWAYS ("extractSPTRoute:  Unknown multiple route policy\n");
    return (-1);
  }

  newcall->extractions++;
  
  if (extractresult == ROUTE_EXT_SUCCESS) {
    if ((newcall->routelen - 1) > newcall->minhops + hopthresh) {  /* route is too long */
#if RS_TRACE_LEVEL >= TRACE_DETAILS
      printf ("extractSPTRoute: Route too long for call %d: routelen %d (min: %d)\n",
	      newcall->callid, newcall->routelen, newcall->minhops);
#endif
      *result = ROUTE_EXT_FAIL_LEN;
      stats->extractfailures++;
    }
    else {  /* route is within hopcount threshold */
      if (getLinks(newcall->route, newcall->routelen, &newcall->links,
		   sstrees[source]->network)) {
	RS_PRINT_ERROR ("extractSPTRoute: getLinks returned failure\n");
	return (-1);
      }
      RS_PRINT_EVENT1 ("extractSPTRoute: Extracted route for call %d\n",newcall->callid);
      *result = ROUTE_EXT_SUCCESS;
    }
  }
  else {  /* could not extract route -- declare failure */
    RS_PRINT_EVENT1 ("extractSPTRoute: Can't find a route for call %d\n", newcall->callid);
    *result = ROUTE_EXT_FAIL;
    stats->extractfailures++;
  }

  return (0);
}
/********************************************************************************/
static int
extractBFRoute (call_t *newcall, int *result)
{
  int	    source;
  int	    extractresult;
  bftab_t   **bftables;
  rsstat_t  *stats;

  source = newcall->source;
  bftables = (bftab_t **) newcall->network->routetable;
  stats = newcall->network->stats;

  if (bfExtractPath(bftables[source], newcall, &newcall->route, &extractresult)) {
    RS_PRINT_ERROR1 ("extractBFRoute: bfExtractPath failed for call %d\n", newcall->callid);
    return (-1);
  }
  newcall->extractions++;
  
  if (extractresult == ROUTE_EXT_SUCCESS) {
    if ((newcall->routelen - 1) > newcall->minhops + hopthresh) {
#if RS_TRACE_LEVEL >= TRACE_DETAILS
      printf ("extractBFRoute: Route too long for call %d: routelen %d (min: %d)\n",
	      newcall->callid, newcall->routelen, newcall->minhops);
#endif
      *result = ROUTE_EXT_FAIL;
      stats->extractfailures++;
    }
    else {  /* route is within hopcount threshold */
      if (getLinks(newcall->route, newcall->routelen, &newcall->links,
		   bftables[source]->network)) {
	RS_PRINT_ERROR ("extractBFRoute: getLinks returned failure\n");
	return (-1);
      }
      RS_PRINT_EVENT1 ("extractBFRoute: Extracted route for call %d\n",newcall->callid);
      *result = ROUTE_EXT_SUCCESS;
    }
  }
  else {  /* could not extract route -- declare failure */
    RS_PRINT_EVENT1 ("extractBFRoute: Can't find a route for call %d\n", newcall->callid);
    *result = ROUTE_EXT_FAIL;
    stats->extractfailures++;
  }

  return (0);
}
/********************************************************************************/
/* extractHSFRoute for extracting the HSF Route
 * Ganga - needs to be fully changed from original BFRoute.
 */
static int
extractHSFRoute (call_t *newcall, int *result)
{
  int	    source;
  int	    extractresult;
  int       **allroutes;
  rsstat_t  *stats;

  source = newcall->source;
  allroutes = newcall->allroutes;

  //  bftables = (bftab_t **) newcall->network->routetable;
  stats = newcall->network->stats;

  if (hsfExtractPath(newcall, &newcall->route, &extractresult)) {
    RS_PRINT_ERROR1 ("extractHSFRoute: hsfExtractPath failed for call %d\n", newcall->callid);
    return (-1);
  }
  newcall->extractions++;
  
  if (extractresult == ROUTE_EXT_SUCCESS) {
    if ((newcall->routelen - 1) > newcall->minhops + hopthresh) {
#if RS_TRACE_LEVEL >= TRACE_DETAILS
      printf ("extractHSFRoute: Route too long for call %d: routelen %d (min: %d)\n",
	      newcall->callid, newcall->routelen, newcall->minhops);
#endif
      *result = ROUTE_EXT_FAIL;
      stats->extractfailures++;
    }
    else {  /* route is within hopcount threshold */
      if (getLinks(newcall->route, newcall->routelen, &newcall->links,
		   newcall->network)) {
	RS_PRINT_ERROR ("extractHSFRoute: getLinks returned failure\n");
	return (-1);
      }
      RS_PRINT_EVENT1 ("extractHSFRoute: Extracted route for call %d\n",newcall->callid);
      *result = ROUTE_EXT_SUCCESS;
    }
  }
  else {  /* could not extract route -- declare failure */
    RS_PRINT_EVENT1 ("extractHSFRoute: Can't find a route for call %d\n", newcall->callid);
    *result = ROUTE_EXT_FAIL;
    stats->extractfailures++;
  }

  return (0);
}






/********************************************************************************/
static int
computeSPTFallback (call_t *newcall, double time, int *result)
/* Compute a minhop SPT when computation or extraction can't find a suitable route
 */
{
  int		source;
  int		calcresult;
  spt_t		**sstrees;
  sptpolicy_t	*sptpolicy;

  source = newcall->source;
  sstrees = (spt_t **) newcall->network->routetable;
  sptpolicy = (sptpolicy_t *) newcall->network->routepolicies;
  /* recompute SPT */
  sptReset (sstrees[source]);
  if (sptExec (sstrees[source], source, newcall->bandw, time, NO_PRUNE,
	       sptpolicy->distfunc, accuracy, &calcresult)) {
    RS_PRINT_ERROR ("computeSPTFallback: sptExec failed\n");
    return (-1);
  }
  if (calcresult == ROUTE_CALC_SUCCESS) {
    sptMultiPathExtractInit (sstrees[source]);	/* reset head pointers */
    *result = ROUTE_CALC_SUCCESS;
  }
  else {
    *result = ROUTE_CALC_FAIL;
  }

  return (0);

}
/********************************************************************************/
int
getLinks (int *route, int routelen, link_t ***links, topology_t *network)
/* Traces the path from source to dest and places the link ptrs in
 * an array.  The links parameter is pointed to this array before
 * returning. Assumes that the route array has been filled in and that
 * routelen is the number of *nodes* in the route.
 */
{
  int		i, numlinks;
  link_t	**pathlinks;

  *links = (link_t **) 0;

  if ( (route == (int *) 0) || (routelen == 0) ){
    RS_PRINT_SOFT_ERROR ("getLinks: Route not initialized\n");
    return (-1);
  }
  numlinks = routelen - 1;
  pathlinks = (link_t **) malloc (sizeof(link_t *) * numlinks);
  if (pathlinks == (link_t **) 0) {
    RS_PRINT_SOFT_ERROR ("getLinks:  Can't allocate links array\n");
    return (-1);
  }

  /* extract links from route array */
  for (i = 0; i < numlinks; i++) {
    pathlinks[i] = network->links[route[i]][route[i+1]];    
  }

  *links = pathlinks;
  return (0);
}  
/********************************************************************************/
int
linkExists (int a, int b, double reqbw, prunepolicy_t policy, staleness_t staleness,
	    topology_t *network)
/* answers the question: is there a link from a to b with the requested bw
 * under the given policy?
 */
{
  double availbw;

  if (network->links[a][b] == (link_t *) 0 || network->links[a][b]->down == 1)
    return (0);
  switch (policy) {
  case NO_PRUNE:
    /* there is always a link */
    return (1);
  case DEFAULT_PRUNE:
    availbw = getLinkAvailBandwidth (network->links[a][b], staleness);
    if (availbw == -INF) {
      RS_PRINT_ERROR ("linkExists: Error in getLinkAvailBandwidth\n");
      return (0);
    }
#if RS_TRACE_LEVEL >= TRACE_DETAILS
    if (availbw < reqbw) 
      printf ("linkExists: pruning insufficient link %d -> %d (%.2f < %.2f)\n",
	      a, b, availbw, reqbw);
#endif      
    return ((availbw >= reqbw) && (network->links[a][b]->active == 1));
    break;
  default:
    RS_PRINT_ALWAYS ("Bad prune policy -- bad bad\n");
    return (0);
  }
}
/********************************************************************************/
double
getLinkCost (link_t *link, distfunction_t func, staleness_t staleness, topology_t *network)
{
  double availbw;
  double occupancy;
  double cost;

  availbw = getLinkAvailBandwidth (link, staleness);
  if (availbw == -INF) {
    RS_PRINT_ERROR ("getLinkCost: Error in getLinkAvailBandwidth\n");
    return ((double) INF);
  }

  occupancy = 1.0 - (availbw / link->capacity);

  switch (func) {
  case HOPCOUNT:
    cost = link->hopcount.current;
    break;
    /* d_i = K + c_i, 1/C <= c_i <= 1, then
     *
     * for an h-hop path:
     *	minimum cost = h(K + 1/C)
     *	maximum cost = h(K + 1)
     *
     * for an h+1 - hop path:
     *	minimum cost = (h+1)(K + 1/C)
     *	maximum cost = (h+1)(K + 1)
     */    
  case EXP_SINGLE_COST:
    cost = calcExpLinkCost (occupancy);
    /* add K term */
    cost += rscostconst;
    break;
  case EXP_MINHOP_BIAS:
    cost = calcExpLinkCost (occupancy);
    /* add N term */
    cost += (double) getNumNodes(network);
    break;
  default:
    RS_PRINT_ALWAYS ("getLinkCost: Unknown distance function\n");
    cost = (double) INF;
    break;
  }
  return (cost);
}

/********************************************************************************/
static double
calcExpLinkCost (double occupancy)
/* computes link cost according to
 *
 *	  --			       --
 *	  | '	         ` alpha	|
 *  c_i = | |(u_i - x_o) |		|
 *	  | |----------- |    * (C - 1)	| + 1	    for u_i > x_o
 *	  | | (1 - x_o)  |		|
 *	  |  `	         '		|
 *	-------------------------------------
 *			C
 *
 *			1/C			    otherwise
 */
{
  double cost;

  if (occupancy < rscostmin)
    /* make sure that for alpha = 0, cost returns 1 no matter what the occupancy */
    rscostalpha == 0.0 ? (cost = 1.0) : (cost = 1.0/(double)rscostlevels);
  else {
    cost = (ceil((pow(((occupancy - rscostmin)/(1.0 - rscostmin)), rscostalpha)) *
		 ((double)rscostlevels - 1.0) ) + 1.0) / (double)rscostlevels;
  }

  return (cost);
  
}
/********************************************************************************/
double
getLinkAvailBandwidth (link_t *link, staleness_t staleness)
/* Subject to bandwidth quantization. We use equal class based updates which partitions
 * link into equal chunks (0,B), (B,2B), (2B,3B)...  Advertised value is the midpoint of this
 * range.
 */
{
  double    bwval, retval;
  int	    classnum;

  /* first get the actual bw value */
  switch (staleness) {
  case CURRENT:
    bwval = link->availbw.current;
    break;
  case STALE:
    bwval = link->availbw.last;
    break;
  default:
    RS_PRINT_ALWAYS ("getLinkAvailBandwidth:  Unknown staleness function\n");
    return (-INF);
  }

  /* now quantize it, if necessary */
  if (bwclasssize > 0) {
    classnum = (int) (bwval / bwclasssize);  /* truncates the result down to the classnum */
    /* compute the midpoint value (average) of the corresponding class */
    retval = ( (classnum * bwclasssize)  + ((classnum + 1) * bwclasssize) ) / 2.0;
    retval = copysign (retval, bwval);
  }
  else {
    /* return the unquantized value */
    retval = bwval;
  }

  return (retval);
    
}

/********************************************************************************/
static int
validRouteExists (int source, topology_t *network)
{
  int		retval = 0;
  routealg_t	routealg;
  spt_t		**sstrees;
  bftab_t	**bftables;

  routealg = network->routingalg;
  switch (routealg) {
  case SPT:
    sstrees = (spt_t **) network->routetable;
    if (sstrees[source]->valid == 1) {
      retval = 1;
    }
    break;
  case WIDE_SHORT:
    bftables = (bftab_t **) network->routetable;
    if (bftables[source]->valid == 1) {
      retval = 1;
    }
    break;
    
    /* dont care for HSF as it is used only in Precomp.
     * Ganga
     */
  case HSF:
    break;

  default:
    RS_PRINT_ALWAYS ("validRouteExists:  Unknown routing algorithm\n");
  }

  return (retval);

}
/********************************************************************************/
double
getLastRouteCompTime (int source, topology_t *network)
{
  routealg_t	routealg;
  spt_t		**sstrees;
  bftab_t	**bftables;  

  routealg = network->routingalg;
  switch (routealg) {
  case SPT:
    sstrees = (spt_t **) network->routetable;
    if (sstrees[source] == (spt_t *) 0) {
      RS_PRINT_ERROR ("getLastRouteCompTime:  Invalid spt\n");
      return (-1.0);
    }
    return (sstrees[source]->comptime);
    break;
  case WIDE_SHORT:
    bftables = (bftab_t **) network->routetable;
    if (bftables[source] == (bftab_t *) 0) {
      RS_PRINT_ERROR ("getLastRouteCompTime:  Invalid bftab\n");
      return (-1.0);
    }
    return (bftables[source]->comptime);
    break;
    /* Dont know how and whether to measure this
     * For now the plan is to just ignore this part of the stats.
     * Ganga
     */
  case HSF:
    return (0.0);
    break;

  default:
    RS_PRINT_ALWAYS ("getLastRouteCompTime:  Unknown routing algorithm\n");
    return (-1.0);
  }
}

int
getNumRouteComps (int source, topology_t *network)
{
  routealg_t	routealg;
  spt_t		**sstrees;
  bftab_t	**bftables;  

  routealg = network->routingalg;
  switch (routealg) {
  case SPT:
    sstrees = (spt_t **) network->routetable;
    if (sstrees[source] == (spt_t *) 0) {
      RS_PRINT_ERROR ("getNumRouteComps:  Invalid spt\n");
      return (-1);
    }
    return (sstrees[source]->computations);
    break;
  case WIDE_SHORT:
    bftables = (bftab_t **) network->routetable;
    if (bftables[source] == (bftab_t *) 0) {
      RS_PRINT_ERROR ("getNumRouteComps:  Invalid bftab\n");
      return (-1);
    }
    return (bftables[source]->computations);
    break;


    /* Not clear how to measure this for HSF
     * Ganga
     */
  case HSF:
    return (0);
    break;

  default:
    RS_PRINT_ALWAYS ("getNumRouteComps:  Unknown routing algorithm\n");
    return (-1);
  }  

}

int
resetRouteComps (int source, topology_t *network)
{
  routealg_t	routealg;
  spt_t		**sstrees;
  bftab_t	**bftables;
  
  routealg = network->routingalg;
  
  switch (routealg) {
  case SPT:
    sstrees = (spt_t **) network->routetable;
    if (sstrees[source] == (spt_t *) 0) {
      RS_PRINT_ERROR ("resetRouteComps:  Invalid spt\n");
      return (-1);
    }
    sstrees[source]->computations = 0;    
    break;
  case WIDE_SHORT:
    bftables = (bftab_t **) network->routetable;
    if (bftables[source] == (bftab_t *) 0) {
      RS_PRINT_ERROR ("resetRouteComps:  Invalid bftab\n");
      return (-1);
    }
    bftables[source]->computations = 0;
    break;

    /* Nothing to do yet for HSF
     * Ganga
     */
  case HSF:
    break;


  default:
    RS_PRINT_ALWAYS ("resetRouteComps:  Unknown routing algorithm\n");
    return (-1);
  }

  return (0);
}
	 
  
/********************************************************************************/
static void
resetCallRoute (call_t *call)
{

  if (call->route != (int *) 0) {
    free (call->route);
    call->route = (int *) 0;
  }
  
  call->routelen = 0;

  if (call->links != (link_t **) 0) {
    free (call->links);
    call->links = (link_t **) 0;
  }
}

/********************************************************************************/
void
printRoutingPolicy (void)
{
  switch (routepolicy) {
  case ON_DEMAND:
    printf ("ON_DEMAND");
    break;
  case PRECOMP:
    printf ("PRECOMP");
    break;
  default:
    printf ("UNKNOWN");
  }
}

void
printRoutingAlgorithm (topology_t *network)
{
  switch (network->routingalg) {
  case SPT:
    printf ("SPT");
    break;
  case WIDE_SHORT:
    printf ("WIDE_SHORT");
    break;
  case HSF:
    printf ("HSF");
    break;
  default:
    printf ("UNKNOWN");
  }
}

void
printAltPolicy (topology_t *network)
{
  sptpolicy_t	*sptpolicy;

  if (network->routepolicies == (void *) 0)
    return;

  sptpolicy = (sptpolicy_t *) network->routepolicies;
  switch (sptpolicy->altpolicy) {
  case ROUND_ROBIN:
    printf ("ROUND_ROBIN");
    break;
  case STICKY:
    printf ("STICKY");
    break;
  case NO_ALT:
    printf ("NO_ALT");
    break;
  default:
    printf ("UNKNOWN");
  }
}

void
printStaleness (void)
{
  switch (accuracy) {
  case CURRENT:
    printf ("CURRENT");
    break;
  case STALE:
    printf ("STALE");
    break;
  default:
    printf ("UNKNOWN");
  }
}

void
printPrunePolicy (topology_t *network)
{
  sptpolicy_t	*sptpolicy;

  if (network->routepolicies == (void *) 0)
    return;

  sptpolicy = (sptpolicy_t *) network->routepolicies;  
  switch (sptpolicy->prunep) {
  case NO_PRUNE:
    printf ("NO_PRUNE");
    break;
  case DEFAULT_PRUNE:
    printf ("DEFAULT_PRUNE");
    break;
  default:
    printf ("UNKNOWN");
  }
}

void
printDistFunction (topology_t *network)
{
  sptpolicy_t	*sptpolicy;

  if (network->routepolicies == (void *) 0)
    return;

  sptpolicy = (sptpolicy_t *) network->routepolicies;
  switch (sptpolicy->distfunc) {
  case HOPCOUNT:
    printf ("HOPCOUNT");
    break;
  case EXP_SINGLE_COST:
    printf ("EXP_SINGLE_COST");
    break;
  case EXP_MINHOP_BIAS:
    printf ("EXP_MINHOP_BIAS");
    break;
  case WIDEST:
    printf ("WIDEST");
    break;    
  default:
    printf ("UNKNOWN");
  }
}

/********************************************************************************/
int
cleanupRouting (topology_t *network)
{
  int	i, numnodes;
  spt_t		**sstrees;
  bftab_t	**bftables;  

  numnodes = getNumNodes(network);
  switch (network->routingalg) {
  case SPT:
    sstrees = (spt_t **) network->routetable;
    for (i = 0; i < numnodes; i++) {
      sptDestroy (sstrees[i]);
    }
    free(sstrees);
    free (network->routepolicies);
    break;
  case WIDE_SHORT:
    bftables = (bftab_t **) network->routetable;
    for (i = 0; i < numnodes; i++) {
      bfDestroy (bftables[i]);
    }
    free (bftables);
    /* WIDE_SHORT doesn't have policies to free */
    break;

    /* Nothing to free for HSF for now
     * Ganga
     */    
    
  case HSF:
    break;
  default:
    RS_PRINT_ALWAYS ("cleanupRouting:  Unknown routing algorithm\n");
    return (-1);
  }

  return (0);
}
