/* $Id: spt.c,v 3.15 1999/05/02 21:14:42 ashaikh Exp $ */

static const char rcsid[] = "$Id: spt.c,v 3.15 1999/05/02 21:14:42 ashaikh Exp $";

/*
*   Description:
*
*	Simulation of the Dijkstra single-source shortest paths algorithm for
*	the construction of shortest-path trees.  These are specific to use
*	with Routesim.
*/

#include "spt.h"
#include "sptgeneric.h"
#include "priorityQ.h"
#include "clinklist.h"
#include "linklist.h"
#include "topology.h"
#include "routesim.h"
#include "rstrace.h"
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <math.h>


static int sptSrcExec (pq_t *, double *, double *, spt_t *, int, double, double,
		       prunepolicy_t, distfunction_t, int *);
static int sptRelax (pq_t *, double *, double *, spt_t *, int, int, distfunction_t, staleness_t);
static int sptGetUniqParent (spt_t *, int, int, uniqpolicy_t, sptparent_t **);
static int checkLinkFeasibility (link_t *, call_t *, boolean);
static int checkRoutingParameters (prunepolicy_t, distfunction_t, staleness_t);

/********************************************************************************/
int
sptExec (spt_t *spt, int source, double bandw, double currenttime,
	 prunepolicy_t policy, distfunction_t func, staleness_t staleness, int *result)
/*  This routine is the main body of the algorithm.  Each node is visited in turn,
 *  and added to the shortest path tree.  At each node distances to all neighbors are
 *  checked, and updated if necessary.  At the end of the routines all nodes have
 *  been added to S.
*/
{
  pqelement_t    *popped, *newel;
  int		current;
  int		neighbor;
  int		i, numnodes;
  double	*keys;	    /* distance keys (estimates) for each node */
  double	*bottleneck; /* max bw avail on path to each node */
  pq_t		*sptQ;	    /* priority Q for network nodes */

  if (checkRoutingParameters(policy, func, staleness)) {
    RS_PRINT_ERROR ("sptExec: bad routing parameter flag\n");
    return (-1);
  }
  numnodes = getNumNodes(spt->network);

  spt->valid = 0;
  spt->source = source;
  sptQ = pqInit(numnodes);
  if (sptQ == (pq_t *) 0) {
    RS_PRINT_SOFT_ERROR ("sptExec: SPT pq allocation failure\n");
    return (-1);
  }

  keys = (double *) malloc (sizeof(double)*numnodes);
  bottleneck = (double *) malloc (sizeof(double)*numnodes);

  if ( (!keys) || (!bottleneck) ) {
    RS_PRINT_SOFT_ERROR ("sptExec: SPT key/bottleneck array allocation failure\n");
    return (-1);
  }
  else {
    for (i = 0; i < numnodes; i++) {
      keys[i] = INF;
      bottleneck[i] = -INF;
    }
  }

  if (source >= numnodes) {
    RS_PRINT_ERROR ("sptExec: Bad source id\n");
    return (-1);
  }

  keys[source] = 0.0;	/* source has zero distance */
  bottleneck[source] = INF; /* and infinite bw to itself */
    
  /* put all the nodes in a heap */
  /* this can be optimized to avoid putting nodes in until they are scanned */
  for (i=0; i < numnodes; i++) {
    newel = (pqelement_t *) malloc (sizeof(pqelement_t));
    newel->id = i;
    newel->key = keys[i];
    newel->element = 0;
    pqAddItem(sptQ, newel);
  }

  /* special treatment for the source */
  if (sptSrcExec (sptQ, keys, bottleneck, spt, source, bandw, currenttime, policy, func, result)) {
    RS_PRINT_SOFT_ERROR ("sptExec: sptSrcExec failed\n");
    sptCleanup (sptQ);
    free (keys);
    free (bottleneck);
    return (-1);
  }
  else if (*result == ROUTE_CALC_FAIL) {  /* routing failed at the source */
    sptCleanup (sptQ);
    free (keys);
    free (bottleneck);    
    return (0);
  }

  /* now the rest of the tree */
  while (!pqIsQEmpty(sptQ)) {
    if (pqRemoveItem (sptQ, &popped)) {	    /* examine node with shortest dist. estimate */
      RS_PRINT_SOFT_ERROR ("sptExec: pqRemoveItem failed\n");
      sptCleanup (sptQ);
      free (keys);
      free (bottleneck);
      return (-1);
    }
    current = popped->id;
    free (popped);
#if RS_TRACE_LEVEL >= TRACE_ALL
    printf ("sptExec: popped %d (%.3f)\n", current, keys[current]);
#endif

    if (keys[current] == INF) {
      /* If current node has never been encountered, it is disconnected from the source. */
      /* Moreover, any other nodes in the queue will have an infinite key and thus are also */
      /* disconnected.  So we can stop now. */
#if RS_TRACE_LEVEL >= TRACE_ALL
      printf ("sptExec: %d is disconnected\n", current);
#endif      
      *result = ROUTE_CALC_DISCONN;
      spt->valid = 1;
      spt->comptime = currenttime;
      spt->computations++;
      sptCleanup (sptQ);
      free (keys);
      free (bottleneck);
      return(0);
    }

    neighbor = -1;
#if RS_TRACE_LEVEL >= TRACE_ALL    
    printf ("\trelaxing:\n");
#endif
    for (i = 0; i < numnodes; i++) {	/* check all neighboring nodes */
      if (spt->network->links[current][i] == (link_t *) 0)
	continue;
      if (processLinkUpdates (currenttime, spt->network->links[current][i], spt->network, "(routing)")) {
	RS_PRINT_ERROR ("sptExec: processLinkUpdates failed\n");
	sptCleanup (sptQ);
	free (keys);
	free (bottleneck);
	return (-1);
      }
      /* check if there is a usable link subject to pruning policy */
      if (linkExists(current, i, bandw, policy, staleness, spt->network)) {
	neighbor = i;
	if (sptRelax (sptQ, keys, bottleneck, spt, current, neighbor, func, staleness)) {
	  RS_PRINT_SOFT_ERROR ("sptExec: sptRelax failed\n");
	  sptCleanup (sptQ);
	  free (keys);
	  free (bottleneck);
	  return (-1);
	}
      }
    }
    
#if RS_TRACE_LEVEL >= TRACE_DETAILS
    if (neighbor < 0)	    /* current node had no neighbors */    
      printf ("sptExec: %d has no neighbors\n", current);
#endif
  
  } /* while (!pqIsQEmpty) */
    
  *result = ROUTE_CALC_SUCCESS;
  spt->valid = 1;
  spt->comptime = currenttime;
  spt->computations++;  
  sptCleanup (sptQ);
  free (keys);
  free (bottleneck);
  return (0);
}
/********************************************************************************/
static int
sptSrcExec (pq_t *sptQ, double *keys, double *bottleneck, spt_t *spt, int source, double bandw,
	    double currenttime, prunepolicy_t policy, distfunction_t func, int *result)
/*  Run the first iteration of sptExec for the source.  Main reason is to allow use
 *  of current link state info for adjacent links
 */
{
  pqelement_t    *popped;
  int	    current;
  int	    neighbor;
  int	    i, numnodes;
 
  if (pqRemoveItem (sptQ, &popped)) {
    RS_PRINT_SOFT_ERROR ("sptSrcExec: pqRemoveItem failed\n");
    return (-1);
  }
  numnodes = getNumNodes(spt->network);

  current = popped->id;
  free (popped);
  if (current != source) {
#if RS_TRACE_LEVEL >= TRACE_SOFT_ERRORS
    fprintf (stderr, "sptSrcExec: popped element (%d) is not source (%d)\n",
	     popped->id, source);
#endif
    return (-1);
  }
  
#if RS_TRACE_LEVEL >= TRACE_ALL
  printf ("sptSrcExec: popped %d (%.3f)\n", current, keys[current]);
  printf ("\trelaxing:\n");
#endif
  neighbor = -1;    

  for (i = 0; i < numnodes; i++) {
    if (spt->network->links[current][i] == (link_t *) 0)
      continue;
    if (processLinkUpdates (currenttime, spt->network->links[current][i],
			    spt->network, "(routing)")) {
      RS_PRINT_ERROR ("processLinkUpdates failed\n");
      return (-1);
    }
    /* source always knows CURRENT info about adjacent links */
    if (linkExists(current, i, bandw, policy, CURRENT, spt->network)) { /* is there a usable link? */
      neighbor = i;
      if (sptRelax (sptQ, keys, bottleneck, spt, current, neighbor, func, CURRENT)) {
	RS_PRINT_SOFT_ERROR ("sptExec: sptRelax failed\n");
	return (-1);
      }      
    }
  }
  if (neighbor < 0) {	    /* current had no neighbors */
#if RS_TRACE_LEVEL >= TRACE_DETAILS
    printf ("sptSrcExec: source has no neighbors\n");
#endif
    *result = ROUTE_CALC_FAIL;
    return (0);    
  }
  *result = ROUTE_CALC_SUCCESS;
  return (0);
}
/********************************************************************************/
static int
sptRelax (pq_t *sptQ, double *keys, double *bottleneck, spt_t *spt, int current, int neighbor,
	  distfunction_t func, staleness_t staleness)
{
  double    key;
  double    minbw, linkbw;

  switch (func) {
  case HOPCOUNT:
  case EXP_MINHOP_BIAS:
  case EXP_SINGLE_COST:
    /* get distance from current to neighbor */
    key = getLinkCost(spt->network->links[current][neighbor], func, staleness, spt->network);  
    
    if (keys[neighbor] > (keys[current] + key)) {	 /* need to update */
      if (pqChangeVal (sptQ, neighbor, (keys[current] + key))) {
      RS_PRINT_SOFT_ERROR ("sptRelax: pqChangeVal failed\n");
      return (-1);
      }
      keys[neighbor] = keys[current] + key;
      /* get rid of any existing parents list */
      if (clearParents (spt, neighbor)) {
	RS_PRINT_SOFT_ERROR ("sptRelax: clearParent failed\n");
	return (-1);
      }
      /* add the new parent ptr */
      if (addParent (spt, neighbor, current)) {
	RS_PRINT_SOFT_ERROR ("sptRelax: addParent failed\n");
	return (-1);
      }
#if RS_TRACE_LEVEL >= TRACE_ALL
      printf ("\t\tmod %d (%.3f)\n", neighbor, keys[neighbor]);
#endif
    }
    else {
      if (keys[neighbor] == (keys[current] + key)) {	 /* equal cost -- add to parent list */
	if (addParent (spt, neighbor, current)) {
	  RS_PRINT_SOFT_ERROR ("sptRelax: addParent failed\n");
	  return (-1);
	}
#if RS_TRACE_LEVEL >= TRACE_ALL      
	printf ("\t\tadd %d (%.3f)\n", neighbor, keys[neighbor]);
#endif
      }
    }
    break;
  case WIDEST:
    /* compute shortest-widest path */
    /* idea is to use hopcount and break ties based on avail bw */
    key = spt->network->links[current][neighbor]->hopcount.current;

    if (keys[neighbor] > (keys[current] + key)) {	 /* need to update */
      if (pqChangeVal (sptQ, neighbor, (keys[current] + key))) {
      RS_PRINT_SOFT_ERROR ("sptRelax: pqChangeVal failed\n");
      return (-1);
      }
      keys[neighbor] = keys[current] + key;

      /* compute min bw to neighbor */
      linkbw = getLinkAvailBandwidth (spt->network->links[current][neighbor], staleness);
      if (bottleneck[current] <= linkbw)
	minbw = bottleneck[current];
      else
	minbw = linkbw;
      bottleneck[neighbor] = minbw;
      
      /* get rid of any existing parents list */
      if (clearParents (spt, neighbor)) {
	RS_PRINT_SOFT_ERROR ("sptRelax: clearParent failed\n");
	return (-1);
      }
      /* add the new parent ptr */
      if (addParent (spt, neighbor, current)) {
	RS_PRINT_SOFT_ERROR ("sptRelax: addParent failed\n");
	return (-1);
      }
#if RS_TRACE_LEVEL >= TRACE_ALL
      printf ("\t\tmod %d (dist %.3f, bw %.3f)\n", neighbor, keys[neighbor], bottleneck[neighbor]);
#endif
    }
    else {
      if (keys[neighbor] == (keys[current] + key)) {	 /* equal distance -- check bw */

	/* compute min bw to neighbor */
	linkbw = getLinkAvailBandwidth (spt->network->links[current][neighbor], staleness);
	if (bottleneck[current] <= linkbw)
	  minbw = bottleneck[current];
	else
	  minbw = linkbw;
	
	if (minbw > bottleneck[neighbor]) { /* this is a wider path */
	  /* get rid of any existing parents list */
	  if (clearParents (spt, neighbor)) {
	    RS_PRINT_SOFT_ERROR ("sptRelax: clearParent failed\n");
	    return (-1);
	  }
	  /* add the new parent ptr */
	  if (addParent (spt, neighbor, current)) {
	    RS_PRINT_SOFT_ERROR ("sptRelax: addParent failed\n");
	    return (-1);
	  }
	  /* update new bottleneck bw */
	  bottleneck[neighbor] = minbw;
#if RS_TRACE_LEVEL >= TRACE_ALL
	  printf ("\t\tmod %d (dist %.3f, bw %.3f)\n", neighbor, keys[neighbor], bottleneck[neighbor]);
#endif	  
	}
	else {
	  if (minbw == bottleneck[neighbor]) {	/* multiple widest-shortest paths? */
	    /* don't need to update bottleneck value -- it's the same */
	    if (addParent (spt, neighbor, current)) {
	      RS_PRINT_SOFT_ERROR ("sptRelax: addParent failed\n");
	      return (-1);
	    }
#if RS_TRACE_LEVEL >= TRACE_ALL      
	    printf ("\t\tadd %d (%.3f)\n", neighbor, keys[neighbor]);
#endif
	  }
	}
      } /* if equal distance neighbor */
    }
    break;
  default:
    RS_PRINT_ALWAYS ("sptRelax:  Unknown distance function\n");
    return (-1);
  }
  return (0);
}
/********************************************************************************/
int
sptExtractMultiPath (spt_t *spt, call_t *newcall, alternate_t altpolicy,
		     boolean feascheck, int **path, int *result)
{
  int		source, dest;
  linklist_t	*pathstack;
  sptparent_t	*nextparent;
  int		*node, *popped;
  int		top, nextid;
  int		pathlen, *treepath;
  int		i;
  int		feasible = 0;

  source = newcall->source;
  dest = newcall->dest;
  *path = (int *) 0;
  *result = ROUTE_EXT_FAIL;

  /* check that the destination has at least one parent pointer */
  if ( (clistSize(spt->parents[dest])) < 1 ) {
#if RS_TRACE_LEVEL >= TRACE_MORE_EVENTS
    printf ("sptExtractMultiPath:  Dest is disconnected (call: %d %d->%d\n",
	    newcall->callid, newcall->source, newcall->dest);
#endif
    *result = ROUTE_EXT_FAIL;
    return (0);
  }
  
  pathstack = listInit();
  node = (int *) malloc (sizeof(int));
  *node = dest;
  listPush (pathstack, (void *) node);	    /* start at dest node */
  clistIteratorAssign (spt->walkptrs[dest], spt->headptrs[dest]);  
  top = dest;
  
  while ( (!listEmpty(pathstack)) && ( (top = *((int *) listGetTop(pathstack))) != source) ) {
    /* explore parents of top node */
    nextparent = (sptparent_t *) clistIteratorGetCurrent(spt->walkptrs[top]);
    nextid = nextparent->parent;
    feasible = checkLinkFeasibility(spt->network->links[nextid][top], newcall, feascheck);
    if (feasible) {	/* test feasibilty */
      node = (int *) malloc (sizeof(int));
      *node = nextid;
      listPush(pathstack, (void *) node);	    /* explore next node in route */
      clistIteratorAssign (spt->walkptrs[nextid], spt->headptrs[nextid]);  /* start at first parent */      
    }
    else { /* link was not feasible -- try another */
      while (clistIteratorGetNext(spt->walkptrs[top]) ==
	     clistIteratorGetCurrent(spt->headptrs[top])) {
	listPop (pathstack, (void **) &popped);	/* backtrack from exhausted parent list */
	free (popped);
	if (!listEmpty(pathstack))
	  top = *((int *) listGetTop(pathstack));
	else
	  break;	
      }
      if (!listEmpty(pathstack)) {
	clistIteratorSetNext(spt->walkptrs[top]);   /* advance to next parent */
      }
    }
  }
  
  if (top == source) {	/* found a route */
    pathlen = listSize(pathstack);
    newcall->routelen = pathlen;
    treepath = (int *) malloc (sizeof(int) * pathlen);	/* allocate for the path */
    i = 0;
    do {
      listPop (pathstack, (void **) &popped);
      treepath[i++] = *popped;
      free (popped);
      top = *((int *) listGetTop(pathstack));      
      switch (altpolicy) {
      case NO_ALT:
	break;
      case ROUND_ROBIN:
	clistIteratorAssign (spt->headptrs[top], spt->walkptrs[top]);
	clistIteratorSetNext (spt->headptrs[top]);
	break;
      case STICKY:
	clistIteratorAssign (spt->headptrs[top], spt->walkptrs[top]);
	break;
      default:
	RS_PRINT_ERROR ("Unknown alternate routing option\n");
	return (-1);
	break;
      }
    }
    while (top != dest);
    listPop (pathstack, (void **) &popped); /* finally, pop off dest */
    treepath[i] = *popped;
    free (popped);
    *path = treepath;	/* set path pointer */
    listDestroy (pathstack);
  }
  else {    /* can't find a route */
#if RS_TRACE_LEVEL >= TRACE_MORE_EVENTS
    printf ("sptExtractMultiPath:  Extraction failed: No route available\n");
#endif
    listDestroy (pathstack);
    *result = ROUTE_EXT_FAIL;
    return (0);
  }

  *result = ROUTE_EXT_SUCCESS;
  return (0);
      
}

/********************************************************************************/
static int
checkLinkFeasibility (link_t *link, call_t *call, boolean feascheck)
/* Check link feasibility based on failed signalling attempts and most recent
 * link state.  We check for links failed in signalling regardless of whether
 * feasibility-checking is on or off.  Link-state checking is done only if
 * feasibility checking is on.
 */
   
{
  int	i;
  int	feasible = 1;

  /* first check that link did not fail in signalling */
  for (i = 0; i < call->sigattempts; i++) {
    if (link == call->failedlinks[i])
      feasible = 0;
  }

  if (feascheck == TRUE) {
    /* now check latest link state for the link */
    if (link->availbw.last < call->bandw)
      feasible = 0;
    else
      feasible = 1;
  }

  return (feasible);
}
/********************************************************************************/
int
sptExtractSinglePath (spt_t *spt, call_t *newcall, boolean feascheck,
		      int **path, int *result)
/* Traces a single path from source to dest and places the node ids in
 * an array.  The path parameter is pointed to this array before
 * returning.  
 */
{
  int	    i, len;
  int	    current, *treepath;
  sptparent_t  *parentptr;
  int	    source, dest;
  int	    feasible;


#if RS_TRACE_LEVEL >= TRACE_ALL
  sptPrintTree(spt);
#endif

  source = newcall->source;
  dest = newcall->dest;

  /* check that the destination has at least one parent pointer */
  if ( (clistSize(spt->parents[dest])) < 1 ) {
#if RS_TRACE_LEVEL >= TRACE_MORE_EVENTS
    printf ("sptExtractSinglePath:  Dest is disconnected (call: %d %d->%d\n",
	    newcall->callid, newcall->source, newcall->dest);
#endif
    *result = ROUTE_EXT_FAIL;
    return (0);
  }

  *path = (int *) 0;
  *result = ROUTE_EXT_FAIL;
  len = sptGetPathLen (spt, source, dest);
  newcall->routelen = len;
  
  if (len < 0)
    return (-1);

  treepath = (int *) malloc (sizeof(int) * len);
  if (treepath == (int *) 0)
    return (-1);
  for (i = 0; i < len; i++) {
    treepath[i] = -1;
  }

  current = dest;
  i = len - 1;
#if RS_TRACE_LEVEL >= TRACE_ALL  
  printf ("Extracting:\n");
#endif

  while (current != source) {

    treepath[i--] = current;    /* add current node to the path */

    /* now grab first parent of current node */
    if (clistGetFirstElement (spt->parents[current], (void **) &parentptr))
      return (-1);
    
    /* check link feasibility between parent and current */
    feasible = checkLinkFeasibility(spt->network->links[parentptr->parent][current], newcall, feascheck);

    if (feasible) {	

#if RS_TRACE_LEVEL >= TRACE_ALL	/* some elaborate tracing for multiparent */
      sptparent_t   *traceptr;
      int	    j;
      if (clistSize(spt->parents[current]) > 1) {
	printf ("%d ( ", treepath[i-1]);
	for (j = 1; j <= clistSize(spt->parents[current]); j++) {
	  clistGetElement (spt->parents[current], j, (void **) &traceptr);
	  printf ("%d ", traceptr->parent);
	}
	printf (")\n");
      }
      else {
	printf ("%d\n", treepath[i-1]);
      }
#endif

      /* go on to parent */
      current = parentptr->parent;
    }
    else {  /* link was infeasible */
      /* no chance to backtrack so we just fail here */
      *result = ROUTE_EXT_FAIL;
      free (treepath);
      return (0);
    }
  }
  treepath[i] = current;

#if RS_TRACE_LEVEL >= TRACE_ALL  
  printf ("%d\n", current);
#endif

  *path = treepath;
  *result = ROUTE_EXT_SUCCESS;
  return (0);
}
/********************************************************************************/
int
sptExtractUniqPath (spt_t *spt, call_t *newcall, int src, int dst, uniqpolicy_t uniqpolicy,
		    boolean feascheck, int **path, int *result)
/* Extracts a path, picking among multiple parent pointers based on
 * source node id -- no feasibility checking.
 */
{
  int		source, dest;
  int		current, *node;
  sptparent_t	*parentptr;
  linklist_t	*treepath;
  int		*newpath, pathlen;
  int		feasible;
  int		i;


  if (newcall == (call_t *) 0)	{ /* call unspecified */
    source = src;
    dest = dst;
  }
  else {    /* use the call spec */
    source = newcall->source;
    dest = newcall->dest;
  }

  if (spt->valid == 0) {
    *result = ROUTE_EXT_FAIL;
    RS_PRINT_ERROR ("sptExtractUniqPath:  spt is invalid\n");
    return (-1);
  }

  /* check that the destination has at least one parent pointer */
  if (clistSize(spt->parents[dest]) < 1) {
#if RS_TRACE_LEVEL >= TRACE_MORE_EVENTS
    printf ("sptExtractSinglePath:  Dest is disconnected %d->%d\n", source, dest);
#endif
    *result = ROUTE_EXT_FAIL;
    return (0);
  }
  
  treepath = listInit();
  current = dest;
  
  while (current != source) {
    /* add current node to the path */    
    node = (int *) malloc (sizeof(int));
    *node = current;
    listInsert (treepath, (void *) node);

    /* get the next parent according to the specified policy */
    if (sptGetUniqParent (spt, source, current, uniqpolicy, &parentptr)) {
      RS_PRINT_ERROR ("sptExtractSinglePath:  sptGetUniqParent returned error\n");
      *result = ROUTE_EXT_FAIL;      
      return (-1);
    }

    if (newcall == (call_t *) 0) {
      current = parentptr->parent;
    }
    else {
      /* check link feasibility between parent and current */
      feasible = checkLinkFeasibility(spt->network->links[parentptr->parent][current], newcall, feascheck);
      if (feasible) {
	current = parentptr->parent;
      }
      else {  /* link was infeasible */
	/* no chance to backtrack so we just fail here */
	*result = ROUTE_EXT_FAIL;
	/* delete the path list */
	for  (i = 1; i <= listSize(treepath); i++) {
	  listGetElement (treepath, i, (void **) &node);
	  free (node);
	}
	listDestroy (treepath);	
	return (0);
      }      
    }
  }
  
  if (listEmpty (treepath))
    return (-1);
  /* add the source node */
  node = (int *) malloc (sizeof(int));
  *node = current;
  listInsert (treepath, (void *) node);
  
  pathlen = listSize (treepath);
  newpath = (int *) malloc (sizeof(int) * pathlen);
  for  (i = 1; i <= pathlen; i++) {
    listGetElement (treepath, i, (void **) &node);
    newpath[i - 1] = *node;
    free (node);
  }
  listDestroy (treepath);
  
  *path = newpath;
  if (newcall != (call_t *) 0)
    newcall->routelen = pathlen;
  
  *result = ROUTE_EXT_SUCCESS;
  
  return (0);
  
}
/********************************************************************************/
static int
sptGetUniqParent (spt_t *spt, int source, int current, uniqpolicy_t uniqpolicy,
		  sptparent_t **parentptr)
{
  int	chosen = -1, numparents;

  /* how many parents does the current node have? */
  numparents = clistSize (spt->parents[current]);

  if (numparents > 1) {
    switch (uniqpolicy) {
    case SRCHASH:
      chosen = source % numparents; /* a simple hash function based on source id */
      chosen++;   /* because parents are indexed starting at 1, not 0 */
      break;
    case RANDOM:
      chosen = (int) floor( 1.0 + (((double)numparents - 1.0 + 1.0)*drand48()) );
      break;
    default:
      RS_PRINT_ALWAYS ("sptGetUniqParent: Invalid uniq route policy!\n");
      return (-1);
    }
    /* grab chosen parent */
    if (clistGetElement (spt->parents[current], chosen, (void **) parentptr)) {
      RS_PRINT_ERROR ("sptExtractUniqPath:  clistGetElement returned error\n");
      return (-1);
    }
  }
  else {  /* only one parent */
    /* now grab first parent of current node */
    if (clistGetFirstElement (spt->parents[current], (void **) parentptr)) {
      RS_PRINT_ERROR ("sptExtractUniqPath:  clistGetFirstElement returned error\n");	
      return (-1);
    }
  }

  return (0);
}
/********************************************************************************/
void
sptPrintUniqPath (spt_t *spt, int dest, uniqpolicy_t uniqpolicy)
{
  int	*path, result;
  int	*nodeptr;

  if (sptExtractUniqPath (spt, (call_t *) 0, spt->source, dest, uniqpolicy, FALSE, &path, &result)) {
    RS_PRINT_ERROR ("sptPrintUniqPath:  sptExtractUniqPath failed\n");
    return;
  }

  printf ("%d -> %d: ", spt->source, dest);
  for (nodeptr = path; *nodeptr != dest; nodeptr++) {
    printf ("%d ", *nodeptr);
  }
  printf ("%d\n", *nodeptr);

  free (path);
}

void
sptPrintFirstPath (spt_t *spt, int dest)
{
  int	*path;
  int	*nodeptr;

  if (sptExtractFirstPath (spt, spt->source, dest, &path)) {
    RS_PRINT_ERROR ("sptPrintUniqPath:  sptExtractUniqPath failed\n");
    return;
  }

  printf ("%d -> %d: ", spt->source, dest);
  for (nodeptr = path; *nodeptr != dest; nodeptr++) {
    printf ("%d ", *nodeptr);
  }
  printf ("%d\n", *nodeptr);

  free (path);
}
  
/********************************************************************************/
static int
checkRoutingParameters (prunepolicy_t policy, distfunction_t func, staleness_t staleness)
{
  if ((policy != NO_PRUNE) &&
      (policy != DEFAULT_PRUNE)) {
    return (-1);
  }

  if ((func != HOPCOUNT) &&
      (func != EXP_SINGLE_COST) &&
      (func != EXP_MINHOP_BIAS) &&
      (func != WIDEST)) {
    return (-1);
  }

  if ((staleness != CURRENT) &&
      (staleness != STALE)) {
    return (-1);
  }

  return (0);
}

/********************************************************************************/
