/* $Id: topkncube.c,v 1.4 1998/12/07 18:55:08 ashaikh Exp $ */

#include "topkncube.h"
#include "topology.h"
#include <math.h>
#include <stdlib.h>

static int calcKaryNcubeDist (int ,int, int, int);
static int calcKaryNcubeNumLinks (int, int);
static double calcKaryNcubeHopcount (int, int);
static int calcKaryNcubeDiameter (int, int);
static double calcKaryNcubeDegree (int);
static int calcKaryNcubeDistTable (int, int, topology_t *);
/********************************************************************************/
int
genKaryNcube(int k, int n, topology_t **network)
/*  compute k-ary n-cube neighbors and generate the corresponding links in
 *  the adjacency matrix.
 */
{
  int		dimension,
		below, above, neighbor,
		i, j;
  int		numneighbors;
  topology_t	*net;
  link_t	*link;

  if (k < 3 || n == 0) {
    RS_PRINT_SOFT_ERROR ("Invalid k-ary n-cube parameters\n");
    return (-1);
  }

  if (initNetwork(&net, TOP_KNCUBE)) {
    RS_PRINT_ERROR ("genKaryNcube:  initNetwork failed\n");
    return (-1);
  }

  net->numnodes = (int) pow((double)k,(double)n);

  if (initLinkMatrix (net)) {
    RS_PRINT_SOFT_ERROR ("Error in initMatrix\n");
    return (-1);
  }
  
  /* Sequence through the nodes to compute edges */
#if RS_TRACE_LEVEL >= TRACE_DETAILS
  printf ("%d nodes in %d-ary %d-cube\n", net->numnodes, k, n);
  printf ("Generating neighbors ...\n");
#endif
  for (i=0; i < net->numnodes; i++) {
    numneighbors = 0;
    /* Sequence through the dimensions */
    for (j = 0; j < n; j++) {
      /* Isolate address fields above, at, and below digit j */
      dimension = (int) pow((double)k,(double)j);
      below = i % dimension;
      above = (i/(k*dimension))*(k*dimension);
      
      /* Adjacent nodes +1 in dimension j */
      neighbor = above + (((i/dimension)+1)%k)*dimension + below;
      link =  genLink(i, neighbor, 0);
      if (link == (link_t *) 0) {
	RS_PRINT_ERROR ("genKaryNcube: Link alloc failed\n");
	return (-1);
      }
      if (genLinkVariables (link)) {
	RS_PRINT_ERROR ("genKaryNcube: genLinkVariables returned error\n");
	return (-1);
      }
      net->links[i][neighbor] = link;
      net->numlinks++;
      numneighbors++;
#if RS_TRACE_LEVEL >= TRACE_ALL
      printf("%d\t%d\n", i, neighbor);
#endif
      
      /* Adjacent nodes -1 in dimension j */
      neighbor = above + (((i/dimension)+k-1)%k)*dimension + below;
      link =  genLink(i, neighbor, 0);
      if (link == (link_t *) 0) {
	RS_PRINT_ERROR ("genKaryNcube: Link alloc failed\n");
	return (-1);
      }
      if (genLinkVariables (link)) {
	RS_PRINT_ERROR ("genKaryNcube: genLinkVariables returned error\n");
	return (-1);
      }      
      net->links[i][neighbor] = link;      
      net->numlinks++;
      numneighbors++;
#if RS_TRACE_LEVEL >= TRACE_ALL
      printf("%d\t%d\n", i, neighbor);
#endif      
    }
    updateSimpleRatioStat (numneighbors, &(net->degree));
  }
#if RS_TRACE_LEVEL >= TRACE_DETAILS
  printf ("Done\n");
#endif

  /* fill in the rest of the topology parameters */

  if (calcKaryNcubeDistTable (k, n, net)) {
    RS_PRINT_SOFT_ERROR ("genKaryNcube: Hoptable calculation failed\n");
    return (-1);
  }

  net->diameter = calcKaryNcubeDiameter(k, n);

  /* just to double check */
  if (net->diameter != net->hopcount.max) {
    RS_PRINT_ALWAYS ("Mismatch in diameter calculation\n");
    return (-1);
  }      
  if (net->numlinks != calcKaryNcubeNumLinks(k, n)) {
    RS_PRINT_ALWAYS ("Mismatch in numlinks calculation\n");
    return (-1);
  }
  if (net->degree.currentrate != calcKaryNcubeDegree (n)) {
     RS_PRINT_ERROR ("genKaryNcube:  Avg node degree mismatch\n");
     return (-1);
  }


  net->info = (void *) 0;

  calcLinkCapacityStats(net);

  *network = net;
  return (0);
}

/********************************************************************************/
static int
calcKaryNcubeNumLinks (int k, int n)
{
  return (2 * n * (int)pow ((double)k,(double)n));
}

static double
calcKaryNcubeHopcount (int k, int n)
{
  double hopcount;

  k % 2 == 0 ? (hopcount = k / 4.0 * n) :
    (hopcount = (k*k - 1.0) / 4.0 / k * n);
  hopcount *= ((pow((double)k,(double)n)) / (pow((double)k,(double)n) - 1.0)) ;
  return (hopcount);
  
}

static int
calcKaryNcubeDiameter (int k, int n)
{
  return ( (k / 2) * n );
}

static double
calcKaryNcubeDegree (int n)
{
  return ( (2.0 * n) );
}
/********************************************************************************/
static int
calcKaryNcubeDistTable (int k, int n, topology_t *net)
{
  int	i, j;

  if (net == (topology_t *) 0)
    return (-1);

  net->hoptable = (int **) malloc (sizeof(int) * net->numnodes);
  if (net->hoptable == (int **) 0)
    return (-1);

  for (i = 0; i < net->numnodes; i++) {
    net->hoptable[i] = (int *) malloc (sizeof(int) * net->numnodes);
    if (net->hoptable[i] == (int *) 0)
      return (-1);
    for (j = 0; j < net->numnodes; j++) {
      if (i == j) {
	net->hoptable[i][j] = 0;
      }
      else {
	net->hoptable[i][j] = calcKaryNcubeDist (k, n, i,j);
	updateSimpleRatioStat (net->hoptable[i][j], &(net->hopcount));	
      }
    }
  }

  return (0);
}
/********************************************************************************/
static int
calcKaryNcubeDist (int k, int n, int src, int dest)
/*  mostly stolen from ppsim code */
{
  int	i, delta_1, delta_2;
  int	offsetsum = 0;

  for (i=0; i < n; i++) {

    /* determine offset */
    delta_1   = mod((dest - src), k);
    delta_2   = mod((src - dest), k);

    if (delta_1 < delta_2)
      offsetsum += delta_1;
    else
      offsetsum += delta_2;

    /* prepare for next iteration */
    src  /= k;
    dest /= k;
  }

  return (offsetsum);
}

