/* $Id: toprandom.c,v 1.8 1999/06/10 17:32:37 ashaikh Exp $ */

#include "toprandom.h"
#include "dfs.h"
#include "rstrace.h"
#include "topology.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>


static double	**eucliddists;

static int genRandEdgesSym (randgraph_t, double, double, double, topology_t *);
static int genRandEdgesAsym (randgraph_t, double, double, double, topology_t *);
static int genEuclidDistMatrix(double, topology_t *);
static int initEuclidDistMatrix (topology_t *);
static void destroyEuclidDistMatrix (topology_t *);
static int processRandGraph (topology_t *net);
/********************************************************************************/
int
genRandTopology (randgraph_t method, int topsize, double alpha, double beta, double planesize,
		 boolean asym, int maxdiam, topology_t **network)
{
  topology_t	*net;
  int		connected = 0;
  int		bigdiameter = 1;
  int		tries = 0;
  randgraph_t	*m;
  
  if (topsize == 0) {
    RS_PRINT_ERROR ("Topology size not initialized\n");
    return (-1);
  }

  if (initNetwork(&net, TOP_RAND)) {
    RS_PRINT_ERROR ("genRandTopology:  initNetwork failed\n");
    return (-1);
  }
  net->numnodes = topsize;

  if (alpha <= 0 || beta > 1) {
    RS_PRINT_ERROR ("genRandTopology:  Invalid random graph parameters\n");
    return (-1);
  }

  if (initLinkMatrix (net)) {
    RS_PRINT_SOFT_ERROR ("Error in initMatrix\n");
    return (-1);
  }

  if (initDistTable (net)) {
    RS_PRINT_ERROR ("genRandTopology: Error in initDistTable()\n");
    return (-1);
  }

  if ( initEuclidDistMatrix(net) ) {	/* allocate distance matrix */
    RS_PRINT_ERROR ("genRandTopology: Error in initEuclidDistMatrix()\n");
    return (-1);
  }
  if ( genEuclidDistMatrix(planesize, net) ) { /* fill in the distance matrix */
    RS_PRINT_ERROR ("genRandTopology: Error in genEuclidDistMatrix()\n");
    return(-1);
  }


  do {
    /* generate edges */
    if (asym == FALSE) {
      if (genRandEdgesSym (method, alpha, beta, planesize, net)) {  /* generate links */
	RS_PRINT_ERROR ("genRandTopology: Error in genRandEdges()\n");
	destroyEuclidDistMatrix (net);	
	return (-1);
      }
    } else {
      if (genRandEdgesAsym (method, alpha, beta, planesize, net)) {  /* generate links */
	RS_PRINT_ERROR ("genRandTopology: Error in genRandEdges()\n");
	destroyEuclidDistMatrix (net);	
	return (-1);
      }
    }
    /* check connectivity */
    connected = isStronglyConnected(net->links, net->numnodes); /* check connectivity */
    if (!connected) {
      RS_PRINT_EVENT1 ("genRandTopology: Graph disconnected (try %d)\n", tries);
      clearLinkMatrix(net);
    }
    else {
      /* graph is connected -- calculate and check diameter */
      if (processRandGraph (net)) {
	RS_PRINT_ERROR ("genRandTopology: processRandGraph() failed\n");
	return (-1);
      }
      bigdiameter = (net->diameter > maxdiam);
      if (bigdiameter) {
	RS_PRINT_EVENT2 ("genRandTopology: Graph diameter %d exceeds max allowed (try %d)\n",
			 net->diameter, tries);
	clearLinkMatrix(net);	
	clearDistTable (net);
      }
    }
    tries++;
    
  } while ((!connected || bigdiameter) && (tries < MAX_GRAPH_GEN_TRIES));

  destroyEuclidDistMatrix (net);    

  if (tries == MAX_GRAPH_GEN_TRIES) {
    RS_PRINT_SOFT_ERROR ("genRandTopology: Failed to generate a connected graph\n");
    return (-1);
  }

  /* got a good graph -- do some more preprocessing on it */

#if RS_TRACE_LEVEL >= TRACE_DETAILS
  printLinkMatrix(net);
#endif

  calcLinkCapacityStats(net);
  m = (randgraph_t *) malloc (sizeof(randgraph_t));
  *m = method;
  net->info = (void *) &method;

  *network = net;
  return (0);  
}
/********************************************************************************/
static int
processRandGraph (topology_t *net)
{

  if (calcDistTable(net)) {
    RS_PRINT_ERROR ("processRandGraph: Error in calcDistTable()\n");
    return (-1);
  }

  if (parseDistTable (net)) {
    RS_PRINT_ERROR ("processRandGraph:  Error in parseDistTable()\n");
    return (-1);
  }

  return (0);
}
/********************************************************************************/
  
static int
genRandEdgesAsym (randgraph_t method, double alpha, double beta, double planesize, topology_t *net)
/* Generates single unidirectional edges independently.  That is, existence of edge (u,v)
 * does not necessarily imply existence of edge (v,u).
 */
{
  int	    i, j;
  int	    dist;
  double    maxdist, p;
  link_t    *link;

  maxdist = planesize * sqrt(2.0);
  for (i = 0; i < net->numnodes; i++) {
    for (j = 0; j < net->numnodes; j++) {
      if (j == i) continue; /* no self edges */
      switch (method) {
      case RG2:
	dist = (int) rint(eucliddists[i][j]);
	/* Waxman's RG2 method */
	p = alpha * exp (-dist/beta/maxdist);
	break;
      case PURE:
	p = alpha;
	break;
      default:
	RS_PRINT_ERROR ("genRandEdges:  Unknown graph method\n");
	return (-1);
      }
      if (drand48() < p) {
	link =  genLink(i, j, 0);
	if (link == (link_t *) 0) {
	  RS_PRINT_ERROR ("genRandEdges:  Error in genLink()\n");
	  return (-1);
	}
	if (genLinkVariables (link)) {
	  RS_PRINT_ERROR ("genRandEdges: genLinkVariables returned error\n");
	  return (-1);
	}
	net->links[i][j] = link;
	net->numlinks++;
      }
    }
  }
  
  return (0);
}
/********************************************************************************/
static int
genRandEdgesSym (randgraph_t method, double alpha, double beta, double planesize, topology_t *net)
/* Generates a pair of unidirectional edges between a node-pair, (u,v), with a
 * a given probability.  Existence of edge (u,v) implies existence of edge (v,u).
 */
{
  int	    i, j;
  int	    dist;
  double    maxdist, p;
  link_t    *link1, *link2;

  maxdist = planesize * sqrt(2.0);
  for (i = 0; i < net->numnodes; i++) {
    for (j = i + 1; j < net->numnodes; j++) {
      switch (method) {
      case RG2:
	dist = (int) rint(eucliddists[i][j]);
	/* Waxman's RG2 method */
	p = alpha * exp (-dist/beta/maxdist);
	break;
      case PURE:
	p = alpha;
	break;
      default:
	RS_PRINT_ERROR ("genRandEdges:  Unknown graph method\n");
	return (-1);
      }
      if (drand48() < p) {
	link1 =  genLink(i, j, 0);
	link2 =  genLink(j, i, 0);	
	if ( (link1 == (link_t *) 0) || (link2 == (link_t *) 0) ) {
	  RS_PRINT_ERROR ("genRandEdges:  Error in genLink()\n");
	  return (-1);
	}
	if ( genLinkVariables(link1) || genLinkVariables(link2) ) {
	  RS_PRINT_ERROR ("genRandEdges: genLinkVariables returned error\n");
	  return (-1);
	}
	net->links[i][j] = link1;
	net->links[j][i] = link2;
	net->numlinks += 2;	/* generated two edges (i,j) and (j,i) */
      }
    }
  }
  
  return (0);
}
/********************************************************************************/
static int
genEuclidDistMatrix (double planesize, topology_t *net)
{
  int	    i, j;
  double    maxdist, dist;

  maxdist = planesize * sqrt(2.0);
  for (i = 0; i < net->numnodes; i++) {
    for (j = i + 1; j < net->numnodes; j++) {
      dist = drand48() * maxdist;
      eucliddists[i][j] = dist;
      eucliddists[j][i] = dist;
    }
  }
#if RS_TRACE_LEVEL >= TRACE_DETAILS
  for (i = 0; i < net->numnodes; i++) {
    for (j = 0; j < net->numnodes; j++) {
      printf ("%.3d ", (int) rint(eucliddists[i][j]));
    }
    printf ("\n");
  }
#endif

  return (0);
}
/********************************************************************************/
static int
initEuclidDistMatrix (topology_t *net)
{
  int	i, j;

  eucliddists = (double **) malloc (sizeof(double *) * net->numnodes);
  if (eucliddists == (double **) 0) {
    RS_PRINT_ERROR ("initEuclidDistMatrix:  Couldn't allocate Euclidean dist matrix\n");
    return (-1);
  }
  for (i = 0; i < net->numnodes; i++) {
    eucliddists[i] = (double *) malloc (sizeof(double) * net->numnodes);
    if (eucliddists[i] == (double *) 0) {
      RS_PRINT_ERROR ("initEuclidDistMatrix:  Couldn't allocate Euclidean dist matrix entry\n");      
      return (-1);
    }
    for (j = 0; j < net->numnodes; j++)
      eucliddists[i][j] = 0.0;
  }

  return (0);
}
/********************************************************************************/
static void
destroyEuclidDistMatrix (topology_t *net)
{
  int	i;

  for (i = 0; i < net->numnodes; i++) {
    free (eucliddists[i]);
  }

  free (eucliddists);
}
  






