#include "sptgeneric.h"
#include "linklist.h"
#include "clinklist.h"
#include "stat.h"
#include <stdlib.h>

/*
 * Description:
 *
 * This program prints all of the shortest paths between a single src-dest pair.
 * The topology is read on stdin and must be in routesim format.
 *
 */

topology_t  *network;

spt_t		*spt;	    /* source-specific shortest-path trees */
int		source, destn;


typedef struct {
  double    capacity;
  int	    id;
} pathstacknode_t;

static int computeShortestPaths (void);
static int traversePaths (spt_t *, int);
static void traverseVisit (spt_t *, int, int);
static int traverseShortestPaths (spt_t *, int);
static void printPathStack (linklist_t *);
/********************************************************************************/

int
main (int argc, char **argv)
{

  if (argc != 3) {
    printf ("argc = %d\n", argc);
    fprintf (stderr, "usage:\t%s\t<source> <dest>\n", "printpath"); 
    exit (1);
  }

  if (readTopology (&network)) {
    RS_PRINT_ERROR ("readTopology failed\n");
    exit (1);
  }

  source = (int) strtol (argv[1], (char **) NULL, 10);
  destn = (int) strtol (argv[2], (char **) NULL, 10);

  if (computeShortestPaths()) {
    RS_PRINT_ERROR ("computeShortestPaths failed\n");
    return (-1);
  }

  traversePaths (spt, destn);

}

/********************************************************************************/
static int
computeShortestPaths (void)
/* Compute shortest-path trees based on hopcount for all nodes.  Uses the multiple-parent
 * Dijkstra modifications.
 */
{
  

  /* initialize spt data structure */
  if (sptInit (&spt, network)) {
    RS_PRINT_ERROR ("computeShortestPaths: SPT allocation failed\n");
    return (-1);
  }

  sptReset (spt);
  if (sptHopMultiExec(spt, source)) {
    RS_PRINT_ERROR ("computeShortestPaths: sptHopMultiExec failed\n");
    return (-1);
  }
  spt->source = source;
  sptMultiPathExtractInit (spt);    /* reset head pointers */

  return (0);
}
/********************************************************************************/
static int
traversePaths (spt_t *spt, int dest)
/* Performs a recursive depth-first traversal of the shortest-path graph.
 */   
{
  int	    source;
  int	    i, numnodes;


  source = spt->source;
  numnodes = getNumNodes(spt->network);

  printf ("Traversing from %d -> %d\n", dest, source);


  /* set walkptr to headptr for all nodes */
  for (i = 0; i < getNumNodes(spt->network); i++) {
    clistIteratorAssign (spt->walkptrs[i], spt->headptrs[i]);
  }

  traverseShortestPaths (spt, dest);
  
  return (0);
}

static void
traverseVisit (spt_t *spt, int current, int finaldest)
/* Recursive function to follow parents of each node in the shortest-path
 * graph.
 */
{
  int	source, numparents;
  sptparent_t	*nextparent;

  source = spt->source;

  if (current == source) {
    printf (" %d\n", current);
    return;
  }

  /* not yet at the source */
  numparents = clistSize (spt->parents[current]);

  do {	/* visit each parent */
    nextparent = (sptparent_t *) clistIteratorGetCurrent(spt->walkptrs[current]);
    printf (" %d", nextparent->parent);
    traverseVisit (spt, nextparent->parent, finaldest);
    /* advance to next parent */        
    clistIteratorSetNext(spt->walkptrs[current]);
  }
  /* while walkptr != headptr */
  while (clistIteratorGetCurrent(spt->walkptrs[current]) !=
	 clistIteratorGetCurrent(spt->headptrs[current]));
}

/********************************************************************************/
static int
traverseShortestPaths (spt_t *spt, int dest)
/* A non-recursive routine to traverse the shortest-path graph.
 */
{
  int		    source;
  linklist_t	    *pathstack;
  sptparent_t	    *nextparent;
  pathstacknode_t   *node, *popped;
  int		    top, nextid;

  source = spt->source;

  pathstack = listInit();
  node = (pathstacknode_t *) malloc (sizeof(pathstacknode_t));
  node->id = dest;
  listPush (pathstack, (void *) node);	    /* start at dest node */
  /*  printf ("pushing %d\n", node->id); */
  clistIteratorAssign (spt->walkptrs[dest], spt->headptrs[dest]);  
  top = dest;
  
  while (!listEmpty(pathstack)) {
    while ( ((top = ((pathstacknode_t *)listGetTop(pathstack))->id) != source) ) {
      /* explore parents of top node */
      nextparent = (sptparent_t *) clistIteratorGetCurrent(spt->walkptrs[top]);
      nextid = nextparent->parent;
      /* allocate a new node */
      node = (pathstacknode_t *) malloc (sizeof(pathstacknode_t));
      node->id = nextid;
      node->capacity = 1.0 / (clistSize (spt->parents[top]));
      listPush(pathstack, (void *) node);	    /* explore next node in route */
      /*       printf ("pushing %d\n", node->id); */
      clistIteratorAssign (spt->walkptrs[nextid], spt->headptrs[nextid]);  /* start at first parent */
    }
    /* reached the source -- pop back to last branch point */
    printPathStack (pathstack);    
    listPop (pathstack, (void **) &popped);
    /*    printf ("popped %d\n", popped->id);     */
    free (popped);
    if (!listEmpty(pathstack)) {
      top = ((pathstacknode_t *)listGetTop(pathstack))->id;  /* reset top */
      while (clistIteratorGetNext(spt->walkptrs[top]) ==
	     clistIteratorGetCurrent(spt->headptrs[top])) {
	listPop (pathstack, (void **) &popped);	/* backtrack from exhausted parent list */
	free (popped);
	if (!listEmpty(pathstack))
	  top = ((pathstacknode_t *)listGetTop(pathstack))->id;  /* reset top */
	else
	  break;	
      }
      /* reached an unexplored branch point -- or stack went empty */
      if (!listEmpty(pathstack)) {
	clistIteratorSetNext(spt->walkptrs[top]);   /* advance to next parent */
      }
      else break;	/* exhausted all parents */
    }
  }

  return (0);
}

/********************************************************************************/
static void
printPathStack (linklist_t *pathstack)
{
  int	size, i;
  pathstacknode_t   *node;

  size = listSize (pathstack);

  for (i = 1; i <= size; i++) {
    listGetElement (pathstack, i, (void **) &node);
    printf ("%d ", node->id);
  }
  printf ("\n");
}
