/* $Id: clinklist.c,v 1.2 1998/02/10 21:48:49 ashaikh Exp $ */

/*
*   Description:
*
*	Implementation of a simple circular linked list data structure.
*	Also supports stack-like functions, i.e. push/pop on lists.
*
*
*		 -------	 -------	 -------
*		| 	|------>|	|	|	|<---- last
*	     -->|	|	|	|------>|	|---
*	    |	 -------    	 -------	 -------    |
*	    |  		    				    |
*	     -----------------------------------------------
*
*
*
*
*/

#include "clinklist.h"
#include <stdio.h>
#include <stdlib.h>


/********************************************************************************/
clinklist_t
*clistInit (void)
{
  clinklist_t	*newlist;

  newlist = (clinklist_t *) malloc (sizeof(clinklist_t));
  if (newlist == (clinklist_t *) 0)
    return ((clinklist_t *) 0);

  newlist->last = (clistnode_t *) 0;    /* empty list last point to NULL */
  newlist->numelements = 0;

  return (newlist);
}
/********************************************************************************/
int
clistInsert (clinklist_t *list, void *newelement)
/* Insert newelement in the front of the list
 */     
{
  clistnode_t	*newnode;
  
  if ((list == (clinklist_t *) 0) ||	/* check for null pointers */
      (newelement == (void *) 0))
    return (-1);

  newnode = (clistnode_t *) malloc (sizeof (clistnode_t));
  if (newnode == (clistnode_t *) 0)
    return (-1);

  newnode->element = newelement;

  if (list->last == (clistnode_t *) 0) {    /* need special case for empty list */
    list->last = newnode;
    newnode->next = newnode;
  }
  else {
    newnode->next = list->last->next;
    list->last->next = newnode;
  }

  list->numelements++;

  return (0);
}

int
clistAttach (clinklist_t *list, void *newelement)
/* Append newelement to the end of the list
 */
{
  clistnode_t	*newnode;
  

  if ((list == (clinklist_t *) 0) ||
      (newelement == (void *) 0))
    return (-1);

  newnode = (clistnode_t *) malloc (sizeof (clistnode_t));
  if (newnode == (clistnode_t *) 0)
    return (-1);

  newnode->element = newelement;

  if (list->last == (clistnode_t *) 0) {  /* check for empty list */
    list->last = newnode;
    newnode->next = newnode;
  }
  else {
    newnode->next = list->last->next;
    list->last->next = newnode;
    list->last = newnode;
  }
  
  list->numelements++;

  return(0);
}

int
clistPush (clinklist_t *list, void *newelement)
/*  Pushes newelement onto the list as if it were a stack
 */
{
  return (clistInsert(list, newelement));
}

/********************************************************************************/
int
clistPop (clinklist_t *list, void **element)
/* Removes first element from the list and sets a pointer to it
 */
{
  clistnode_t	*popped;

  *element = (void *) 0;

  if (list == (clinklist_t *) 0)
    return (-1);

  if (list->last == (clistnode_t *) 0)	/* empty list */
    return (-1);

  popped = list->last->next;
  *element = popped->element;
  
  if (list->last == list->last->next) {	/* only one element */
    list->last = (clistnode_t *) 0;
  }
  else {
    list->last->next = popped->next;
  }

  free(popped);

  list->numelements--;

  return (0);
}

int
clistDestroy (clinklist_t *list)
/*  Deletes a link list, including all nodes.  Does NOT delete element
 *  pointers -- that's the responsibility of the list owner.
 */
{
  clistnode_t	*ptr, *del;

  if (list == (clinklist_t *) 0)
    return (-1);

  /* check that list has elements in it */
  if (list->last != (clistnode_t *) 0) {
    for (ptr = list->last->next, del = list->last;
	 ptr != list->last;) {
      del = ptr;
      ptr = ptr->next;
      free (del);
    }
    free (ptr);
  }

  free (list);
  
  return (0);
}
    
/********************************************************************************/
int
clistSize (clinklist_t *list)
{
  if (list == (clinklist_t *) 0)
    return (-1);
  else
    return (list->numelements);
}

int
clistEmpty (clinklist_t *list)
{
  if (list == (clinklist_t *) 0)
    return (-1);
  else
    return (list->numelements == 0);
}
/********************************************************************************/
int
clistGetElement(clinklist_t *list, int index, void **element)
/*  A list iterator function when used with listSize().  Sets element to the index_th
 *  element in the list but doesn't remove it from the list.
 */
{
  clistnode_t	*ptr;
  int		i;

  if (list == (clinklist_t *) 0)
    return (-1);

  *element = (void *) 0;

  if ((index < 1) || (index > list->numelements)) {
    return (-1);
  }

  for (ptr = list->last, i = 0;
       i < index;
       ptr = ptr->next, i++);

  *element = ptr->element;
  return (0);
}

int
clistGetFirstElement (clinklist_t *list, void **element)
/*  Special case to get the first element in the list
 */
{
  if ((list == (clinklist_t *) 0) || (list->numelements < 1))
    return (-1);

  *element = list->last->next->element;

  return (0);
}

/********************************************************************************/
clistiterator_t *
clistIteratorInit (clinklist_t *list)
{
  clistiterator_t   *newiterator;
  
  if (list == (clinklist_t *) 0)	/* check for valid list */
    return ((clistiterator_t *) 0);

  newiterator = (clistiterator_t *) malloc (sizeof(clistiterator_t));
  if (newiterator == (clistiterator_t *) 0)
    return ((clistiterator_t *) 0);

  newiterator->list = list;
  if (newiterator->list->last == (clistnode_t *) 0) /* set current to NULL for empty list */
    newiterator->current = (clistnode_t *) 0;
  else
    newiterator->current = list->last->next;

  return (newiterator);
}

int
clistIteratorReset (clistiterator_t *iterator)
{
  if (iterator->list == (clinklist_t *) 0)	/* check for valid list */
    return (-1);

  iterator->current = (clistnode_t *) 0;
  
  return (0);
}

int
clistIteratorSetFirst (clistiterator_t *iterator)
{
  if (iterator->list->last != (clistnode_t *) 0) {  /* disallow for empty list */
    iterator->current = iterator->list->last->next;
    return (0);
  }
  else {
    iterator->current = (clistnode_t *) 0;
    return (-1);
  }
}

void *
clistIteratorGetCurrent (clistiterator_t *iterator)
{
  if (iterator->current != (clistnode_t *) 0) {	/* check that current is valid */
    return (iterator->current->element);
  }
  else 
    return ((void *) 0);
}

int
clistIteratorSetNext (clistiterator_t *iterator)
{
  if (iterator->list->last != (clistnode_t *) 0) {  /* disallow for empty list */
    iterator->current = iterator->current->next;
    return (0);
  }
  else {
    iterator->current = (clistnode_t *) 0;
    return (-1);
  }

}

int
clistIteratorAssign (clistiterator_t *dest, clistiterator_t *src)
{
  if (dest->list != src->list)	/* make sure iterators are for the same list */
    return (-1);
  
  dest->current = src->current;
  return (0);
}

void *
clistIteratorGetNext (clistiterator_t *iterator)
{
  if (iterator->current != (clistnode_t *) 0) {	/* check that current is valid */
    return (iterator->current->next->element);
  }
  else 
    return ((void *) 0);
}

int
clistIteratorDestroy (clistiterator_t *iterator)
{
  free (iterator);
  return (0);
}

/********************************************************************************/
void
clistPrint (clinklist_t *list)
/* A pretty useless routine that will print out pointer addresses of the list
 * elements -- maybe useful for debugging.
 */
{
  clistnode_t	*ptr;
  int		i;

  if ( (list == (clinklist_t *) 0) || (list->last == (clistnode_t *) 0) )
    return;

  for (ptr = list->last->next, i = 0;
       ptr != list->last;
       ptr = ptr->next, i++) {
    printf ("node[%d]: %p\n", i, ptr->element);
  }
  printf ("node[%d]: %p\n", i, ptr->element);

}

/********************************************************************************/
