/* $Id: linklist.c,v 1.4 1998/02/10 21:49:21 ashaikh Exp $ */

/*
*   Description:
*
*	Implementation of a simple linked list data structure.  Also supports
*	stack-like functions, i.e. push/pop on lists.
*
*/

#include "linklist.h"
#include <stdio.h>
#include <stdlib.h>


typedef struct listnode_t {
  void			*element;
  struct listnode_t	*next;
} listnode_t;


/********************************************************************************/
linklist_t
*listInit (void)
{
  linklist_t	*newlist;

  newlist = (linklist_t *) malloc (sizeof(linklist_t));
  if (newlist == (linklist_t *) 0)
    return ((linklist_t *) 0);
  
  newlist->head = (listnode_t *) malloc (sizeof(listnode_t));
  newlist->tail = (listnode_t *) malloc (sizeof(listnode_t));

  if ((newlist->head == (listnode_t *) 0) ||
      (newlist->tail == (listnode_t *) 0)) {
    free (newlist);
    return ((linklist_t *) 0);
  }

  (newlist->head)->next = newlist->tail;
  (newlist->head)->element = (void *) 0;
  (newlist->tail)->next = newlist->tail;
  (newlist->tail)->element = (void *) 0;
  newlist->numelements = 0;

  return (newlist);
}
/********************************************************************************/
int
listInsert (linklist_t *list, void *newelement)
/* Insert newelement in the front of the list, i.e. after head
 */     
{
  listnode_t	*newnode;
  
  if ((list == (linklist_t *) 0) ||
      (newelement == (void *) 0))
    return (-1);

  newnode = (listnode_t *) malloc (sizeof (listnode_t));
  if (newnode == (listnode_t *) 0)
    return (-1);

  newnode->element = newelement;
  newnode->next = list->head->next;
  list->head->next = newnode;
  list->numelements++;

  return (0);
}

int
listAttach (linklist_t *list, void *newelement)
/* Append newelement to the end of the list, i.e. before tail
 */
{
  listnode_t	*newnode, *ptr;
  

  if ((list == (linklist_t *) 0) ||
      (newelement == (void *) 0))
    return (-1);

  newnode = (listnode_t *) malloc (sizeof (listnode_t));
  if (newnode == (listnode_t *) 0)
    return (-1);

  newnode->element = newelement;
  for (ptr = list->head; ptr->next != list->tail; ptr = ptr->next);
  ptr->next = newnode;
  newnode->next = list->tail;
  list->numelements++;

  return(0);
}

int
listPush (linklist_t *list, void *newelement)
/*  Pushes newelement onto the list as if it were a stack
 */
{
  return (listInsert(list, newelement));
}

/********************************************************************************/
int
listPop (linklist_t *list, void **element)
/* Removes first element from the list and sets a pointer to it
 */
{
  listnode_t	*ptr;

  *element = (void *) 0;

  if (list == (linklist_t *) 0)
    return (-1);

  if (list->numelements == 0)
    return (-1);

  ptr = list->head->next;
  list->head->next = ptr->next;
  
  *element = ptr->element;
  free(ptr);
  list->numelements--;

  return (0);
}

int
listDestroy (linklist_t *list)
/*  Deletes a link list, including all nodes.  Does NOT delete element
 *  pointers -- that's the responsibility of the list owner.
 */
{
  listnode_t	*ptr, *del;

  if (list == (linklist_t *) 0)
    return (-1);

  for (ptr = list->head->next, del = list->head;
       ptr != list->tail;) {
       del = ptr;
       ptr = ptr->next;
       free (del);
  }

  free (list->head);
  free (list->tail);
  free (list);
  
  return (0);
}
    
       

  

/********************************************************************************/
int
listSize (linklist_t *list)
{
  if (list == (linklist_t *) 0)
    return (-1);
  else
    return (list->numelements);
}

int
listEmpty (linklist_t *list)
{
  if (list == (linklist_t *) 0)
    return (-1);
  else
    return (list->numelements == 0);
}
/********************************************************************************/
int
listGetElement(linklist_t *list, int index, void **element)
/*  A list iterator function when used with listSize().  Sets element to the index_th
 *  element in the list but doesn't remove it from the list.
 */
{
  listnode_t	*ptr;
  int		i;

  if (list == (linklist_t *) 0)
    return (-1);

  *element = (void *) 0;

  if ((index < 1) || (index > list->numelements)) {
    return (-1);
  }

  for (ptr = list->head, i = 0;
       i < index;
       ptr = ptr->next, i++);

  if (ptr == list->tail) {	/* shouldn't happen with the index check but just in case */
    return (-1);
  }
    
  *element = ptr->element;
  return (0);
}

int
listGetFirstElement (linklist_t *list, void **element)
/*  Special case to get the element pointed to by head->
 */
{
  if ((list == (linklist_t *) 0) || (list->numelements < 1))
    return (-1);

  *element = list->head->next->element;

  return (0);
}

void *
listGetTop (linklist_t *list)
{
  if ((list != (linklist_t *) 0) && (list->numelements > 0)) {
    return (list->head->next->element);
  }
  else
    return ((void *) 0);
}
     
  
/********************************************************************************/
void
listPrint (linklist_t *list)
/* A somewhat useless routine that will print out pointer addresses of the list
 * elements.
 */
{
  listnode_t	*ptr;
  int		i;

  if (list == (linklist_t *) 0)
    return;
  
  for (ptr = list->head->next, i = 0; ptr != list->tail; ptr = ptr->next, i++) {
    printf ("node[%d]: %p\n", i, ptr->element);
  }

}

/********************************************************************************/
int
listReverse (linklist_t *list)
{
  listnode_t	*ptr, *prev, *fwd;

  if (list == (linklist_t *) 0)
    return (-1);
  
  for (ptr = list->head->next, fwd = ptr, prev = list->tail; \
	 fwd != list->tail; \
	 fwd = ptr->next, ptr->next = prev, prev = ptr, ptr = fwd);

  list->head->next = prev;

  return(0);
}
