/* $Id: routesim.c,v 5.1 1999/06/02 20:53:18 ashaikh Exp $ */

static const char rcsid[] = "$Id: routesim.c,v 5.1 1999/06/02 20:53:18 ashaikh Exp $";

#include "routesim.h"
#include "rsopt.h"
#include "rsroute.h"
#include "rsevents.h"
#include "opt-parse.h"
#include "rstopology.h"
#include "rstraffic.h"
#include "rsstat.h"
#include "rslog.h"
#include "rslinklog.h"
#include "rstrace.h"
#include "rsmail.h"
#include <stdlib.h> 
#include <stdio.h>
#include <iostream.h>
#include <math.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <strings.h>

static int initRoutesim (void);
static int cleanupRoutesim (void);
static int getNewCall (int, int, double, topology_t *, call_t **);
static termination_t *getCallTermination (call_t *);
static int signalCall (call_t *, int *);
static int teardownCall (call_t *, double, int *);
static int updateLinkStats (double, link_t *);
static event_t *getNewEvent (double, eventtype_t, void *);
static int eventHandler (event_t *, eventQ&);
static int newCallEvHandler (call_t *, eventQ&, int * );
static int newCallRouteHandler (call_t *, double, int *);
static int newCallSigHandler (call_t *, double, int *);
static int terminationEvHandler (termination_t *, eventQ&);
static int recompEvHandler (event_t *, eventQ&);
static int callRerouteEvHandler (termination_t *, eventQ&);
static int createNewCall (int, int, double, topology_t *, eventQ&);
static int destroyCall (call_t *);
static int createInitEvents (eventQ& );
static int createCallEvent (call_t *, eventQ&);
static int createCallTermEv (call_t *, eventQ& );
static int createNewSrcDestCall (int, int, double, topology_t *, eventQ&);
static int copyCallRequest (call_t *, call_t **);
static double calcBWChange (link_t *, double);
void setFailedLinksInactive (call_t *);
void resetFailedLinks (call_t *);

static void printStats (void);
static int startActualSim (double);
static int stopWarmup (topology_t *);
static int stopSim (topology_t *);
static void updateSimLog (int, double, topology_t *);

static void printRoutesimVersion (void);

#if RS_TRACE_LEVEL >= TRACE_EVENTS
static void printSignalCode (int, FILE *);
static void printEvent (event_t *);
#endif

#if RS_TRACE_LEVEL >= TRACE_ALL
static void printCallTable (int);
#endif
/********************************************************************************/

/* Global variables */
extern netmodel_t	netmodel;
extern topology_t	*rsnetwork;
extern topology_t	*staticnetwork;
extern topology_t	*dynamicnetwork;

/* Some routing parameters */
boolean periodicrecomps = (boolean) FALSE;	/* use periodic route recomputations */
double	recompinterval = 0.0;			/* period at which trees are recomputed */
double	recompskew = 0.0;			/* skew on the recomputation period */
boolean reextsigfailure = (boolean) FALSE;	/* whether to reextract new route on sig failure */
extern boolean trigrecompsig;			/* trigger route recomputations on sig failures */
extern boolean prunerecomp;			/* removed failed links from top db on recomp */
extern boolean recompafterfail;			/* recompute even when crankbacks are exhausted */

/* Traffic pattern */
extern arrivalspec_t arrspec;
double	flowtrig = 0.0;

/* Signalling */
blockingpolicy_t	    blkpolicy;	/* calls block or not */
int	maxsigattempts = MAX_SIG_ATTEMPTS_PER_CALL;

/* Link updates */
double	trigger = 0.0;		/* bandwidth change percentage that triggers a link update */
double	minupdinterval = 0.0;	/* hold-down time interval between updates */
double	refreshinterval = 0.0;	/* period at which link state updates are refreshed */
double  refreshskew = 0.0;	/* skew on period expressed as percentage of base period */

/* Simulation control */
double globaltime = 0.0;
int minrequests = 0;		/* number of requests in experiment	*/
int minwarmup = 0;		/* min number of calls in warmup */
double minsimtime = 0.0;	/* min simulation time to run */
double warmupproportion = 0.0;	/* proportion of warmup that sim must run */
long seedval = 0;		/* random number generator seed		*/

boolean dumptop = (boolean) FALSE;	/* should we dump topology to a file */
boolean printtop = (boolean) FALSE;
boolean monitorlink = (boolean) FALSE;
int	rslogid = -1;
int	linkmonlogid = -1;
boolean printversion = (boolean) FALSE;
int	mailreport = 0;

/* these tolerances specify how close the mean must be to the edges of the confidence
 * interval before stopping either the sim or warmup periods
 */
double conflevel = 0.0;	    /* confidence level */
double simtolerance = 0.0;
double warmuptolerance = 0.0;

int callscreated = 0;
int warmupcalls = 0;
int actualcalls = 0;
int warmup;		/* flag to indicate whether we're in warmup */
double warmuptime = 0.0;
double actualtime = 0.0;
int routesimcomplete;	/* flag to indicate we can stop running */
int nextcallid = 0;	/* call id counter */

proportionstat_t  overutilstat;
proportionstat_t  underutilstat;
proportionstat_t  violutilstat;
proportionstat_t  overutilflowstat;
proportionstat_t  underutilflowstat;
proportionstat_t  violutilflowstat;
double		  woverutil;
double		  wunderutil;
double		  wviolutil;
int		  reroutedcalls;
int		  rerouteattempts;

#if RS_TRACE_LEVEL >= TRACE_ALL
int **calltable;
#endif
/********************************************************************************/
int
main(int argc, char *argv[])
{
  static  eventQ	evq;	    /* primary event queue */
  eventptr_t	nextev;

  ios::sync_with_stdio();   /* sync up the io streams for C and C++ */
  
  /* Process input */
  createOptions();
  parse_command_line (argc, argv);
  if (printversion == TRUE) {
    printRoutesimVersion ();
    exit (0);
  }
  /*   checkInputParms(); */

  /* set up other data structures */
  if (initRoutesim()) {
    RS_PRINT_ALWAYS ("initRoutesim failed.  Exiting...\n");
    if (mailreport) {
      if (sendRSNotification (RS_ERROR_NOTICE)) {
	RS_PRINT_ERROR ("Could not send notification\n");
      }
    }
    exit (1);
  }
  
  RS_PRINT_GROSS_EVENT ("Routesim initialization complete\n");

  /* start warmup phase */
  warmup = 1;
  routesimcomplete = 0;
  rsWriteLog (rslogid, "Warming up ...\n", 0);
  
  /* generate initial set of call arrivals */
  /* note that we don't count them toward calls created */
  if (createInitEvents(evq)) {
    RS_PRINT_ERROR ("Error creating bootstrap events: createInitEvents returned error\n");
    if (mailreport) {
      if (sendRSNotification (RS_ERROR_NOTICE)) {
	RS_PRINT_ERROR ("Could not send notification\n");
      }
    }    
    exit (1);
  }

  /* core event loop */
  while (1) {
    if (evq.empty()) {
      RS_PRINT_GROSS_EVENT ("Events exhausted.\n");
      break;
    }
    nextev = evq.top();
    if (nextev == (eventptr_t) 0) {
      RS_PRINT_ALWAYS ("Error removing event from queue\n");
      if (mailreport) {
	if (sendRSNotification (RS_ERROR_NOTICE)) {
	  RS_PRINT_ERROR ("Could not send notification\n");
	}
      }      
      exit (1);
    }    
    evq.pop();

    globaltime = nextev->time;
    if (warmup)
      warmuptime = nextev->time;
    else
      actualtime = nextev->time;
    if (eventHandler(nextev, evq)) {
      RS_PRINT_ERROR ("eventHandler returned error. Exiting ...\n");
      if (mailreport) {
	if (sendRSNotification (RS_ERROR_NOTICE)) {
	  RS_PRINT_ERROR ("Could not send notification\n");
	}
      }      
      exit (1);
    }
  } /* end while */

  if (cleanupRoutesim()) {
    fprintf (stderr, "cleanupRoutesim returned error\n");
    if (mailreport) {
      if (sendRSNotification (RS_ERROR_NOTICE)) {
	RS_PRINT_ERROR ("Could not send notification\n");
      }
    }    
    exit (1);
  }

  return (0);

}
/********************************************************************************/
static int
initRoutesim (void)
/* Set up some data structures based on sim parameters
 */
{
  int	diameter;

#if RS_TRACE_LEVEL >= TRACE_DETAILS
  dump_state();
#endif

  if (sendRSNotice())
    mailreport = 1;

  /* init random num generator */
  srand48(seedval);

  /* generate topology */
  if (genRSTopology()) {
    fprintf (stderr, "Error generating topology\n");
    return (-1);
  }

  if (initRSRouting()) {
    RS_PRINT_ERROR ("initRoutesim:  initRSRouting returned error\n");
    return (-1);
  }

  if (dumptop == TRUE)
    /* dump the reference network regardless of network model (i.e. SINGLE/DUAL) */
    if (dumpTopologyFile(OUTTOPFILENAME, rsnetwork)) {
      RS_PRINT_ERROR ("Error writing topology to file\n");
      return (-1);
    }

  if (printtop == TRUE) {
    printTopologyStats (rsnetwork);
    exit (0);
  }

  /* initialize network traffic data structs */
  if (initRSTraffic(getNumNodes(rsnetwork))) {
    RS_PRINT_ERROR ("initRSTraffic failed\n");
    return (-1);
  }

  /* diameter better be computed by now */
  diameter = getDiameter(rsnetwork);
  if (diameter == 0) {
    RS_PRINT_ERROR ("Network diameter not initialized!\n");
    return (-1);
  }

  switch (netmodel) {
  case SINGLE:
    if ( (initRoutesimStats(conflevel, rsnetwork)) ||
	 (resetLinkStatistics (0.0, conflevel, rsnetwork)) ) {
      RS_PRINT_ERROR ("Error initializing stats\n");
      return (-1);
    }
    break;
  case DUAL:
    if ( (initRoutesimStats(conflevel, staticnetwork)) ||
	 (resetLinkStatistics (0.0, conflevel, staticnetwork)) ||
	 (initRoutesimStats (conflevel, dynamicnetwork)) ||
	 (resetLinkStatistics (0.0, conflevel, dynamicnetwork)) ) {
      RS_PRINT_ERROR ("Error initializing stats\n");
      return (-1);
    }
    /* reset max sigattempts very large to allow unrestricted rerouting */
    maxsigattempts = DEFAULT_REROUTE_ATTEMPTS;
    break;
  default:
    RS_PRINT_ERROR ("Unknown netmodel\n");
    return (-1);
  }


#if RS_TRACE_LEVEL >= TRACE_ALL
  calltable = (int **) malloc (sizeof(int *) * getNumNodes(rsnetwork));
  if (calltable == (int **) 0) {
    RS_PRINT_ERROR ("Error allocating calltable\n");
    return (-1);
  }
  for (i = 0; i < numnodes; i++) {
    calltable[i] = (int *) malloc (sizeof(int) * getNumNodes(rsnetwork));
    if (calltable[i] == (int *) 0) {
      RS_PRINT_ERROR ("Error allocating calltable row\n");
      return (-1);
    }
    for (j = 0; j < getNumNodes(rsnetwork); j++)
      calltable[i][j] = 0;
  }
#endif

  /* set up logs */
  rslogid = rsInitLog (RSLOGFILE);
  if (rslogid < 0) {
    fprintf (stderr, "Error in rsInitLog\n");
    return (-1);
  }

  if (monitorlink == TRUE) {
    linkmonlogid = rsInitLog (LINKMONLOGFILE);
    if (linkmonlogid < 0) {
      fprintf (stderr, "Error in rsInitLog\n");
      return (-1);
    }
  }

  return (0);

}
/********************************************************************************/

static int
cleanupRoutesim (void)
{
  switch (netmodel) {
  case SINGLE:
    cleanupRSStats (rsnetwork);
    cleanupRSTraffic ();
    cleanupRouting (rsnetwork);
    break;
  case DUAL:
    cleanupRSStats (staticnetwork);
    cleanupRSStats (dynamicnetwork);    
    cleanupRSTraffic ();
    cleanupRouting (staticnetwork);
    cleanupRouting (dynamicnetwork);
    break;
  default:
    RS_PRINT_ALWAYS ("cleanupRoutesim:  Unknown netmodel!\n");
    return (-1);
  }

  return (0);
  
}

/********************************************************************************/

static int
eventHandler (event_t *ev, eventQ& evq)
{
  int		retval;
  int		returncode;

  retval = -1;

#if RS_TRACE_LEVEL  >= TRACE_EVENTS
  printEvent (ev);
#endif

  switch (ev->type) {
  case STARTCALL:	    /* call creation */
    collectReqCallStats ((call_t *) ev->info);
    if (newCallEvHandler((call_t *) ev->info, evq, &returncode)) {
      RS_PRINT_ERROR ("eventHandler: new call event handler failed\n");
      retval = -1;
    }
    else 
      retval = 0;
    free (ev);
    break;	
  case ENDCALL:		/* call termination */
    if (terminationEvHandler((termination_t *) ev->info, evq)) {
	RS_PRINT_ERROR ("eventHandler: termination event handler failed\n");
      retval = -1;
    }
    else 
      retval = 0;
    free (ev);
    break;
  case RECOMP:	/* routing tree recomputation */
    if (recompEvHandler(ev, evq)) {
      RS_PRINT_ERROR ("eventHandler: recomputation event handler failed\n");
      retval = -1;
    }
    else 
      retval = 0;
    break;      
  default:
    RS_PRINT_ERROR ("eventHandler: Unknown event type\n");
    retval = -1;
    break;
  }
  return (retval);
}

/********************************************************************************/
static int
newCallEvHandler (call_t *call, eventQ& evq, int *returncode)
{
  int		routecode = 0;
  int		signalcode = 0;
  double	time;
  int		source, dest, reroute;
  int		retval;
  rsstat_t	*stats;
  topology_t	*network;

  *returncode = REQUEST_FAILURE;
  time = globaltime;	/* set to current time, not call->gentime */
  /* save these values since call may be freed */
  source = call->source;
  dest = call->dest;
  network = call->network;
  stats = network->stats;
  reroute = call->reroute;  

  /* try to find a route for the call */
  if (newCallRouteHandler (call, time, &routecode)) {
    RS_PRINT_ERROR ("newCallEvHandler: Error routing call\n");
    return (-1);
  }
  if (routecode == ROUTING_FAILURE) {
    RS_PRINT_GROSS_EVENT1 ("newCallEvHandler: Call %d routing failed\n", call->callid);
    stats->routefailures++;
    /* no need to continue with signalling */
  }
  else {    /* got a route -- try signalling */
    RS_PRINT_GROSS_EVENT1 ("newCallEvHandler:  Found a route for call %d.  Signalling ...\n",
			   call->callid);
    if (newCallSigHandler (call, time, &signalcode)) {
      RS_PRINT_ERROR ("newCallEvHandler: Error routing call\n");
      return (-1);
    }
    if (signalcode != SIGNALLING_FAILURE) { /* signalling was successful */
      /* successful call */
      RS_PRINT_GROSS_EVENT1 ("newCallEvHandler:  Call %d signalling success (no recomp)\n",
			     call->callid);
    }
    else if (trigrecompsig == TRUE) { /* signalling failed -- recompute tree */
      RS_PRINT_GROSS_EVENT1 ("newCallEvHandler: Call %d initial signalling failed\n",
			     call->callid);
      if (call->sigattempts < maxsigattempts) {
	/* still have some signalling attempts left -- recompute the tree */
	/* and try to signal again */
	if (prunerecomp == FALSE) {
	  if (reComputeRoute (call->source, time, call->bandw, network, &routecode)) {
	    RS_PRINT_ERROR ("newCallEvHandler:  reComputeRoute returned failure\n");
	    return (-1);
	  }
	}
	else {
	  setFailedLinksInactive (call);	/* make sure routing ignores failed links */
	  if (reComputeRoute (call->source, time, call->bandw, network, &routecode)) {
	    RS_PRINT_ERROR ("newCallEvHandler:  reComputeRoute returned failure\n");
	    return (-1);
	  }
	  resetFailedLinks (call);    /* reset links in topology database */
	}
	call->routecomps++;
	if (updateSimpleRatioStat (time - getLastRouteCompTime(source, network), &(stats->rcompinterval))) {
	  RS_PRINT_ERROR ("newCallEvHandler: Error updating updinterval\n");
	}	
	if (routecode == ROUTE_CALC_FAIL) {  /* no new route available  -- block */
	  RS_PRINT_GROSS_EVENT1 ("newCallEvHandler: Call %d second route recomp failed\n",
				 call->callid);
	  stats->sigfailures++;
	}
	else {	/* route recomp was okay -- try to extract and signal again */
	  RS_PRINT_EVENT2 ("newCallEvHandler: Recomputed route for source %d, call %d\n",
			   call->source, call->callid);
	  if (extractRoute (call, time, &routecode)) {
	    RS_PRINT_ERROR ("newCallEvHandler:  extractRoute returned failure\n");
	    return (-1);
	  }
	  if (routecode == ROUTE_EXT_SUCCESS) {
	    routecode = ROUTING_SUCCESS_R;
	    if (newCallSigHandler (call, time, &signalcode)) {
	      RS_PRINT_ERROR ("newCallEvHandler: Error routing call\n");
	      return (-1);
	    }
	    if (signalcode != SIGNALLING_SUCCESS) {
	      RS_PRINT_GROSS_EVENT1 ("newCallEvHandler: Call %d final signalling failed\n", call->callid);
	      stats->sigfailures++;
	    }
	    else {    /* finally success on signalling attempt after recomp */
	      /* successful call */
	      RS_PRINT_GROSS_EVENT2 ("newCallEvHandler: Call %d signalling success (%d tries)\n",
			       call->callid, call->sigattempts);
	    }
	  }
	  else { /* extraction failed */
	    RS_PRINT_EVENT1 ("newCallEvHandler: Call %d second route extraction failed\n",
			     call->callid);
	    routecode = ROUTING_FAILURE;
	    stats->sigfailures++;
	  }
	}
      }
      else { /* out of crankback attempts -- block */
	if (recompafterfail == TRUE) {   /* should we recompute for "the next guy"? */
	  if (reComputeRoute (call->source, time, call->bandw, network, &routecode)) {
	    RS_PRINT_ERROR ("newCallEvHandler:  reComputeRoute returned failure\n");
	    return (-1);
	  }
	  call->routecomps++;	  	  
	  if (updateSimpleRatioStat (time - getLastRouteCompTime(source, network), &stats->rcompinterval)) {
	    RS_PRINT_ERROR ("newCallEvHandler: Error updating updinterval\n");
	  }
	  RS_PRINT_EVENT2 ("newCallEvHandler: Recomputed route for source %d, call %d\n",
			   call->source, call->callid);	  
	}
	RS_PRINT_GROSS_EVENT1 ("newCallEvHandler: Call %d signalling attempts exhausted\n", call->callid);
	stats->sigfailures++;
      }
    } /* else if trigrecompsig == TRUE */
    else {  /* sig failed and no triggered recomputation */
      RS_PRINT_GROSS_EVENT1 ("newCallEvHandler: Call %d signalling failed\n", call->callid);
      stats->sigfailures++;
    }
  } /* else routing was successful */

  if (signalcode == SIGNALLING_SUCCESS && routecode != ROUTING_FAILURE) {
    *returncode = REQUEST_SUCCESS;    
    collectSuccCallStats (call);
    /* make a termination event for this call */
    if (createCallTermEv (call, evq)) {
      RS_PRINT_SOFT_ERROR ("newCallEvHandler:  Error in createCallTermEv\n");
      return (-1);
    }
  }
  else {    /* call failed -- free call */
    collectFailCallStats(call);    
    RS_PRINT_GROSS_EVENT1 ("newCallEvHandler: Call %d failed\n", call->callid);
    destroyCall (call);
  }

  /* in any case we create a new call for this source */
  /* depending on how arrivals are specified --
   * either per-src calls, or per-(src,dest) calls */
  /* but only if this is not a rerouted call! */
  if (reroute == 0) {
    switch (arrspec) {
    case ARR_UNIFORM:
      retval = createNewCall (source, -1, time, network, evq);
      break;
    case ARR_FILE:
      retval = createNewCall (source, dest, time, network, evq);
      break;
    }
    if (retval) {
      RS_PRINT_ERROR ("newCallEvHandler: createNewCall failed\n");
      return (-1);
    }
  }
  return (0);
}
/********************************************************************************/
static int
newCallRouteHandler (call_t *call, double time, int *routecode)
{
  int	code;

  if (routeCall (call, time, &code)) {
    RS_PRINT_ERROR ("newCallRouteHandler: routeCall returned error\n");
    return (-1);
  }

  *routecode = code;
  
  return (0);
}
/********************************************************************************/
static int
newCallSigHandler (call_t *call, double time, int *returncode)
{
  int	sigcode;
  int	routecode;

  *returncode = SIGNALLING_FAILURE;

  while (call->sigattempts < maxsigattempts) { 
    if (signalCall (call, &sigcode)) {
      RS_PRINT_ERROR ("newCallSigHandler: Error signalling call\n");
      return (-1);
    }
    if (sigcode == SIG_SUCCESS) {   /* signalling succeeded */
      *returncode = SIGNALLING_SUCCESS;
      break;
    }      
    else if (reextsigfailure == TRUE &&
	     call->sigattempts < maxsigattempts) {
      /* signalling failed -- reextract but only if we're going to signal again */      
#if RS_TRACE_LEVEL >= TRACE_EVENTS
      printf ("newCallSigHandler: Call %d signalling failed (attempt %d): ",
	      call->callid, call->sigattempts);
      printSignalCode (sigcode, stdout);
#endif
      /* when signalling fails, we just extract another route, using the failed */
      /* link info now stored in the call. We don't recompute the tree here. */
      if (extractRoute (call, time, &routecode)) {
	RS_PRINT_ERROR ("newCallSigHandler: extractRoute failed\n");
	return (-1);
      }
      if (routecode == ROUTE_EXT_FAIL) {  /* can't find another route */
	RS_PRINT_EVENT1 ("newCallSigHandler:  Reextraction for call %d failed -- no more routes\n",
			 call->callid);
	*returncode = SIGNALLING_FAILURE;
	break;  /* exit while loop */
      }
      RS_PRINT_EVENT1 ("newCallSigHandler:  Successful reextraction for call %d\n", call->callid);
    } /* if (reextsigfailure == TRUE) */
    else {  /* sig failed and no retry */
#if RS_TRACE_LEVEL >= TRACE_EVENTS      
      printf ("newCallSigHandler: Call %d signalling failed: ",
	      call->callid);
      printSignalCode (sigcode, stdout);
#endif      
      break;
    }
  } /* while sigattempts < max */

  return (0);
}
/********************************************************************************/
static int
createNewCall (int source, int dest, double time, topology_t *network, eventQ &evq)
{
  char		logbuf[75];
  
  /* make a new call event for this source */
  if (warmup) {
    if (stopWarmup(network)) {
#ifdef PRINT_WARMUP_STATS
      printf ("Warmup statistics:\n");
      printStats ();
      fflush (stdout);
#endif /* PRINT_WARMUP_STATS */          
      if (startActualSim(time)) {
	RS_PRINT_ERROR ("createNewCall: Error in startActualSim\n");
	return (-1);
      }
      warmup = 0;
      sprintf (logbuf, "\nWarmup complete: %d calls\n", warmupcalls);
      rsWriteLog (rslogid, logbuf, 0);
    }

    /* create a new src call regardless of whether we're leaving warmup */
    if (createNewSrcDestCall (source, dest, time, network, evq)) {
      RS_PRINT_ERROR ("createNewCall: Error in createNewSrcCall\n");
      return (-1);
    }
    if (warmupcalls % CALLS_BETWEEN_LOG == 0) {
      updateSimLog (warmupcalls, warmuptime, network);
    }
  }
  else {    /* not in warmup */
    if (!routesimcomplete) {
      /* can we stop the sim? */
      if (stopSim (network)) {
	routesimcomplete = 1;
	/* dump stats here before all calls run to completion and network drains */
	rsWriteLog (rslogid, "\nSimulation complete.  Dumping stats ...\n", 0);
	rsCloseLog (rslogid);
	printf ("\nIntermed. statistics:\n");
	printStats ();
	if (mailreport) {
	  if (sendRSNotification (RS_COMPLETE_NOTICE)) {
	    RS_PRINT_ERROR ("Could not send notification\n");
	  }
	}	
      }
      else {	/* not done yet, keep making calls */
	if (createNewSrcDestCall (source, dest, time, network, evq)) {
	  RS_PRINT_ERROR ("eventHandler: Error in createNewSrcCall\n");
	  return (-1);
	}
	if (actualcalls % CALLS_BETWEEN_LOG == 0) {
	  updateSimLog (actualcalls, actualtime, network);
	}
      }
    } /* if !routesimcomplete */
  }
  return (0);
}
/********************************************************************************/
static int
terminationEvHandler (termination_t *term, eventQ& evq)
{
  int	    signalcode;
  double    actualtermtime;
  

  actualtermtime = term->call->gentime + term->call->length;
  /* teardown the the call if 1) using the SINGLE netmodel, or 2) using the DUAL
   * but the call is really over */
  if ( (netmodel == SINGLE) ||
       ((netmodel == DUAL) && (term->termtime >= actualtermtime)) ) {
    if (teardownCall (term->call, term->termtime, &signalcode)) {
      RS_PRINT_ERROR ("terminationEvHandler: Error signalling call\n");
      return(-1);
    }
    term->network->stats->terminatedcalls++;
#if RS_TRACE_LEVEL >= TRACE_MORE_EVENTS
    if (signalcode != SIG_SUCCESS) {
      printf ("terminationEvHandler: Call teardown failed: ");
      printSignalCode (signalcode, stdout);
    }
#endif
    destroyCall (term->call);
    free (term);
  }
  else {    /* using the DUAL model and call is not really over */
    if (callRerouteEvHandler (term, evq)) {
      RS_PRINT_ERROR ("terminationEvHandler: callRerouteHandler returned error\n");
      return (-1);
    }
  }
  return (0);
}
/********************************************************************************/
static int
callRerouteEvHandler (termination_t *term, eventQ& evq)
/* Try to find a QoS route for the call.  If we can't reroute the call, stay on original
 * static route.
 */
{
  int	    reroutecode, teardowncode;
  call_t    *call, *origcall;

  if (netmodel != DUAL) {   /* this should never happen */
    RS_PRINT_ALWAYS ("callRerouteEvHandler:  Invalid network model!\n");
    return (-1);
  }

  call = term->call;

  if (call->network != staticnetwork) {	 /* also should not happen */
    RS_PRINT_ERROR1 ("callRerouteEvHandler:  Reroute handler invoked for call %d on dynamic net\n", call->callid);
    return (-1);
  }

  /* save a copy the original call, including route and links */
  if (copyCallRequest (call, &origcall)) {
    RS_PRINT_ERROR1 ("callRerouteEvHandler:  copyCallRequest (call %d) returned error\n", call->callid);
    return (-1);
  }
  origcall->route = (int *) malloc (sizeof(int) * call->routelen);
  bcopy (call->route, origcall->route, (sizeof(int) * call->routelen));
  if (getLinks (origcall->route, origcall->routelen, &origcall->links, origcall->network)) {
    RS_PRINT_ERROR1 ("callRerouteEvHandler:  getLinks (call %d) returned error\n", call->callid);
    return (-1);
  }  

  /* try to reroute call */
  call->network = dynamicnetwork;   /* designate call for dynamic network */
  if (!call->reroute) {
    rerouteattempts++;	/* only count first attempt for each call */
  }
  call->reroute++;	/* set reroute flag */
  call->network->stats->rerouteattempts++;  
  if (newCallEvHandler (call, evq, &reroutecode)) {
    RS_PRINT_ERROR1 ("callRerouteEvHandler:  newCallEvHandler (call %d) returned error\n", call->callid);
    return (-1);
  }
  if (reroutecode == REQUEST_SUCCESS) {    /* successful reroute */
    RS_PRINT_EVENT2 ("Time %.3f: Call %d rerouted successfully\n", globaltime, call->callid);
    reroutedcalls++;
    /* new termination event was created in newCallEvHandler for dynamic network */
    /* teardown original call on static network */
    if (teardownCall (origcall, term->termtime, &teardowncode)) {
      RS_PRINT_ERROR1 ("callRerouteEvHandler:  teardownCall (call %d) returned error for original call\n", call->callid);
      return (-1);
    }
    /* don't need call copy now */
    destroyCall (origcall);
  }
  else {        /* reroute failed */
    RS_PRINT_EVENT2 ("Time %.3f: Call %d reroute failed\n", globaltime, origcall->callid);
    /* restore original call since reroute attempt destroys call in newCallEvHandler */
    origcall->reroute++;
    /* create new termination event for original call */
    if (createCallTermEv (origcall, evq)) {
      RS_PRINT_ERROR1 ("callRerouteEvHandler: createCallTermEv (call %d) returned error\n", origcall->callid);
      return (-1);
    }
  }

  free (term);
  term = (termination_t *) 0;
  return (0);

}
/********************************************************************************/
static int
recompEvHandler (event_t *ev, eventQ& eventq)
{
  recomp_t	*recompev;
  double	time, elapsed, offset;
  int		source, calcresult;
  int		computenow;
  rsstat_t	*stats;
  topology_t	*network;

  time = ev->time;
  recompev = (recomp_t *) ev->info;
  source = recompev->source;
  network = recompev->network;
  stats = network->stats;

  if (!routesimcomplete) {
    elapsed = time - getLastRouteCompTime(source, network);
    offset = recompinterval - elapsed;
    /* We recompute if:	1. it's been more than a period since last recomputation OR	*/
    /*			2. it's been less but the difference is within the skew		*/
    if ( (offset <= 0) || ((offset > 0) && (fabs(offset) <= (recompinterval * recompskew))) ) 
      computenow = 1;
    else
      computenow = 0;
    
    if (computenow) {
      /* okay to recompute the tree */
      if (reComputeRoute (source, time, 0.0, network, &calcresult)) {
	RS_PRINT_ERROR ("recompEvHandler: reComputeRoute returned error\n");
	return (-1);
      }
      if (updateSimpleRatioStat (elapsed, &(stats->rcompinterval))) {
	RS_PRINT_ERROR ("recompEvHandler: Error updating rcompinterval\n");
      }      
      if (calcresult != ROUTE_CALC_FAIL) {
	ev->time += recompinterval + ((drand48()-0.5) * recompskew * recompinterval);
	eventq.push(ev);	/* reinsert into priority heap */
      }
      else {  /* route calculation failed for some reason -- shouldn't happen */
	RS_PRINT_ERROR ("recompEvHandler: Route calc failed\n");
      }
    }
    else {    /* too early -- reschedule this recomputation */
      /* reschedule to make sure that next time we will be able to recompute */
      ev->time = time + offset + ((drand48()-0.5) * recompskew * recompinterval);
      eventq.push(ev);	/* reinsert into priority heap */
      RS_PRINT_EVENT3 ("Recomputation for source %d postponed to %.3f (last: %.3f)\n",
		       source, ev->time, getLastRouteCompTime(source, network));
    }
  }
  else {    /* routesim is complete -- start freeing these events */
    free (recompev);
    free (ev);
  }
  
  return (0);
}
/********************************************************************************/
static int
stopWarmup (topology_t *network)
/* Return 1 if warmup is over, 0 otherwise
 */
{
  rsstat_t  *stats;
  int	    donecheck;
  int	    retval;

  stats = network->stats;

  switch (netmodel) {
  case SINGLE:
    (network->blkpolicy == NON_BLOCKING) ?
      (donecheck = checkProportionStat (&overutilstat, warmuptolerance)) :
	(donecheck = checkSimpleRatioStat (&stats->callblk, warmuptolerance));
    /* can we leave warmup? */
    if ( (donecheck) && (warmupcalls >= minwarmup) ) {
      retval = 1;
    }
    else {
      retval = 0;
    }
    break;
  case DUAL:
    /* use overutilization on the static network */
    stats = staticnetwork->stats;
    donecheck = checkProportionStat (&overutilstat, warmuptolerance);
    if ( (donecheck) && (warmupcalls >= minwarmup) ) {
      retval = 1;
    }
    else {
      retval = 0;
    }
    break;
  default:
    RS_PRINT_ALWAYS ("stopWarmup: Unknown netmodel\n");
  }

  if (retval == 1) {
    RS_PRINT_GROSS_EVENT ("Warmup period over\n");

  }
  return (retval);
}
/********************************************************************************/
static int
stopSim (topology_t *network)
/* Return 1 if sim is over according to stopping criteria, 0 otherwise
 */
{
  rsstat_t  *stats;
  int	    donecheck;
  int	    retval;

  switch (netmodel) {
  case SINGLE:
    stats = network->stats;
    (network->blkpolicy == NON_BLOCKING) ?
      (donecheck = checkProportionStat (&overutilstat, simtolerance)) :
	(donecheck = checkSimpleRatioStat (&stats->callblk, simtolerance));  
    if ((donecheck) &&
	(actualcalls >= (float)warmupcalls * warmupproportion) &&
	(actualcalls >= minrequests) &&
	(actualtime >= minsimtime) ) {
      retval = 1;
    }
    else {
      retval = 0;
    }
    break;
  case DUAL:
    /* use overutilization on the static network */
    stats = staticnetwork->stats;
    donecheck = checkProportionStat (&overutilstat, simtolerance);
    if ((donecheck) &&
	(actualcalls >= (float)warmupcalls * warmupproportion) &&
	(actualcalls >= minrequests) &&
	(actualtime >= minsimtime) ) {
      retval = 1;
    }
    else {
      retval = 0;
    }
    break;
  default:
    RS_PRINT_ALWAYS ("stopWarmup: Unknown netmodel\n");
  }
  if (retval == 1) {
    RS_PRINT_GROSS_EVENT ("Simulation complete\n");
  }

  return (retval);
}
/********************************************************************************/
static void
updateSimLog (int calls, double time, topology_t *network)
{
  char	    logbuf[75];
  rsstat_t  *stats;

  switch (netmodel) {
  case SINGLE:
    stats = network->stats;
    switch (network->blkpolicy) {
    case BLOCKING:
      sprintf (logbuf, "%d (%.4f [%.4f, %.4f]) (blocking) Time: %.2f",
	       calls, stats->callblk.currentrate, stats->callblk.conf.lb,
	       stats->callblk.conf.ub, time);
      break;
    case NON_BLOCKING:
      sprintf (logbuf, "%d (%.4f [%.4f, %.4f]) (overutil) Time: %.2f",
	       calls, overutilstat.proportion, overutilstat.conf.lb,
	       overutilstat.conf.ub, time);
      break;
    default:
      RS_PRINT_ALWAYS ("updateSimLog:  Unknown blocking policy\n");
    }
    break;
  case DUAL:
    stats = staticnetwork->stats;
    sprintf (logbuf, "%d (%.4f [%.4f, %.4f]) (overutil) Time: %.2f",
	     calls, overutilstat.proportion, overutilstat.conf.lb,
	     overutilstat.conf.ub, time);
    break;
  default:
    RS_PRINT_ALWAYS ("updateSimLog:  Unknown netmodel\n");
  }

  rsWriteLog (rslogid, logbuf, 1);
}
/********************************************************************************/
static int
startActualSim (double currenttime)
{
  
  int	i, numnodes;

  numnodes = getNumNodes (rsnetwork);  

  /* reset most stats between warmup and actual simulation */
  switch (netmodel) {
  case SINGLE:
    if ( (resetRoutesimStats(rsnetwork)) ||
	 resetLinkStatistics (currenttime, conflevel, rsnetwork) ) {       
      RS_PRINT_ERROR ("startActualSim: Error initializing stats\n");
      return (-1);
    }
    for (i = 0; i < numnodes; i++)
      resetRouteComps (i, rsnetwork);
    break;
  case DUAL:
    if ( (resetRoutesimStats(staticnetwork)) ||
	 (resetLinkStatistics (currenttime, conflevel, staticnetwork)) ||
	 (resetRoutesimStats (dynamicnetwork)) ||
	 (resetLinkStatistics (currenttime, conflevel, dynamicnetwork)) ) {
      RS_PRINT_ERROR ("Error initializing stats\n");
      return (-1);
    }
    for (i = 0; i < numnodes; i++) {
      resetRouteComps (i, staticnetwork);
      resetRouteComps (i, dynamicnetwork);
    }
    break;
  default:
    RS_PRINT_ERROR ("Unknown netmodel\n");
    return (-1);
  }

  return (0);
}


/********************************************************************************/
static int
createCallTermEv (call_t *call, eventQ& evq)
{
  eventptr_t	newev;
  termination_t	*callterm;
  double	nexttrig, actualtermtime;;
  
  callterm = getCallTermination (call);
  if (callterm == (termination_t *) 0) {
    RS_PRINT_SOFT_ERROR ("createCallTermEv: Error allocating termination\n");
    return (-1);
  }

  if (netmodel == DUAL) {
    /* if the call is routed on the dynamic network, the termination should be for the
     * actual call termination.  otherwise, we have to use the flow trigger to
     * decide when to schedule termination.
     */
    if ( (call->network == staticnetwork) &&
	 (call->length > flowtrig) ) {
      nexttrig = call->gentime + flowtrig * (call->reroute + 1);
      actualtermtime = callterm->termtime;  /* set in getCallTermination() */
      (nexttrig < actualtermtime) ?
	(callterm->termtime = nexttrig) : (callterm->termtime = actualtermtime);
    }
  }

  newev = getNewEvent (callterm->termtime, ENDCALL, callterm);
  if (newev == (eventptr_t) 0) {
    RS_PRINT_SOFT_ERROR ("createCallTermEv: Error allocating new event\n");
    free (callterm);
    return(-1);
  }

  /* create heap entry */
  evq.push(newev);

  return (0);
}

/********************************************************************************/
static int
createNewSrcDestCall (int src, int dest, double time, topology_t *network, eventQ& eventq)
{
  call_t	*call;
  eventptr_t	ev;
  topology_t	*callnet;

  if (netmodel == DUAL) {
    callnet = staticnetwork;
  }
  else {
    callnet = network;
  }
  
  if (getNewCall (src, dest, time, callnet, &call)) {
    RS_PRINT_SOFT_ERROR ("Error in getNewCall\n");
    return (-1);
  }
  assert (call != (call_t *) 0);
  ev = getNewEvent (call->gentime, STARTCALL, call);
  if (ev == (eventptr_t) 0) {
    RS_PRINT_SOFT_ERROR ("Error in getNewEvent\n");
    free (call->failedlinks);
    free (call);
    return (-1);
  }
  /* put event in the heap */
  eventq.push (ev);

  call->callid = nextcallid++;
  warmup == 1 ? warmupcalls++ : actualcalls++;

  return (0);
}  
/********************************************************************************/  
static int
createInitEvents (eventQ& eventq)
/* Create the initial call and route recomputation events for each node
 */
{
  double	time;
  int		i, j, numnodes;
  int		retval;
  call_t	*call;
  eventptr_t	ev;  
  recomp_t	*recomp;
  topology_t	*network;
  

  RS_PRINT_EVENT ("Creating bootstrap events ...");

  switch (netmodel) {
  case SINGLE:
    network = rsnetwork;
    break;
  case DUAL:
    network = staticnetwork;	/* all calls are initially on the static net */
    break;
  default:
    RS_PRINT_ALWAYS ("createInitEvents: Unknown netmodel\n");
    return (-1);
  }
	    
  time = 0.0;
  numnodes = getNumNodes(network);
  for (i = 0; i < numnodes; i++) {  /* one call from ea node */
    switch (arrspec) {
    case ARR_UNIFORM:
      if (getNewCall (i, -1, time, network, &call)) {
	RS_PRINT_SOFT_ERROR ("Error in getNewCall\n");
	return (-1);
      }
      if (call == (call_t *) 0)
	continue;   /* no traffic from this node */      
      if (createCallEvent(call, eventq)) {
	RS_PRINT_ERROR ("createInitEvents:  createCallEvent failed\n");
	return (-1);
      }      
      break;
    case ARR_FILE:
      for (j = 0; j < numnodes; j++) {
	if (i == j) continue;
	retval = getNewCall (i, j, time, network, &call);
	if (retval == RS_NO_TRAFFIC) {
	  continue;
	}	
	else if (retval < 0) {
	  RS_PRINT_SOFT_ERROR ("Error in getNewCall\n");
	  return (-1);
	}
	if (createCallEvent(call, eventq)) {
	  RS_PRINT_ERROR ("createInitEvents:  createCallEvent failed\n");
	  return (-1);
	}
      }
    }
	
    /* create periodic recomputation event for this source if necessary */
    if (periodicrecomps == TRUE) {
      recomp = (recomp_t *) malloc (sizeof(recomp_t));
      recomp->source = i;
      recomp->period = recompinterval;
      recomp->recomptime = recompinterval + ((drand48()-0.5) * recompskew * recompinterval);
      (netmodel == SINGLE) ? (recomp->network = rsnetwork) : (recomp->network = dynamicnetwork);
      ev = getNewEvent (recomp->recomptime, RECOMP, recomp);
      if (ev == (eventptr_t) 0) {
	RS_PRINT_SOFT_ERROR ("Error in getNewEvent\n");
	return (-1);
      }
      eventq.push(ev);
    }
  } /* for all nodes i */
  warmupcalls++;

  RS_PRINT_EVENT ("Done\n");

  return (0);
}
/********************************************************************************/
static int
createCallEvent (call_t *call, eventQ& eventq)
{
    eventptr_t	ev;

    call->callid = nextcallid++;
#if RS_TRACE_LEVEL >= TRACE_MORE_EVENTS
    printf ("Initial call %d: gentime %.2f bw %.2f duration %.2f\n", call->callid,
	    call->gentime, call->bandw, call->length);
#endif
    ev = getNewEvent (call->gentime, STARTCALL, call);
    if (ev == (eventptr_t) 0) {
      RS_PRINT_SOFT_ERROR ("Error in getNewEvent\n");
      return (-1);
    }
    /* put event in the heap */
    eventq.push(ev);

    return (0);
}
/********************************************************************************/
static int
getNewCall(int source, int dest, double time, topology_t *network, call_t **newcall)
/* Generates a new connection request and returns it in newcall
 */

{
  call_t    *call;
  double    interarrival;
  int	    i;
  int	    tries = 0;
  int	    numnodes;

  *newcall = (call_t *) 0;
  
  call = (call_t *) malloc (sizeof(call_t));
  if (call == (call_t *) 0)
    return (-1);

  numnodes = getNumNodes (network);

  /* find source and destination */
  if (source >= 0) {
    call->source = source;
  }
  else {
    call->source = (int) (numnodes * drand48());
  }

  if (dest < 0 ) {
    /* destination unspecified, select uniformly */
    call->dest = ( ((int)ceil((numnodes-1) * drand48())) + call->source) % numnodes;
  }
  else {
    call->dest = dest;
  }
    

  /* get the next call interarrival time */
  interarrival = getInterarrival (call->source, call->dest);

  if (interarrival > 0) {   /* interarrival was okay */
    call->gentime = time + interarrival;
  }
  else if (interarrival == 0.0) { /* no traffic from src to dest */
#if RS_TRACE_LEVEL >= TRACE_DETAILS
    printf ("getNewCall: No traffic rate from %d -> %d\n", call->source, call->dest);
#endif
    if (dest < 0) {
      /* if destination was originally unspecified, try to find a destination
	 with a nonzero traffic rate */
      while (interarrival == 0.0 && tries < MAX_INTERARRIVAL_TRIES) {
	call->dest = ( ((int)ceil((numnodes-1) * drand48())) + call->source) % numnodes;
	interarrival = getInterarrival (call->source, call->dest);
	tries++;
      }
      /* got a nonzero interarrival or tries expired */
      if (interarrival == 0.0)
	return (RS_NO_TRAFFIC);
      call->gentime = time + interarrival;
    }
    else {	/* dest was specified but no traffic */
      free (call);
      return (RS_NO_TRAFFIC);
    }
  } /* if interarrival == 0 */
  else {    /* interarrival was < 0 */
    RS_PRINT_SOFT_ERROR ("getNewCall: getInterarrival failed\n");
    free (call);
    return (-1);
  }
  updateSimpleRatioStat (interarrival, &(network->stats->arrivalratestat));
  
  call->bandw = getBandwidth();
  /* spread better be set such that negative or zero bandwidths are disallowed -- we don't
   * adjust the bandwidth because it skews the expected value of the bandwidth
   */
  if (call->bandw <= 0.0) {
#if RS_TRACE_LEVEL >= TRACE_SOFT_ERRORS
    fprintf (stderr, "getNewCall: Invalid bandwidth %.2f\n", call->bandw);
#endif
    free(call);
    return (-1);
  }
  call->length = getHoldTime();

  call->network = network;
  call->minhops = getMinHops (call->source, call->dest, network);
  call->routelen = 0;
  call->route = (int *) 0;
  call->links = (link_t **) 0;
  call->callid = -1;
  call->routecomps = 0;
  call->extractions = 0;
  call->sigattempts = 0;
  call->reroute = 0;
  call->failedlinks = (link_t **) malloc (sizeof(link_t *) * maxsigattempts);
  for (i = 0; i < maxsigattempts; i++)
    call->failedlinks[i] = (link_t *) 0;

  *newcall = call;
  return (0);
}
/********************************************************************************/
static int
copyCallRequest (call_t *call, call_t **call2)
/* Copy call request parameters from call to call2
 */
{
  call_t    *newcall;
  int	    i;

  *call2 = (call_t *) 0;
  
  newcall = (call_t *) malloc (sizeof(call_t));
  if (newcall == (call_t *) 0)
    return (-1);

  newcall->callid = call->callid;
  newcall->network = call->network;  
  newcall->source = call->source;
  newcall->dest = call->dest;
  newcall->gentime = call->gentime;
  newcall->length = call->length;
  newcall->bandw = call->bandw;
  
  newcall->minhops = call->minhops;
  newcall->routelen = call->routelen;
  newcall->route = (int *) 0;	/* route and links aren't filled in */
  newcall->links = (link_t **) 0;
  newcall->routecomps = call->routecomps;
  newcall->extractions = call->extractions;
  newcall->sigattempts = call->sigattempts;
  newcall->reroute = call->reroute;
  newcall->failedlinks = (link_t **) malloc (sizeof(link_t *) * maxsigattempts);
  for (i = 0; i < maxsigattempts; i++)
    newcall->failedlinks[i] = call->failedlinks[i];

  *call2 =  newcall;
  
  return (0);
}
/********************************************************************************/
static int
destroyCall (call_t *call)
{

  if (call->routelen != 0) {
    free (call->route);
    free (call->links);
  }
  call->route = (int *) 0;
  call->links = (link_t **) 0;
  free (call->failedlinks);
  free (call);
  call = (call_t *) 0;

  return (0);

}
/********************************************************************************/
static termination_t *
getCallTermination (call_t *call)
{
  termination_t	    *term;

  term = (termination_t *) malloc (sizeof(termination_t));
  if (term == (termination_t *) 0)
    return ((termination_t *) 0);

  term->termtime = call->gentime + call->length;
  term->call = call;
  term->network = call->network;

  return (term);
}

/********************************************************************************/
static event_t *
getNewEvent (double evtime, eventtype_t evtype, void *ptr)
{
  event_t	*newevent;

  /* create event */  
  newevent = (event_t *) malloc (sizeof(event_t));
  if (newevent == (event_t *) 0)
    return ((event_t *) 0);
  newevent->time = evtime;
  newevent->type = evtype;
  newevent->info = ptr;

  return(newevent);
}
/********************************************************************************/
static int
signalCall (call_t *call, int *result)
/* The result parameter contains result of the signalling attempt.
 *
 * Signalling must check several things on each link (i.e. perform
 * admission control):
 *	1. link is up, i.e. active
 *	2. link can handle an additional call, i.e. maxcalls has not been reached
 *	3. link has sufficient bandwidth for new call
 *
 * If signalling fails on any link, any bandwidth allocated on other links for
 * this call is released.
 */
{
  int		i, j;
  int		success = 1;
  double	bwchange, time;
  double	utilseen;
  topology_t	*network;
  rsstat_t	*stats;

  /* this routine may be called during reroute in addition to
   * call arrival, so we can't set time to call->gentime */
  time = globaltime;

  if (!call->routelen) {
    RS_PRINT_SOFT_ERROR ("signalCall: Call has no route\n");
    return (-1);
  }
  network = call->network;
  stats = network->stats;
  
  stats->sigattempts++;    /* update sigattempts regardless of signalling outcome */
  call->sigattempts++;
  for (i = 0; i < call->routelen - 1; i++) {	/* for each link in route */

    if (call->links[i]->down) {	/* link up? */
      success = 0;
      *result = SIG_LINK_DOWN;
      break;
    }

    if (call->links[i]->numcalls >= call->links[i]->maxcalls) { /* max calls reached? */
      success = 0;
      *result = SIG_MAX_REACHED;
      break;
    }

    
    /* avg util seen on this link by arriving call (as a fraction of link capacity */
    /* record this whether or not adm test succeeds */
    utilseen = (call->links[i]->capacity - call->links[i]->availbw.current) /
      call->links[i]->capacity;
    updateSimpleRatioStat (utilseen, &(call->links[i]->util));
    
    updateLinkStats (time, call->links[i]);
    
    /* admission test */
    if ((network->blkpolicy == NON_BLOCKING) ||
	(call->links[i]->availbw.current - call->bandw >= 0) ) {

      /* successful admission test -- reserve bandwidth on this link */
      call->links[i]->availbw.current -= call->bandw;
      
      /* update some post-admission link statistics */
      call->links[i]->numcalls++;
      call->links[i]->callsceil++;      

      /* log bw seen after call arrives on link */
      if (!warmup && !routesimcomplete && call->links[i]->monitor) {
	if (logLinkEvent(LINK_ARR, time, call->links[i], call, linkmonlogid)) {
	  RS_PRINT_ERROR ("Error writing link log event\n");
	}
      }

      bwchange = calcBWChange(call->links[i], (call->links[i]->availbw.current - call->bandw));
			     
      if (bwchange > trigger) {   /* generate an update */
	if (time >= call->links[i]->lastupdtime + minupdinterval) {
	  if (updateSingleLink(call->links[i]->from, call->links[i]->to, network)) {
	    RS_PRINT_ERROR ("signalCall: updateSingleLink returned error\n");
	    return (-1);
	  }
	  if (updateSimpleRatioStat ((time - call->links[i]->lastupdtime), &stats->updinterval)) {
	    RS_PRINT_ERROR ("signalCall: Error updating updinterval\n");
	  }

	  /* check for monitoring flag */
	  if (call->links[i]->monitor && !warmup && !routesimcomplete) {
	    if (logLinkEvent(LINK_TRIG, time, call->links[i], call, linkmonlogid)) {
	      RS_PRINT_ERROR ("Error writing link log event\n");
	    }
	  }
	  
	  call->links[i]->lastupdtime = time;
	  call->links[i]->pendingupdate = 0;
#if RS_TRACE_LEVEL >= TRACE_EVENTS
	  printf ("Time %.2f: ", time);
	  printf ("Triggered update (s %.1f->%.1f %.1f%%) lut:%.2f\n",
		  call->links[i]->availbw.last,
		  (call->links[i]->availbw.current - call->bandw), bwchange * 100,
		  call->links[i]->lastupdtime - minupdinterval);	    
#endif
	  stats->tlinkupdates++;
	}
	else {	/* it's too soon for this update */
	  call->links[i]->pendingupdate = 1;
#if RS_TRACE_LEVEL >= TRACE_EVENTS
	  printf ("Time %.2f: ", time);
	  printf ("Deferred triggered update (s %.1f%%) to (%.2f)\n", bwchange * 100,
		  call->links[i]->lastupdtime + minupdinterval);
#endif
	  stats->deferredupdates++;
	}
      } /* if bwchange > trigger */
    }
    else {  /* insufficient bandwidth */
      success = 0;      
      *result = SIG_NO_BANDWIDTH;
      /* record which link caused the failure */
      call->failedlinks[(call->sigattempts - 1)] = call->links[i];
      if (call->links[i]->monitor && !warmup && !routesimcomplete) {
	if (logLinkEvent(LINK_SIGFAIL, time, call->links[i], call, linkmonlogid)) {
	  RS_PRINT_ERROR ("Error writing link log event\n");
	}
      }	
      break;	/* skip the rest of the links */
    }

  } /* for each link i */

  if (!success) {   /* restore any used bandwidth to links */
    for (j = 0; j < i; j++) {
      /* recalculate bwchange for the link */
      bwchange = calcBWChange (call->links[j], call->links[j]->availbw.current);
      call->links[j]->availbw.current += call->bandw;
      call->links[j]->numcalls--;
      if (bwchange > trigger)
	stats->extraupdates++;
    }
  }
  else {
    *result = SIG_SUCCESS;

  }
  return (0);
}

/********************************************************************************/
static int
teardownCall (call_t *call, double time, int *result)
{
  int		i;
  int		success = 1;
  double	bwchange;
  topology_t	*network;
  rsstat_t	*stats;

  network = call->network;
  stats = network->stats;

  for (i = 0; i < call->routelen - 1; i++) {	/* for each link in route */
    if (processLinkUpdates (time, call->links[i], network, "(teardown)")) {
      RS_PRINT_ERROR ("processLinkUpdates failed\n");
      return (-1);
    }
    
    if (call->links[i]->down) {	/* link up? */
      success = 0;
      *result = SIG_LINK_DOWN;
      /* must terminate on all links or none */
      break;
    }

    /* record link stats before changing state of the link */
    updateLinkStats (time, call->links[i]);

    bwchange = calcBWChange (call->links[i], (call->bandw + call->links[i]->availbw.current));
    call->links[i]->availbw.current += call->bandw; /* restore bandwidth */

    /* check for monitoring flag */
    if (call->links[i]->monitor && !warmup && !routesimcomplete) {
      if (logLinkEvent(LINK_TERM, time, call->links[i], call, linkmonlogid)) {
	RS_PRINT_ERROR ("Error writing link log event\n");
      }
    }      

    if (bwchange > trigger) { /* generate an update */
      if (time >= call->links[i]->lastupdtime + minupdinterval) {
	if (updateSingleLink(call->links[i]->from, call->links[i]->to, network)) {
	  RS_PRINT_ERROR ("teardownCall: updateLinks() failed\n");
	  return (-1);
	}
	if (updateSimpleRatioStat ((time - call->links[i]->lastupdtime), &stats->updinterval)) {
	  RS_PRINT_ERROR ("teardownCall: Error updating updinterval\n");
	}
	
	/* check for monitoring flag */
	if (call->links[i]->monitor && !warmup && !routesimcomplete) {
	  if (logLinkEvent(LINK_TRIG, time, call->links[i], call, linkmonlogid)) {
	    RS_PRINT_ERROR ("Error writing link log event\n");
	  }
	}
		
	call->links[i]->lastupdtime = time;
	call->links[i]->pendingupdate = 0;

#if RS_TRACE_LEVEL >= TRACE_EVENTS
	printf ("Time %.2f: ", time);
	printf ("Triggered update (t %.1f->%.1f %.1f%%) lut:%.2f\n",
		call->links[i]->availbw.last,
		(call->bandw +  call->links[i]->availbw.current), bwchange * 100,
		call->links[i]->lastupdtime - minupdinterval);
#endif
	stats->tlinkupdates++;
      }
      else {	/* it's too soon for this update */
	call->links[i]->pendingupdate = 1;
#if RS_TRACE_LEVEL >= TRACE_EVENTS	
	printf ("Time %.2f: ", time);
	printf ("Deferred triggered update (t %.1f->%.1f %.1f%%) to (%.2f)\n",
		call->links[i]->availbw.last,
		(call->bandw +  call->links[i]->availbw.current),
		bwchange * 100, call->links[i]->lastupdtime + minupdinterval);
#endif
	stats->deferredupdates++;
      }
    }
    call->links[i]->numcalls--; 
  }
  if (success)
    *result = SIG_SUCCESS;
  return(0);
}
/********************************************************************************/
static int
updateLinkStats (double time, link_t *link)
{
  double    lastinterval;
  double    totaltime;
  int	    numlinks, from, to, numcalls;
  link_t    *mainlink, *dynamiclink, *staticlink;
  double    pctunreserved, level;

  numlinks = getNumLinks (rsnetwork);
  (warmup == 1) ? (totaltime = warmuptime) : (totaltime = actualtime - warmuptime);
  from = link->from;
  to = link->to;

  switch (netmodel) {
  case SINGLE:
    if (link->monitor && !warmup) {
      if (logLinkEvent(LINK_UTIL, time, link, (call_t *) 0, linkmonlogid)) {
	RS_PRINT_ERROR ("Error writing link log event\n");
      }
    }
    numcalls = link->numcalls;
    lastinterval = (time - link->lastutilupd) / (double) numlinks;  
    /* figure out what mode the link has been in -- no violations
     * in SINGLE model
     */
    if (link->availbw.current < 0) {   /* over utilized */
      updateProportionStat (lastinterval, totaltime, &overutilstat);
      updateProportionStat ((lastinterval * numcalls), totaltime, &overutilflowstat);
      level = fabs(link->availbw.current / link->capacity);
      woverutil += (lastinterval * level);
    }
    else {
      if (link->availbw.current >= 0) { /* utilized */
	updateProportionStat (lastinterval, totaltime, &underutilstat);
	updateProportionStat ((lastinterval * numcalls), totaltime, &underutilflowstat);
	level = link->availbw.current / link->capacity;
	wunderutil += (lastinterval * level);
      }
    }
    link->lastutilupd = time;
    break;
  case DUAL:
    mainlink = getLinkRef (from, to, rsnetwork);
    dynamiclink = getLinkRef (from, to, dynamicnetwork);
    staticlink = getLinkRef (from, to, staticnetwork);

    /* set total link availbw -- could be negative. */
    /* this is really a hack -- the main link doesn't get updated
     * during setup or teardown, only the link on which the operation
     * is performed.  therefore this is necessary if we want to keep track
     * of util on the "merged" network. */
    mainlink->availbw.current = dynamiclink->availbw.current + staticlink->availbw.current;
    if (mainlink->monitor && !warmup) {
      if ( logLinkEvent(LINK_UTIL, time, link, (call_t *) 0, linkmonlogid) ||
	   logLinkEvent(LINK_UTIL_DYN, time, dynamiclink, (call_t *) 0, linkmonlogid) ||
	   logLinkEvent(LINK_UTIL_STAT, time, staticlink, (call_t *) 0, linkmonlogid) ) {
	RS_PRINT_ERROR ("Error writing link log event\n");
      }
    }  

    /* choice of link to set time interval is arbitrary -- all three
     * should be correct */
    lastinterval = (time - staticlink->lastutilupd) / (double) numlinks;
    /* only flows on static partition are affected */
    numcalls = staticlink->numcalls;

    /* percent unreserved capacity used is static utilization divided by total
     * unreserved capacity */
    pctunreserved = (staticlink->capacity - staticlink->availbw.current) /
      (staticlink->capacity + dynamiclink->availbw.current);

    if (pctunreserved > 1.0) {	/* violation */
      updateProportionStat (lastinterval, totaltime, &violutilstat);
      updateProportionStat ((lastinterval * numcalls), totaltime, &violutilflowstat);      
      level = pctunreserved - 1.0;
      wviolutil += (level * lastinterval);
    }
    else {
      if (staticlink->availbw.current < 0) {   /* overutilized */
	updateProportionStat (lastinterval, totaltime, &overutilstat);
	updateProportionStat ((lastinterval * numcalls), totaltime, &overutilflowstat);	
	level = fabs(staticlink->availbw.current / staticlink->capacity);
	woverutil += (lastinterval * level);
      }
      else {
	if (staticlink->availbw.current >= 0) { /* underutilized */
	  updateProportionStat (lastinterval, totaltime, &underutilstat);
	  updateProportionStat ((lastinterval * numcalls), totaltime, &underutilflowstat);
	  level = staticlink->availbw.current / staticlink->capacity;
	  wunderutil += (lastinterval * level);
	}
      }      
    }
    mainlink->lastutilupd = time;
    dynamiclink->lastutilupd = time;
    staticlink->lastutilupd = time;    
    break;
  default:
    RS_PRINT_ALWAYS ("updateLinkStats:  Unknown network model\n");
    return (-1);
  }
    
  return (0);

}
/********************************************************************************/
static double
calcBWChange (link_t *link, double newbw)
{
  double    change;

  /* special cases:
   *	fully utilized to utilized --> bwchange = 1.0
   *	fully utilized to overutilized --> bwchange = 1.0 (?)
   */
  if (link->availbw.last == 0.0) {
    change = 1.0;   /* sets off any update trigger */
  }
  else {
    change = fabs(link->availbw.last - newbw) / fabs(link->availbw.last);
  }

  return (change);
  

}
/********************************************************************************/
int
processLinkUpdates (double time, link_t *link, topology_t *network, char *info)
{
  int	    elapsedperiods;
  int	    baseperiods;
  double    refreshtime;
  double    bwchange;
  int	    i;
  rsstat_t  *stats;

  if (link == (link_t *) 0) {
    RS_PRINT_SOFT_ERROR ("passed NULL link pointer\n");
    return (-1);
  }
  stats = network->stats;
  
  /* handle pending updates */
  if (link->pendingupdate && (time >= link->lastupdtime + minupdinterval)) {
    /* check if update should still be sent */

    bwchange = calcBWChange(link, link->availbw.current);
    if (bwchange > trigger) { /* generate an update */
      if (updateSingleLink(link->from, link->to, network)) {
	RS_PRINT_ERROR ("updateSingleLink failed\n");
	return (-1);
      }
      stats->tlinkupdates++;
      if (updateSimpleRatioStat (minupdinterval, &stats->updinterval)) {
	RS_PRINT_ERROR ("signalCall: Error updating updinterval\n");
      }
      
      /* check for monitoring flag */
      if (link->monitor && !warmup && !routesimcomplete) {
	if (logLinkEvent(LINK_PEND, time, link, 0, linkmonlogid)) {
	  RS_PRINT_ERROR ("Error writing link log event\n");
	}
      }
      link->lastupdtime = link->lastupdtime + minupdinterval;
      link->pendingupdate = 0;
#if RS_TRACE_LEVEL >= TRACE_EVENTS
      printf ("Time %.2f: ", link->lastupdtime);
      printf ("Completed pending link update %s\n", info);
#endif
    }
  }
  /* don't reset the pendingupdate flag */


  /* now handle any refresh updates */

  /* compute randomly skewed refresh interval */
  refreshtime = refreshinterval + ((drand48()-0.5) * refreshskew * refreshinterval);

  if (time >= link->lastupdtime + refreshtime) {    /* time for an update? */
    /* elapsed periods calculation is based on mean refresh period */
    elapsedperiods = (int) floor((time - link->lastupdtime) / refreshinterval);
    if (updateSingleLink(link->from, link->to, network)) {
      RS_PRINT_ERROR ("updateSingleLink failed\n");
      return (-1);
    }
    /* set update time to (elapsed - 1) worth of mean refreshintervals and then one
     * random one
     */
       
    /*    refreshtime = refreshinterval + ((drand48()-0.5) * refreshskew * refreshinterval); */
    (elapsedperiods - 1) < 0 ? (baseperiods = 0) : (baseperiods = elapsedperiods - 1);

    /* check for monitoring flag */
    if (link->monitor && !warmup && !routesimcomplete) {
      for (i = 0; i < baseperiods; i++) {
	if (logLinkRefreshEvent ((link->lastupdtime + (refreshinterval * 1)),
				 refreshinterval, link, linkmonlogid)) {
	  RS_PRINT_SOFT_ERROR ("Error writing link log\n");
	}
      }
    }

    link->lastupdtime = link->lastupdtime +
      ((float)(baseperiods) * refreshinterval) + refreshtime; 

    /* check for monitoring flag */
    if (link->monitor && !warmup && !routesimcomplete) {
      if (logLinkRefreshEvent (link->lastupdtime,
			       refreshtime, link, linkmonlogid)) {
	RS_PRINT_SOFT_ERROR ("Error writing link log\n");
      }
    }      

    /* once for each elaspsed refreshinterval and finally once for the random interval */
    stats->rlinkupdates += (baseperiods + 1);

    /* update interval statistic the same way */
    for (i = 0; i < baseperiods; i++) {
      if (updateSimpleRatioStat (refreshinterval, &stats->updinterval)) {
	RS_PRINT_ERROR ("signalCall: Error updating updinterval\n");
      }
    }
    
    if (updateSimpleRatioStat (refreshtime, &stats->updinterval)) {
      RS_PRINT_ERROR ("signalCall: Error updating updinterval\n");
    }
#if RS_TRACE_LEVEL >= TRACE_EVENTS    
    printf ("Time %.3f: ", link->lastupdtime);
    printf ("Periodic refresh link update %s\n", info);
#endif
  }
  return (0);
}

/********************************************************************************/
void
setFailedLinksInactive (call_t *call)
/* Disables links (making them look failed) that failed in signalling for this
 * call.
 */
{
  int	i;

  for (i = 0; i < call->sigattempts; i++)
    call->failedlinks[i]->active = 0;
}

void
resetFailedLinks (call_t *call)
/* Reenables links in the topology database that failed in signalling for this call.
 */
{
  int	i;

  for (i = 0; i < call->sigattempts; i++)
    call->failedlinks[i]->active = 1;
}
/********************************************************************************/
static void
printStats (void)
{
  switch (netmodel) {
  case SINGLE:
    printSingleStats (rsnetwork);
    printf ("%-20s\t%s\n", "Routesim version:", rcsid);
    printf ("---------------------\n");    
    break;
  case DUAL:
    printDualStats (rsnetwork, dynamicnetwork, staticnetwork);
    printf ("%-20s\t%s\n", "Routesim version:", rcsid);
    printf ("---------------------\n");    
    break;
  default:
    RS_PRINT_ALWAYS ("printStats:  Unknown net model\n");
  }
}
/********************************************************************************/
	    
static void
printRoutesimVersion (void)
{
  char		rcsbuf[100];
  char		separator[2];
  char		*stringbuf;
  int		i;
  /* string looks like this:
   *	"$Id: routesim.c,v 5.1 1999/06/02 20:53:18 ashaikh Exp $"
   * or this:
   *	"$Id: routesim.c,v 5.1 1999/06/02 20:53:18 ashaikh Exp $"
   * depending on whether file is checked in or not.
   */
  
  strcpy (separator, " ");
  strncpy (rcsbuf, rcsid, 100);
  printf ("Routesim ver. ");
  stringbuf = strtok(rcsbuf, separator);
  for (i = 0; i< 2; i++)
    stringbuf = strtok(NULL, separator);
  /* version numer */
  printf ("%s ", stringbuf);
  stringbuf = strtok(NULL, separator);
  /* date */
  printf ("%s ", stringbuf);
  for (i = 0; i< 2; i++)
    stringbuf = strtok(NULL, separator);
  printf ("by %s", stringbuf);
  /* now figure out whether file has been checked in */
  for (i = 0; i< 2; i++)
    stringbuf = strtok(NULL, separator);
  if (strcmp (stringbuf, "$\0")) {
      /* file not checked in */
    printf (" (intermediate version)\n");
  }
  else printf ("\n");
  
}


/********************************************************************************/
#if RS_TRACE_LEVEL >= TRACE_EVENTS
static void
printEvent (event_t *event)
{
  call_t	*call;
  termination_t	*term;
  recomp_t	*recomp;

  printf ("Time %.3f: ", event->time);
  switch (event->type) {
  case STARTCALL:
    call = (call_t *) event->info;
    printf ("Call request %d from %d -> %d bw: %.2f dur: %.2f\n", call->callid, call->source,
	    call->dest, call->bandw, call->length);
    break;
  case ENDCALL:
    term = (termination_t *) event->info;
    printf ("Call %d terminated\n", term->call->callid);
    break;
  case RECOMP:
    recomp = (recomp_t *) event->info;
    printf ("Recomputation event for source %d\n", recomp->source);
    break;
  default:
    printf ("Unknown event type\n");
  }
  fflush(stdout);
}

static void
printSignalCode (int code, FILE *fd)
{
  switch (code) {
  case SIG_NO_BANDWIDTH:
    fprintf (fd, "Insufficient bandwidth\n");
    break;
  case SIG_LINK_DOWN:
    fprintf (fd, "Link on route is down\n");
    break;
  case SIG_MAX_REACHED:
    fprintf (fd, "Maximum supported calls reached\n");
    break;
  default:
    fprintf (fd, "Unknown signalling failure\n");
    break;
  }
  fflush (fd);
}
#endif

#if RS_TRACE_LEVEL >= TRACE_DETAILS
void
printCall(call_t *call)
{
  int	i;

  printf("Call information:\n");
  printf("\tCall id: %d\n", call->callid);
  printf("\tGenerated at: %.2f\n",call->gentime);
  printf("\tDuration: %.3f\n",call->length);
  printf("\tBandwidth: %.3f\n",call->bandw);
  printf("\tSource: %d\n",call->source);
  printf("\tDestination: %d\n",call->dest);
  if (call->routelen) {
    printf("\tRoute len: %d\n", call->routelen);
    printf ("\tRoute: ");  
    for (i = 0; i < call->routelen; i++)
      printf ("%d ", call->route[i]);
    printf ("\n");
    printf ("\tLinks: ");
    for (i = 0; i < call->routelen - 1; i++) {
      assert (call->links[i]);
      printf ("%d->%d ", call->links[i]->from, call->links[i]->to);
    }
    printf ("\n");
  }
  else {
    printf ("\tRoute len: 0\n");
    printf ("\tRoute: None\n");
    printf ("\tLinks: None\n");
  }
}
#endif
