/* $Id: rstopology.c,v 2.3 1999/06/02 20:46:27 ashaikh Exp $ */

static const char rcsid[] = "$Id: rstopology.c,v 2.3 1999/06/02 20:46:27 ashaikh Exp $";

#include "rstopology.h"
#include "topology.h"
#include "spt.h"
#include "rsroute.h"
#include "rstrace.h"
#include "rsdefs.h"
#include "general.h"
#include <stdlib.h>
#include <stdio.h>

/* Network topology */
netmodel_t	netmodel = SINGLE;
double		netsplit;
topology_t	*rsnetwork;	    /* network for SINGLE net model, or reference network for DUAL */
topology_t	*staticnetwork;	    /* for short-lived traffic */
topology_t	*dynamicnetwork;    /* for long-lived traffic */

topologytype_t toptype = TOP_KNCUBE;    /* type of topology to use */

/* parameters for random graph */
int		topsize;
int		randgraphmaxdiam = 10;
randgraph_t	randgraphmethod;
double		randgraphalpha;
double		randgraphbeta;
double		randgraphplanesize;
boolean		randgraphasym = FALSE;
/* parameters for k-ary n-cube */
int	edge;
int	dimen;

/* routing-related parameters -- don't really belong here but are necessary for the DUAL
 * network model */
routealg_t routealg = (routealg_t) -1;	/* algorithm for SINGLE netmodel */

routealg_t dualalg = (routealg_t) -1;	/* for DUAL network */
extern prunepolicy_t prunep;
extern distfunction_t distfunc;
extern alternate_t altpolicy;
extern multipolicy_t multipolicy;
extern uniqpolicy_t uniqpolicy;


extern blockingpolicy_t blkpolicy;
extern double refreshinterval;
extern boolean monitorlink;

static int setInitialLinkUpdateTime (topology_t *);
/********************************************************************************/
int
genRSTopology (void)
/*  Creates the simulation network and keeps it in rsnetwork.
 */
{
  int	    result;
  
  switch (toptype) {
  case TOP_KNCUBE:
    result = createKaryNcube(edge, dimen, &rsnetwork);
    break;
  case TOP_FILE:
    result = createFileTopology(TOPFILENAME, &rsnetwork);
    break;
  case TOP_FULL:
    result = createFullTopology(topsize, &rsnetwork);
    break;
  case TOP_RAND:
    result = createRandTopology(randgraphmethod, topsize, randgraphalpha, randgraphbeta,
				randgraphplanesize, randgraphasym, randgraphmaxdiam, &rsnetwork);
    break;
  default:
    RS_PRINT_ERROR ("Unknown topology type\n");
    return (-1);
  }

  if (monitorlink == TRUE) {
    if (initLinkMonitor(rsnetwork)) {
      RS_PRINT_ERROR ("initLinkMonitor failed\n");
      return (-1);
    }
  }

  if (setBlockingPolicy (blkpolicy, rsnetwork)) {
    RS_PRINT_ERROR ("genRSTopology:  setBlockingPolicy returned error\n");
    return (-1);
  }

  setInitialLinkUpdateTime (rsnetwork);


#if RS_TRACE_LEVEL >= TRACE_DETAILS
  printLinkMatrix (rsnetwork);
  printCapacityTable(rsnetwork);
#endif

  if (netmodel == DUAL) {
    if (splitNetCapacity (rsnetwork, netsplit, &staticnetwork, &dynamicnetwork)) {
      RS_PRINT_ERROR ("genRSTopology: splitNetCapacity returned error\n");
      return (-1);
    }
    setBlockingPolicy (NON_BLOCKING, staticnetwork);

    setBlockingPolicy (BLOCKING, dynamicnetwork);
    setInitialLinkUpdateTime (dynamicnetwork);
  }
  
  return (result);
}
/********************************************************************************/
int
initRSRouting (void)
{
  sptpolicy_t	*sptpolicy;
  
  switch (netmodel) {
  case SINGLE:    
    if (initRouting(routealg, rsnetwork)) {
      RS_PRINT_ERROR ("initRSRouting:  initRouting returned error\n");
      return (-1);
    }      
    /* network gets policies from global parms */
    if (setRoutePolicy (rsnetwork, (void *) 0)) {
      RS_PRINT_ERROR ("initRSRouting:  setRoutePolicy returned error\n");
      return (-1);
    }
    break;
  case DUAL:
    if ( (initRouting(SPT, staticnetwork)) ||
	 (initRouting(dualalg, dynamicnetwork)) ) {
      RS_PRINT_ERROR ("initRSRouting:  initRouting returned error\n");
      return (-1);
    }

    /* now install policies -- dynamic network gets policies from global parms */
    if (setRoutePolicy (dynamicnetwork, (void *) 0)) {
      RS_PRINT_ERROR ("initRSRouting:  setRoutePolicy returned error\n");
      return (-1);
    }

    /* set up policies for static network */
    sptpolicy = (sptpolicy_t *) malloc (sizeof(sptpolicy_t));
    if (sptpolicy == (sptpolicy_t *) 0) {
      RS_PRINT_ERROR ("initRSRouting:  unable to allocate sptpolicy\n");
      return (-1);
    }
    sptpolicy->prunep =  NO_PRUNE;
    sptpolicy->distfunc = HOPCOUNT;
    sptpolicy->altpolicy = NO_ALT;
    sptpolicy->multipolicy = UNIQ;
    sptpolicy->uniqpolicy = RANDOM;
    if (setRoutePolicy (staticnetwork, (void *) sptpolicy)) {
      RS_PRINT_ERROR ("initRSRouting:  setRoutePolicy returned error\n");
      return (-1);
    }
    break;
  default:
    RS_PRINT_ALWAYS ("initRSRouting:  Unknown network model!\n");
    return (-1);
  }

  return (0);
}
/********************************************************************************/
int
resetLinkStatistics (double time, double conflevel, topology_t *network)
{
  int	i, j, numnodes;

  numnodes = getNumNodes (network);

  for (i = 0; i < numnodes; i++) {
    for (j = 0; j < numnodes; j++) {
      if (areConnected (i, j, network)) {
	if ( (setLinkUtilUpdateTime (i, j, time, network)) ||
	     (initLinkUtilStat (getLinkRef (i,j,network), conflevel)) ) {
	  RS_PRINT_ERROR ("resetLinkStatistics:  Error resetting\n");
	  return (-1);
	}
      }
      else continue;
    }
  }

  return (0);
}
/********************************************************************************/
static int
setInitialLinkUpdateTime (topology_t *network)
{
  int	    i, j;
  double    initupdtime;


  if (network == (topology_t *) 0)
    return (-1);
  
  for (i = 0; i < getNumNodes (network); i++) {
    for (j = 0; j < getNumNodes (network); j++) {
      if (areConnected(i, j, network)) {
	initupdtime = 0.0 - drand48()*refreshinterval;  /* init upd time is betw 0 and -T */
	setLinkUpdateTime (i, j, initupdtime, network);
      }
      else continue;
    }
  }
  return (0);
}
  
