/* $Id: rstraffic.c,v 1.12 1999/01/29 04:26:50 ashaikh Exp $ */

static const char rcsid[] = "$Id: rstraffic.c,v 1.12 1999/01/29 04:26:50 ashaikh Exp $";

#include "rstraffic.h"
#include "rsdefs.h"
#include "rstrace.h"
#include "usercdf.h"
#include <math.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>


typedef struct {
  double    morningsc, morningsh,
	    afternoonsc, afternoonsh,
	    eveningsc, eveningsh;
} tmatrixentry_t;

typedef struct {
  int	    size;
  double    avgrate;
  tmatrixentry_t    ***tmatrix;
} tmatrix_t;

static tmatrix_t	*trafficmatrix;

/* arrival rate specs */
arrivalspec_t	arrspec = (arrivalspec_t) -1;;
arrdist_t	arrdist = (arrdist_t) -1;
double		arrshape;
double		arrscale;
timeofday_t	timeofday = AFTERNOON;
double		trafficmult;

/* holding time specs */
holdspec_t	holdspec = (holdspec_t) -1;
double		holdshape;
double		holdscale;
cdf_t		*holdcdf;

/* bandwidth specs */
double bandwidth = 0.0;	    /* average bandwidth of calls		  */
double spread = 0.0;	    /* relative spread of bandwidth requirements */


static int getFileInterarrival (int, int, double *, double *);
static void calcAvgFileArrival (void);
static double getGamma (double);
/********************************************************************************/
int
initRSTraffic (int numnodes)
{
  /* set call arrival specification */
  if (arrspec == ARR_FILE) {
    if (initTrafficMatrix(numnodes)) {
      RS_PRINT_ERROR ("initRSTraffic: initTrafficMatrix returned error\n");
      return (-1);
    }
    if (readTrafficMatrix()) {
      RS_PRINT_ERROR ("initRSTraffic:  readTrafficMatrix returned error\n");
      return (-1);
    }
  }

  /* init CDF if necessary */
  if (holdspec == USERDEF) {
    if (readCDFFile (&holdcdf, HOLDTIME_CDF_FILE)) {
      RS_PRINT_ERROR ("initRSTraffic:  readCDFFile returned error\n");
      return (-1);
    }
  }
  else {
    holdcdf = (cdf_t *) 0;
  }

  return (0);
}

/********************************************************************************/
int
cleanupRSTraffic (void)
{
  if (arrspec == ARR_FILE) {
    if (destroyTrafficMatrix()) {
      RS_PRINT_SOFT_ERROR ("cleanupRSTraffic:  destroyTrafficMatrix failed\n");
      return (-1);
    }
  }

  if (holdspec == USERDEF) {
    destroyCDF (holdcdf);
  }
    
  return (0);
}
/********************************************************************************/
double
getBandwidth (void)
{
  double    bw;

  bw = bandwidth + ((drand48()-0.5) * (spread * bandwidth));

  return (bw);
}

double
getAvgBandwidth (void)
{
  return (bandwidth);
}

/********************************************************************************/
double
getInterarrival (int source, int dest)
{
  double    inter;
  double    shape, scale;

  /* first get the correct shape and scale */
  switch (arrspec) {
  case ARR_UNIFORM:
    scale = arrscale;
    shape = arrshape;
    break;
  case ARR_FILE:
    if (getFileInterarrival (source, dest, &scale, &shape)) {
      RS_PRINT_ERROR ("getInterarrival:  getFileInterarrival failed\n");
      return (-1.0);
    }
    scale = 1.0 / scale;    /* convert to interarrival avg */
    break;
  default:
    RS_PRINT_ALWAYS ("Invalid arrivalspec\n");
    return (-1.0);
    break;
  }
  /* now compute interarrival according to desired distribution */
  switch (arrdist) {
  case WEIBULL:
    inter = scale * pow((-log(drand48())), (1.0 / shape));
    break;
  case POISSON:
    inter = (-1.0 * scale * log(drand48()));
    break;
  default:
    RS_PRINT_ALWAYS ("Unknown arrival distribution\n");
    return (-1.0);
  }
  
  return (inter);
}
/********************************************************************************/
static int
getFileInterarrival (int source, int dest, double *scale, double *shape)
{

  if ( (trafficmatrix->tmatrix[source][dest] == (tmatrixentry_t *) 0) ||
       (source == dest) ) {
    *scale = 0.0;
    *shape = 1.0;
    return (0); /* no traffic rate from src to dest */
  }
  switch (timeofday) {
  case MORNING:
    *scale = trafficmatrix->tmatrix[source][dest]->morningsc;
    *shape = trafficmatrix->tmatrix[source][dest]->morningsh;
    break;
  case AFTERNOON:
    *scale = trafficmatrix->tmatrix[source][dest]->afternoonsc;
    *shape = trafficmatrix->tmatrix[source][dest]->afternoonsh;    
    break;
  case EVENING:
    *scale = trafficmatrix->tmatrix[source][dest]->eveningsc;
    *shape = trafficmatrix->tmatrix[source][dest]->eveningsh;
    break;
  }
  
  *scale *= trafficmult;

  return (0);
}
/********************************************************************************/
static void
calcAvgFileArrival (void)
{
  double    avgscale, avgshape;
  double    avg, sumsc, sumsh;
  double    sumscnodes, sumshnodes;
  int	    i, j, numnodes;

  numnodes = trafficmatrix->size;
  /* overall avg is computed by adding up the arrival rates from each node to
   * every other node and calling it the overall interarrival rate at the given
   * node.  Then all of these are averaged to compute the avg. interarrival
   * rate over all nodes.
   */
  sumscnodes = sumshnodes = 0.0;
  for (i = 0; i < numnodes; i++) {
    sumsc = sumsh  = 0.0;	
    for (j = 0; j < numnodes; j++) {
      if ( (i == j) || (trafficmatrix->tmatrix[i][j] == (tmatrixentry_t *) 0) )
	continue;
      switch (timeofday) {
      case MORNING:
	sumsc += trafficmatrix->tmatrix[i][j]->morningsc;
	sumsh += trafficmatrix->tmatrix[i][j]->morningsh;	    
	break;
      case AFTERNOON:
	sumsc += trafficmatrix->tmatrix[i][j]->afternoonsc;
	sumsh += trafficmatrix->tmatrix[i][j]->afternoonsh;
	break;
      case EVENING:
	sumsc += trafficmatrix->tmatrix[i][j]->eveningsc;
	sumsh += trafficmatrix->tmatrix[i][j]->eveningsh;	    
	break;
      }
    }
    /* sumsc contains the sum of arrival rates across a row of the matrix */
    sumscnodes += sumsc * trafficmult;
    /* sumsh in general is just (n-1)*shape */
    sumshnodes += (sumsh / (double) (numnodes - 1));
  }
  /* this is probably not the best way to do this */
  avgscale = sumscnodes / (double) numnodes;
  avgshape = sumsh / (double) numnodes;

  /* now compute average */
  switch (arrdist) {
  case WEIBULL:
    avg = (avgscale / avgshape) * getGamma ((1.0 / avgshape));
    break;
  case POISSON:
    avg = avgscale;
    break;
  default:
    RS_PRINT_ALWAYS ("getAvgInterarrival: Invalid arrival distribution\n");
  }

  trafficmatrix->avgrate = avg;
}
/********************************************************************************/
double
getAvgInterarrival (int source, int dest)
{
  double    avgscale, avgshape;
  double    avg;

  /* first compute avg shape and scale */
  switch (arrspec) {
  case ARR_UNIFORM:
    avgscale = arrscale;
    avgshape = arrshape;
    break;
  case ARR_FILE:
    if ( (source < 0) || (dest < 0) ) {	/* calculate overall avg */
      return (1.0 / trafficmatrix->avgrate);
    }
    else {  /* avg for one src - dest pair */
      if (trafficmatrix->tmatrix[source][dest] == (tmatrixentry_t *) 0)
	return (-1.0); /* no traffic rate from src to dest */
      switch (timeofday) {
      case MORNING:
	avgscale = trafficmatrix->tmatrix[source][dest]->morningsc;
	avgshape = trafficmatrix->tmatrix[source][dest]->morningsh;
	break;
      case AFTERNOON:
	avgscale = trafficmatrix->tmatrix[source][dest]->afternoonsc;
	avgshape = trafficmatrix->tmatrix[source][dest]->afternoonsh;	
	break;
      case EVENING:
	avgscale = trafficmatrix->tmatrix[source][dest]->eveningsc;
	avgshape = trafficmatrix->tmatrix[source][dest]->eveningsh;	
	break;
      }
      avgscale *= trafficmult;
      avgscale = 1.0 / avgscale;
    }
    break;
  default:
    RS_PRINT_ALWAYS ("getAvgInterarrival: Invalid arrivalspec\n");
    avg = -1.0;
    break;
  }
  /* now compute average */
  switch (arrdist) {
  case WEIBULL:
    avg = (avgscale / avgshape) * getGamma ((1.0 / avgshape));
    break;
  case POISSON:
    avg = avgscale;
    break;
  default:
    RS_PRINT_ALWAYS ("getAvgInterarrival: Invalid arrival distribution\n");
    return (-1.0);
  }

  return (avg);
}

double
getAvgArrival (int source, int dest)
{
  return (1.0 / getAvgInterarrival (source, dest));
}

void
printArrivalSpec (void)
{
  double    avg;

  avg = getAvgArrival (-1, -1);

  switch (arrdist) {
  case WEIBULL:
    printf ("WEIBULL (mean: %.3f, shape: %.3f)", avg, arrshape);
    break;
  case POISSON:
    printf ("POISSON (mean: %.3f)", avg);
    break;
  default:
    RS_PRINT_ALWAYS ("printArrivalSpec:  Unknown arrival distribution\n");
  }
}
/********************************************************************************/
double
getHoldTime (void)
{
  double    holdtime;
  
  switch (holdspec) {
  case EXP:
    holdtime = -1.0 * holdscale * log(drand48());
    break;
  case PARETO:
    if (holdshape < 0 || holdscale < 0)
      return (-1.0);    
    holdtime = holdscale / pow ( (drand48()), (1.0/holdshape) );
    break;
  case USERDEF:
    holdtime = sampleCDF (holdcdf);
    break;
  default:
    RS_PRINT_ERROR ("getHoldTime:  Invalid holdspec\n");
    holdtime = -1.0;
  }
  return (holdtime);
}

double
getAvgHoldTime (void)
{
  double    avg;
  
  switch (holdspec) {
  case EXP:
    avg = holdscale;
    break;
  case PARETO:
    if (holdshape <= 1)
      avg = -1.0;
    else 
      avg = holdshape * holdscale / (holdshape - 1);
    break;
  case USERDEF:
    avg = getCDFAvg (holdcdf);
    break;
  default:
    RS_PRINT_ERROR ("getAvgHoldTime:  Invalid holdspec\n");
    avg = -1.0;
  }

  return (avg);
  
}

void
printHoldSpec (void)
{
  switch (holdspec) {
  case EXP:
    printf ("EXPONENTIAL (mean: %.3f)", holdscale);
    break;
  case PARETO:
    printf ("PARETO (shape: %.3f, scale: %.3f)\n", holdshape, holdscale);
    break;
  case USERDEF:
    printf ("USER DEFINED (mean: %.3f)", getCDFAvg (holdcdf));
    break;
  default:
    RS_PRINT_ERROR ("printHoldSpec: Invalid holdspec\n");
  }
}


/********************************************************************************/
int
initTrafficMatrix (int numnodes)
{
  int	i,j;
  
  
  if (arrspec != ARR_FILE) {
    RS_PRINT_ERROR ("Invalid arrivalspec: expecting ARR_FILE\n");
    return (-1);
  }
  
#if RS_TRACE_LEVEL >= TRACE_EVENTS
  printf ("Allocating traffic matrix ...");
  fflush (stdout);
#endif
  trafficmatrix = (tmatrix_t *) malloc (sizeof(tmatrix_t));
  if (trafficmatrix == (tmatrix_t *) 0) {
    RS_PRINT_SOFT_ERROR ("Couldn't allocate traffic matrix\n");
    return (-1);
  }
  
  trafficmatrix->tmatrix = (tmatrixentry_t ***) malloc (sizeof(tmatrixentry_t *)*numnodes);
  if (trafficmatrix->tmatrix == (tmatrixentry_t ***) 0) {
    RS_PRINT_SOFT_ERROR ("Couldn't allocate traffic matrix entries\n");
    return (-1);
  }
  trafficmatrix->size = numnodes;
  for (i = 0; i < numnodes; i++) {
    trafficmatrix->tmatrix[i] = (tmatrixentry_t **) malloc (sizeof(tmatrixentry_t *)*numnodes);
    if (trafficmatrix->tmatrix[i] == (tmatrixentry_t **) 0) {
      RS_PRINT_SOFT_ERROR ("Couldn't allocate traffic matrix entry\n");
      return (-1);
    }
  }
  /* put NULL pointers in each array position */
  for (i = 0; i < numnodes; i++)
    for (j = 0; j < numnodes; j++) {
      trafficmatrix->tmatrix[i][j] = (tmatrixentry_t *) 0;
    }
#if RS_TRACE_LEVEL >= TRACE_EVENTS  
  printf ("Done.\n");
#endif
  return (0);
}

/********************************************************************************/
int
readTrafficMatrix (void)
/* Traffic matrix input format:
 *  <from> <to> <morningscale> <morningshape> <afternoonscale> <afternoonshape> <eveningscale> <eveningshape>
 *
 *  scales are actual arrival rates -- not interarrival time!
 *
 */
{
  FILE		    *trafficfile;
  char		    *line, *result;
  tmatrixentry_t    *newentry;
  int		    from, to;
  double	    morningscale, morningshape;
  double	    afternoonscale, afternoonshape;
  double	    eveningscale, eveningshape;
  int		    scanresult;
  
  trafficfile = fopen (TRAFFICFILENAME, "r");
  if (trafficfile == NULL) {
#if RS_TRACE_LEVEL >= TRACE_ERRORS
    fprintf (stderr, "Error opening %s for reading: %s\n",
	     TRAFFICFILENAME, strerror(errno));
#endif
    return (-1);
  }
  
#if RS_TRACE_LEVEL >= TRACE_EVENTS
  printf ("Reading traffic matrix ...");
  fflush(stdout);
#endif
  
  line = (char *) malloc (MAX_TRAFFICFILE_LINE);
  if (line == (char *) 0) {
    RS_PRINT_ERROR ("readTrafficMatrix: Allocation failure\n");
    return (-1);
  }
  
  result = fgets (line, MAX_TRAFFICFILE_LINE, trafficfile);
  while (result != (char *) 0) {
    
    scanresult = sscanf (line, "%d %d %lf %lf %lf %lf %lf %lf\n",
			 &from,
			 &to,
			 &morningscale,
			 &morningshape,
			 &afternoonscale,
			 &afternoonshape,
			 &eveningscale,
			 &eveningshape);
    
    if (scanresult < 8) {
      /* not very forgiving */
#if RS_TRACE_LEVEL >= TRACE_SOFT_ERRORS
      fprintf (stderr, "Error parsing traffic matrix file %s\n", TRAFFICFILENAME);
#endif
      return (-1);
    }

    /* no duplicates */
    if (trafficmatrix->tmatrix[from][to] != (tmatrixentry_t *) 0) {
#if RS_TRACE_LEVEL >= TRACE_ERRORS
      fprintf (stderr, "Duplicate entries (%d -> %d) in traffic matrix file\n", from, to);
#endif
      return (-1);
    }
    
    /* allocate tmatrix structure */
    newentry = (tmatrixentry_t *) malloc (sizeof(tmatrixentry_t));
    if (newentry == 0) {
      RS_PRINT_SOFT_ERROR ("tmatrixentry allocation failed\n");
      return (-1);
    }
    /* check for defaults */
    if (morningscale == -1.0) {
      newentry->morningsc = DEFAULT_SCALE;
      newentry->morningsh = DEFAULT_SHAPE;
    }
    else {
      newentry->morningsc = morningscale;
      newentry->morningsh = morningshape;
    }
    if (afternoonscale == -1.0) {
      newentry->afternoonsc = DEFAULT_SCALE;
      newentry->afternoonsh = DEFAULT_SHAPE;
    }
    else {
      newentry->afternoonsc = afternoonscale;
      newentry->afternoonsh = afternoonshape;
    }
    if (eveningscale == -1.0) {
      newentry->eveningsc = DEFAULT_SCALE;
      newentry->eveningsh = DEFAULT_SHAPE;
    }
    else {
      newentry->eveningsc = eveningscale;
      newentry->eveningsh = eveningshape;
    }    

    trafficmatrix->tmatrix[from][to] = newentry;

    /* read the next line */
    result = fgets (line, MAX_TRAFFICFILE_LINE, trafficfile);
  }
  
  fclose (trafficfile);
  
#if RS_TRACE_LEVEL >= TRACE_EVENTS
  printf ("Done\n");
#endif

  calcAvgFileArrival ();
  
  return (0);
}

/********************************************************************************/
int
destroyTrafficMatrix (void)
{
  int	i,j, numnodes;

  numnodes = trafficmatrix->size;
  for (i = 0; i < numnodes; i++)
    for (j = 0; j < numnodes; j++) {
      if (trafficmatrix->tmatrix[i][j] != (tmatrixentry_t *) 0)
	free (trafficmatrix->tmatrix[i][j]);
    }
  for (i = 0; i < numnodes; i++) {
    if (trafficmatrix->tmatrix[i] != (tmatrixentry_t **) 0) {
      free (trafficmatrix->tmatrix[i]);
    }
  }
  free (trafficmatrix->tmatrix);
  free (trafficmatrix);
  
  return (0);
}
/********************************************************************************/
static double
getGamma (double x)
{
  double lg;
  
  lg = lgamma(x);
  return (signgam*exp(lg));
}

