/* $Id: topology.c,v 3.2 1999/01/08 05:26:59 ashaikh Exp $ */

static const char rcsid[] = "$Id: topology.c,v 3.2 1999/01/08 05:26:59 ashaikh Exp $";

#include "topology.h"
#include "topkncube.h"
#include "topfile.h"
#include "topfull.h"
#include "toprandom.h"
#include "rstrace.h"
#include "general.h"
#include "sptgeneric.h"
#include <math.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>



static void printLinkCapacityStats (topology_t *);
/********************************************************************************/
int
initNetwork (topology_t **net, topologytype_t toptype)
{
  topology_t	*network;

  *net = (topology_t *) 0;
  
  /* allocate network */
  network = (topology_t *) malloc (sizeof(topology_t));
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("initNetwork: Could not allocate network structure\n");
    return (-1);
  }
  network->links = (link_t ***) 0;
  network->hoptable = (int **) 0;
  network->info = (void *) 0;

  network->stats = (rsstat_t *) malloc (sizeof (rsstat_t));
  if (network->stats == (rsstat_t *) 0) {
    RS_PRINT_ERROR ("initNetwork:  Could not allocate network statistics\n");
    return (-1);
  }
  
  switch (toptype) {
  case TOP_KNCUBE:
    network->type = TOP_KNCUBE;
    break;
  case TOP_FILE:
    network->type = TOP_FILE;
    break;
  case TOP_FULL:
    network->type = TOP_FULL;
    break;
  case TOP_RAND:
    network->type = TOP_RAND;
    break;
  default:
    RS_PRINT_ERROR ("Unknown topology type\n");
    return (-1);
  }

  /* initialize these elements -- they get filled in later */
  network->numnodes = network->numlinks = network->diameter = 0;
  initSimpleRatioStat (&(network->hopcount), DEFAULT_CONFIDENCE_LEVEL);
  initSimpleRatioStat (&(network->degree), DEFAULT_CONFIDENCE_LEVEL);
  network->blkpolicy = BLOCKING;    /* default blocking policy */

  *net = network;

  return (0);
}
/********************************************************************************/
int
initLinkMatrix(topology_t *network)
{
  int i, j;
  int	num;

#if RS_TRACE_LEVEL >= TRACE_EVENTS
  printf ("Creating network matrix ...");
#endif
  if (network->numnodes == 0) {
    RS_PRINT_SOFT_ERROR ("initLinkMatrix: network size not initialized\n");
    return (-1);
  }
  num = network->numnodes;

  network->links = (link_t ***) malloc (sizeof(link_t *)*num);
  if (network->links == (link_t ***) 0) {
    RS_PRINT_ERROR ("Couldn't allocate network matrix\n");
    return (-1);
  }
  for (i = 0; i < num; i++) {
    network->links[i] = (link_t **) malloc (sizeof(link_t *)*num);
    if (network->links[i] == (link_t **) 0) {
      RS_PRINT_SOFT_ERROR ("Couldn't allocate network matrix entry\n");
      return (-1);
    }
  }
  /* put NULL pointers in each array position */
  for (i = 0; i < num; i++)
    for (j = 0; j < num; j++) {
      network->links[i][j] = (link_t *) 0;
    }
#if RS_TRACE_LEVEL >= TRACE_EVENTS
  printf ("Done.\n");
#endif
  return (0);
}
/********************************************************************************/
int
createKaryNcube (int k, int n, topology_t **network)
{
  return (genKaryNcube (k, n, network));
}

int
createFileTopology (char *infile, topology_t **network)
{
  return (genFileTopology (infile, network));
}

int
readTopology (topology_t **network)
{
  return (readStdinTopology (network));
}

int
createFullTopology (int size, topology_t **network)
{
  return (genFullTopology (size, network));
}

int
createRandTopology (randgraph_t method, int topsize, double alpha, double beta, double planesize,
		 boolean asym, int maxdiam, topology_t **network)
{
  return (genRandTopology (method, topsize, alpha, beta, planesize, asym, maxdiam, network));
}


/********************************************************************************/
int
splitNetCapacity (topology_t *net, double split, topology_t **net1, topology_t **net2)
/* Partitions the network net into two networks net1 and net2 with the original capacity
 * split according to the split parameter.  This function does a *lot* of memory copying.
 */
{
  topology_t	*neta, *netb;
  link_t	*linka, *linkb;
  int		i, j;

  if (net == (topology_t *) 0) {
    RS_PRINT_ERROR ("splitNetCapacity: Invalid network\n");
    return (-1);
  }

  if ( (split < 0) || (split > 1) ) {
    RS_PRINT_ERROR ("splitNetCapacity: split parameter out of range\n");
    return (-1);
  }

  if ( (initNetwork (&neta, net->type)) ||
       (initNetwork (&netb, net->type)) ) {
    RS_PRINT_ERROR ("splitNetCapacity:  Could not initialize new networks\n");
    return (-1);
  }

  /* copy topology characteristics over */
  neta->numnodes = netb->numnodes = net->numnodes;
  neta->numlinks = netb->numlinks = net->numlinks;
  neta->diameter = netb->diameter = net->diameter;
  copySimpleRatioStat (&(net->hopcount), &(neta->hopcount));
  copySimpleRatioStat (&(net->hopcount), &(netb->hopcount));    
  copySimpleRatioStat (&(net->degree), &(neta->degree));
  copySimpleRatioStat (&(net->degree), &(netb->degree));  
  neta->blkpolicy = netb->blkpolicy = net->blkpolicy;
  neta->routingalg = netb->routingalg = net->routingalg;
  /* all three info pointers point to the same memory location */
  neta->info = netb->info = net->info;
  
  if ( (initLinkMatrix(neta)) ||
       (initLinkMatrix(netb)) ) {
    RS_PRINT_ERROR ("splitNetCapacity:  Unable to initialize link matrices\n");
    return (-1);
  }

  if ( (initDistTable(neta)) ||
       (initDistTable(netb)) ) {
    RS_PRINT_ERROR ("splitNetCapacity:  Unable to initialize hop tables\n");
    return (-1);
  }

  /* copy hop table and create new link matrices */
  for (i = 0; i < net->numnodes; i++) {
    for (j = 0; j < net->numnodes; j++) {
      neta->hoptable[i][j] = netb->hoptable[i][j] = net->hoptable[i][j];
      if (net->links[i][j] != (link_t *) 0) {
	linka =  genLink(i, j, 0);
	linkb =  genLink(i, j, 0);
	if ( (linka == (link_t *) 0) || (linkb == (link_t *) 0) ) {
	  RS_PRINT_ERROR ("splitNetCapacity:  Error in genLink()\n");
	  return (-1);
	}
	linka->capacity = split * net->links[i][j]->capacity;
	linkb->capacity = (1.0 - split) * net->links[i][j]->capacity;
	if ( (genLinkVariables (linka)) || (genLinkVariables (linkb)) ){
	  RS_PRINT_ERROR ("splitNetCapacity: genLinkVariables returned error\n");
	  return (-1);
	}
	linka->monitor = linkb->monitor = net->links[i][j]->monitor;
	neta->links[i][j] = linka;
	netb->links[i][j] = linkb;
      }
    }
  }

  calcLinkCapacityStats(neta);
  calcLinkCapacityStats(netb);
 
  *net1 = neta;
  *net2 = netb;

  return (0);
}

/********************************************************************************/
int
dumpTopologyFile (char *outfile, topology_t *network)
{
  FILE	    *topfile;
  int	    i,j;
  int	    printresult;

  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("dumpTopologyFile: Invalid network\n");
    return (-1);
  }

  topfile = fopen (outfile, "w");
  if (topfile == NULL) {
#if RS_TRACE_LEVEL >= TRACE_ERRORS
    fprintf (stderr, "Error opening %s for writing: %s\n",
	     outfile, strerror(errno));
#endif
    return (-1);
  }

  printresult = fprintf (topfile, "%d\n", network->numnodes);
  if (printresult < 0) {
#if RS_TRACE_LEVEL >= TRACE_ERRORS    
    fprintf (stderr, "Error writing num. nodes to %s\n", outfile);
#endif
    fclose (topfile);
    return (-1);
  }
  for (i = 0; i < network->numnodes; i++) {
    for (j = 0; j < network->numnodes; j++) {
      if (i == j || network->links[i][j] == (link_t *) 0)
	continue;
      fprintf (topfile, "%d %d %f %f %f %d\n",
	       network->links[i][j]->from,
	       network->links[i][j]->to,
	       network->links[i][j]->capacity,
	       network->links[i][j]->propdelay,
	       network->links[i][j]->adminweight,
	       network->links[i][j]->maxcalls);
    }
  }
  fclose (topfile);
  return (0);

}
/********************************************************************************/
void
printTopologyStats (topology_t *network)
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("printTopologyStats: Invalid network\n");
    return;
  }
  printf ("Topology Statistics:\n");
  printf ("\tType: ");
  printTopFunction (network);
  printf ("\n");
  printf ("\tNodes:\t%d\n", network->numnodes);
  printf ("\tLinks:\t%d\n", network->numlinks);
  printf ("Hopcount:\n");
  printf ("\tDiameter\t%d\n", network->diameter);  
  printf ("\tAvg. Hopcount\t%.3f\n", network->hopcount.currentrate);
  printf ("\tVariance\t%.3f\n", network->hopcount.currentvar);
  printf ("\tMax. Hopcount\t%.3f\n", network->hopcount.max);
  printf ("\tMin. Hopcount\t%.3f\n", network->hopcount.min); 
  printf ("Node Degree:\n");
  printf ("\tAvg. Degree\t%.3f\n", network->degree.currentrate);
  printf ("\tDegree Variance\t%.3f\n", network->degree.currentvar);
  printf ("\tMax. Degree\t%.3f\n", network->degree.max);
  printf ("\tMin. Degree\t%.3f\n", network->degree.min); 
  printLinkCapacityStats(network);

}

/********************************************************************************/
int
clearLinkMatrix (topology_t *network)
{
  int	i,j;

  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("clearLinkMatrix: Invalid network\n");
    return (-1);
  }
  
  /* put NULL pointers in each array position */
  for (i = 0; i < network->numnodes; i++) {
    for (j = 0; j < network->numnodes; j++) {
      if (network->links[i][j] != (link_t *) 0)
	free (network->links[i][j]);
      network->links[i][j] = (link_t *) 0;
    }
  }
  network->numlinks = 0;
  network->diameter = 0;
  initSimpleRatioStat (&(network->hopcount), DEFAULT_CONFIDENCE_LEVEL);
  initSimpleRatioStat (&(network->degree), DEFAULT_CONFIDENCE_LEVEL);

  return (0);
}

/********************************************************************************/
int
initLinkMonitor (topology_t *network)
{
  FILE	    *monfile;
  char	    *line, *result;
  int	    scanresult;
  int	    src, dest;

  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("initLinkMonitor: Invalid network\n");
    return (-1);
  }
  
  monfile = fopen (MONFILENAME, "r");
  if (monfile == NULL) {
#if RS_TRACE_LEVEL >= TRACE_ERRORS
    fprintf (stderr, "Error opening %s for reading: %s\n",
	     MONFILENAME, strerror(errno));
#endif
    return (-1);
  }

  line = (char *) malloc (MAX_MONFILE_LINE);

  /* read links */
  result = fgets (line, MAX_MONFILE_LINE, monfile);

  while (result != (char *) 0) {
    scanresult = sscanf (line, "%d %d \n", &src, &dest);
    if (scanresult != 2) {
#if RS_TRACE_LEVEL >= TRACE_SOFT_ERRORS
      fprintf (stderr, "Error parsing link from file %s. File empty?\n", MONFILENAME);
#endif
      return (-1);
    }

    if (network->links[src][dest] == (link_t *)0) {	    /* no link there */
#if RS_TRACE_LEVEL >= TRACE_ERRORS
      fprintf (stderr, "Specified link %d->%d does not exist, skipping...\n", src, dest);
#endif
    }
    else {
      network->links[src][dest]->monitor = 1;
    }
    
    result = fgets (line, MAX_MONFILE_LINE, monfile);
  }
  fclose (monfile);
  
  return (0);
}

/********************************************************************************/
int
calcDistTable (topology_t *net)
{
  int	i, j;
  int	**srcpaths;

  if (net == (topology_t *) 0) {
    RS_PRINT_ERROR ("calcDistTable: Invalid network\n");
    return (-1);
  }
  
  /* create an array for each source to store its source-specific SPT */
  srcpaths = (int **) malloc (sizeof(int) * net->numnodes);
  for (i = 0; i < net->numnodes; i++)
    srcpaths[i] = (int *) malloc (sizeof(int) * net->numnodes);

  for (i = 0; i < net->numnodes; i++)
    calcSimpleSPT(net, i, srcpaths[i]);	/* generate SPT rooted at i */

  /* now that each srcpaths[i] has an SPT, extract distance */
  for (i = 0; i < net->numnodes; i++) {
    for (j = 0; j < net->numnodes; j++) {
      if (i == j) continue;
      net->hoptable[i][j] = sptGetSpecPathLen (i, j, srcpaths[i]);
    }
  }

#if RS_TRACE_LEVEL >= TRACE_DETAILS
    printDistTable (net);
#endif

  for (i = 0; i < net->numnodes; i++)
    free (srcpaths[i]);
  free (srcpaths);

  return (0);
}
/********************************************************************************/
int
parseDistTable (topology_t *net)
{
  int	i, j;
  int	max, sumhops, paths;
  int	neighbors;

  if (net == (topology_t *) 0) {
    RS_PRINT_ERROR ("parseDistTable: Invalid network\n");
    return (-1);
  }

  sumhops = 0;
  max = 0;
  paths = 0;

  for (i = 0; i < net->numnodes; i++) {
    neighbors = 0;
    for (j = 0; j < net->numnodes; j++) {
      if (net->links[i][j] != (link_t *) 0) {
	neighbors++;
      }
      if (i == j) continue;
      if (net->hoptable[i][j] > 0) {
	updateSimpleRatioStat (net->hoptable[i][j], &(net->hopcount));
      }
      else {
	RS_PRINT_SOFT_ERROR ("Bad entry in disttable\n");
	return (-1);
      }
    }
    updateSimpleRatioStat (neighbors, &(net->degree));
  }
  
  net->diameter = (int) net->hopcount.max;

  return (0);
}
/********************************************************************************/
int
initDistTable (topology_t *net)
{
  int	i, j;
  
  /* initialize the distance table */
  net->hoptable = (int **) malloc (sizeof(int) * net->numnodes);
  if (net->hoptable == (int **) 0) {
    RS_PRINT_ERROR ("initDistTable:  Couldn't allocate distance table\n");
    return (-1);
  }
  for (i = 0; i < net->numnodes; i++) {
    net->hoptable[i] = (int *) malloc (sizeof(int) * net->numnodes);
    if (net->hoptable[i] == (int *) 0) {
      RS_PRINT_ERROR ("initDistTable:  Couldn't allocate distance table entry\n");
      return (-1);
    }
    for (j = 0; j < net->numnodes; j++)
      net->hoptable[i][j] = -1;
  }

  return (0);
}

int
clearDistTable (topology_t *net)
{
  int	i, j;

  if (net == (topology_t *) 0) {
    RS_PRINT_ERROR ("clearDistTable: Invalid network\n");
    return (-1);
  }
  for (i = 0; i < net->numnodes; i++) {
    for (j = 0; j < net->numnodes; j++) {
      net->hoptable[i][j] = -1;
    }
  }
  net->diameter = 0;
  initSimpleRatioStat (&(net->hopcount), DEFAULT_CONFIDENCE_LEVEL);
  initSimpleRatioStat (&(net->degree), DEFAULT_CONFIDENCE_LEVEL);

  return (0);
}

int
destroyDistTable (topology_t *net)
{
  int	i;

  if (net == (topology_t *) 0) {
    RS_PRINT_ERROR ("destroyDistTable: Invalid network\n");
    return (-1);
  }
  
  for (i = 0; i < net->numnodes; i++)
    free (net->hoptable[i]);

  free (net->hoptable);
  net->hoptable = (int **) 0;

  return (0);
}
/********************************************************************************/  
#if RS_TRACE_LEVEL >= TRACE_DETAILS
void
printDistTable (topology_t *network)
{
  int	i, j;

  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("printDistTable: Invalid network\n");
    return;
  }
  printf ("Hop table:\n");
  printf ("  ");
  for (i = 0; i < network->numnodes; i++)
    printf ("%2d ", i);
  printf ("\n  ");		    /* set up table heading */
  for (i = 0; i < network->numnodes; i++)
    printf ("---");
  printf ("\n");
  for (i = 0; i < network->numnodes; i++) {
    printf ("%d|", i);
    for (j = 0; j < network->numnodes; j++) {
      printf ("%2d ", network->hoptable[i][j]);
    }
    printf ("\n");
  }
}
#endif
/********************************************************************************/

void
printTopFunction (topology_t *network)
{
  randgraph_t	*method;
  
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("printTopFunction: Invalid network\n");
    return;
  }
  switch (network->type) {
  case TOP_KNCUBE:
    printf ("K-ARY N-CUBE");
    break;
  case TOP_FILE:
    printf ("FILE");
    break;
  case TOP_FULL:
    printf ("FULLY CONN");
    break;
  case TOP_RAND:
    printf ("RANDOM ");
    method =  (randgraph_t *) network->info;
    switch (*method) {
    case PURE:
      printf ("(PURE)");
      break;
    case RG2:
      printf ("(RG2)");
      break;
    default:
      printf ("(UNKNOWN)");
    }
    break;
  default:
    printf ("UNKNOWN");
  }
}

/********************************************************************************/
int
getNumNodes (topology_t *network)
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("getNumNodes: Invalid network\n");
    return (-1);
  }
		  
  return (network->numnodes);
}

int
getDiameter (topology_t *network)
{
    if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("getDiameter: Invalid network\n");
    return (-1);
  }
		  
  return (network->diameter);

}

double
getAvgHopcount (topology_t *network)
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("getAvgHopcount: Invalid network\n");
    return (-1);
  }
		  
  return (network->hopcount.currentrate);
}

int
getMinHops (int src, int dest, topology_t *network)
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("getMinHops: Invalid network\n");
    return (-1);
  }

  return (network->hoptable[src][dest]);

}

int
getNumLinks (topology_t *network)
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("getNumLinks: Invalid network\n");
    return (-1);
  }
		  
  return (network->numlinks);
}

double
getAvgDegree (topology_t *network)
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("getAvgDegree: Invalid network\n");
    return (-1);
  }
		  
  return (network->degree.currentrate);
}

/********************************************************************************/
int
updateLinks (topology_t *network)
/* updates bandwidth on all links */
{
  int	i, j;

  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("updateLinks: Invalid network\n");
    return (-1);
  }
  for (i = 0; i < network->numnodes; i++)
    for (j = 0; j < network->numnodes; j++) {
      /* update each link */
      if (network->links[i][j])
	network->links[i][j]->availbw.last = network->links[i][j]->availbw.current;
    }
  return (0);
}
/********************************************************************************/
int
updateSingleLink (int a, int b, topology_t *network)
/* updates bandwidth on a single links */
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("updateSingleLink: Invalid network\n");
    return (-1);
  }

  if (network->links[a][b])
    network->links[a][b]->availbw.last = network->links[a][b]->availbw.current;
  return (0);
}
/********************************************************************************/
int
setLinkUpdateTime (int a, int b, double time, topology_t *network)
/* set the last update time for a single link */
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("setLinkUpdateTime: Invalid network\n");
    return (-1);
  }

  network->links[a][b]->lastupdtime = time;

  return (0);
}

int
setLinkUtilUpdateTime (int a, int b, double time, topology_t *network)
/* set the last update time for a single link */
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("setLinkUtilUpdateTime: Invalid network\n");
    return (-1);
  }

  network->links[a][b]->lastutilupd = time;

  return (0);
}
/********************************************************************************/
int
setBlockingPolicy (blockingpolicy_t policy, topology_t *network)
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("setBlockingPolicy: Invalid network\n");
    return (-1);
  }

  network->blkpolicy = policy;

  return (0);

}

void
printBlockingPolicy (topology_t *network)
{
  switch (network->blkpolicy) {
  case NON_BLOCKING:
    printf ("NON BLOCKING");
    break;
  case BLOCKING:
    printf ("BLOCKING");
    break;
  default:
    printf ("UNKNOWN");
    break;
  }
}
/********************************************************************************/
link_t *
genLink (int a, int b, int id)
/* Allocate and fill in link structure
 */
{
 link_t *newlink;

 /* allocate link structure */
 newlink = (link_t *) malloc (sizeof(link_t));
 if (newlink == 0) {
   RS_PRINT_ERROR ("genLink: link allocation failed\n");
   return ((link_t *) 0);
 }
 /* fill in link parameters -- defaults for now */
 newlink->id = id;
 newlink->from = a;
 newlink->to = b;
 newlink->capacity = DEFAULT_LINK_CAPACITY;
 newlink->adminweight = DEFAULT_LINK_ADMINWEIGHT;
 newlink->propdelay = DEFAULT_LINK_PROPDELAY;
 newlink->maxcalls = DEFAULT_LINK_MAXCALLS;
 
 

 return (newlink);
}
/********************************************************************************/
int
genLinkVariables (link_t *newlink)
/* This function fills in initial values for link variables after fixed link parameters
 * have been specified.
 */
{
  if (newlink == (link_t *) 0) {
    RS_PRINT_ERROR ("genLinkVariables: invalid link structure\n");
    return (-1);
  }

  newlink->pendingupdate = 0;
  newlink->lastupdtime = 0.0;
  newlink->availbw.current = newlink->availbw.last = newlink->capacity;
  newlink->delay.current = newlink->delay.last = newlink->propdelay;
  newlink->hopcount.current = newlink->hopcount.last = (double)DEFAULT_LINK_HOPCOUNT;
  newlink->lossratio.current = newlink->lossratio.last = (double)DEFAULT_LINK_LOSSRATIO;
  newlink->cost.current = newlink->cost.last = (double)DEFAULT_LINK_COST;
  newlink->numcalls = 0;
  newlink->callsceil = 0;
  if (initSimpleRatioStat(&(newlink->util), DEFAULT_CONFIDENCE_LEVEL)) {
    RS_PRINT_ERROR ("genLinkVariables:  could not initialize link util statistic\n");
    free (newlink);
    return (-1);
  }
  newlink->lastutilupd = 0.0;
  newlink->calls = 0;
  newlink->active = 1;
  newlink->down = 0;
  newlink->monitor = 0;

  return (0);
}

int
initLinkUtilStat (link_t *link, double conf)
/* In case the DEFAULT_CONFIDENCE_LEVEL is not satisfactory, or stat needs to be reset
 * in the course of simulation.
 */
{
  if (link == (link_t *) 0) {
    RS_PRINT_ERROR ("initLinkUtilStat:  Invalid link structure\n");
    return (-1);
  }

  if (initSimpleRatioStat(&(link->util), conf)) {
    RS_PRINT_ERROR ("initLinkUtilStat:  could not initialize link util statistic\n");
    return (-1);
  }

  return (0);
}
/********************************************************************************/
int
areConnected (int a, int b, topology_t *network)
/* Check if there is a link between nodes a and b.
 */
{
  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("areConnected: Invalid network\n");
    return (0);
  }

  if (network->links[a][b] != (link_t *) 0)
    return (1);
  else
    return (0);
}

link_t *
getLinkRef (int a, int b, topology_t *network)
/* Return a handle on the link between nodes a and b.  Will return NULL
 * if no such link exists.
 */
{
  return (network->links[a][b]);
}

    
/********************************************************************************/
void
printCallsPerLinkStats (topology_t *network)
{
  int	    i, j;
  int	    sum, globalmax, unused;

  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("printCallsPerLinkStats: Invalid network\n");
    return;
  }
  sum = 0;
  globalmax = 0;
  unused = 0;
  for (i = 0; i < network->numnodes; i++) {
    for (j = 0; j < network->numnodes; j++) {
      /* update each link */
      if (network->links[i][j] != (link_t *) 0) {
	sum += network->links[i][j]->callsceil;
	if (network->links[i][j]->callsceil > globalmax)
	  globalmax = network->links[i][j]->callsceil;
	if (network->links[i][j]->callsceil  == 0)
	  unused++;
      }
    }
  }
  printf ("  %-25s\t%.2f\n", "Avg. Max calls per link:", ((float) sum / (float) network->numlinks));
  printf ("  %-25s\t%d\n", "Max calls on any link:", globalmax);
  printf ("  %-25s\t%d\n", "Unused links:", unused);

}

void
printDualCallsPerLinkStats (topology_t *net1, topology_t *net2)
{
  int	    i, j;
  int	    sum1, globalmax1, unused1;
  int	    sum2, globalmax2, unused2;  

  if ( (net1 == (topology_t *) 0) ||
       (net2 == (topology_t *) 0) ){
    RS_PRINT_ERROR ("printCallsPerLinkStats: Invalid network\n");
    return;
  }
  sum1 = sum2 = 0;
  globalmax1 = globalmax2 = 0;
  unused1 = unused2 = 0;
  for (i = 0; i < net1->numnodes; i++) {
    for (j = 0; j < net1->numnodes; j++) {
      /* update each link */
      if (net1->links[i][j] != (link_t *) 0) {
	sum1 += net1->links[i][j]->callsceil;
	sum2 += net2->links[i][j]->callsceil;	
	if (net1->links[i][j]->callsceil > globalmax1)
	  globalmax1 = net1->links[i][j]->callsceil;
	if (net2->links[i][j]->callsceil > globalmax2)
	  globalmax2 = net2->links[i][j]->callsceil;	
	if (net1->links[i][j]->callsceil  == 0)
	  unused1++;
	if (net2->links[i][j]->callsceil  == 0)
	  unused2++;	
      }
    }
  }
  printf ("  %-25s\t%.2f\t\t%.2f\n", "Avg. Max calls per link:",
	  ((float) sum1 / (float) net1->numlinks),
	 ((float) sum2 / (float) net2->numlinks));
  printf ("  %-25s\t%d\t\t%d\n", "Max calls on any link:", globalmax1, globalmax2);
  printf ("  %-25s\t%d\t\t%d\n", "Unused links:", unused1, unused2);

}
/********************************************************************************/

void
printLinkUtilStats (topology_t *network)
{
  int		    i, j;
  simpleratio_t	    globalavg;
  double	    globalmax, globalmin;
  double	    weightedavg;


  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("printLinkUtilStats: Invalid network\n");
    return;
  }
  
  initSimpleRatioStat (&globalavg, DEFAULT_CONFIDENCE_LEVEL);
  globalmax = -(DBL_MAX);
  globalmin = (DBL_MAX);
  weightedavg = 0;
  for (i = 0; i < network->numnodes; i++) {
    for (j = 0; j < network->numnodes; j++) {
      /* update each link */
      if (network->links[i][j] != (link_t *) 0) {
	/* this takes an average over all avg. link utilizations */
	updateSimpleRatioStat (network->links[i][j]->util.currentrate, &globalavg);
	if (network->links[i][j]->util.currentrate > globalmax) {
	  globalmax = network->links[i][j]->util.currentrate;
	}
	if (network->links[i][j]->util.currentrate < globalmin) {
	  globalmin = network->links[i][j]->util.currentrate;
	}
	weightedavg += network->links[i][j]->util.currentrate * network->links[i][j]->capacity;
      }
    }
  }
  weightedavg /= getTotalLinkCapacity(network);
  printf ("  %-25s\t%.3f\n", "Avg. link util (all links):", globalavg.currentrate);
  printf ("  %-25s\t%.3f\n",  "Link util variance (all links):", globalavg.currentvar);
  printf ("  %-25s\t%.3f\n", "Max avg. link util:", globalmax);
  printf ("  %-25s\t%.3f\n", "Min avg. link util:", globalmin);
  printf ("  %-25s\t%.3f\n", "Normalized avg. link util:", weightedavg);
}

void
printDualLinkUtilStats (topology_t *net1, topology_t *net2)
{
  int		    i, j;
  simpleratio_t	    globalavg1, globalavg2;
  double	    globalmax1, globalmin1, globalmax2, globalmin2;
  double	    weightedavg1, weightedavg2;

  if ( (net1 == (topology_t *) 0) || (net2 == (topology_t *) 0)) {
    RS_PRINT_ERROR ("printLinkUtilStats: Invalid network\n");
    return;
  }
  
  initSimpleRatioStat (&globalavg1, DEFAULT_CONFIDENCE_LEVEL);
  initSimpleRatioStat (&globalavg2, DEFAULT_CONFIDENCE_LEVEL);  
  globalmax1 = globalmax2 = -(DBL_MAX);
  globalmin1 = globalmin2 = (DBL_MAX);
  weightedavg1 = weightedavg2 = 0.0;
  for (i = 0; i < net1->numnodes; i++) {
    for (j = 0; j < net1->numnodes; j++) {
      /* update each link */
      if (net1->links[i][j] != (link_t *) 0) {
	/* this takes an average over all avg. link utilizations */
	updateSimpleRatioStat (net1->links[i][j]->util.currentrate, &globalavg1);
	updateSimpleRatioStat (net2->links[i][j]->util.currentrate, &globalavg2);	
	if (net1->links[i][j]->util.currentrate > globalmax1) {
	  globalmax1 = net1->links[i][j]->util.currentrate;
	}
	if (net2->links[i][j]->util.currentrate > globalmax2) {
	  globalmax2 = net2->links[i][j]->util.currentrate;
	}	
	if (net1->links[i][j]->util.currentrate < globalmin1) {
	  globalmin1 = net1->links[i][j]->util.currentrate;
	}
	if (net2->links[i][j]->util.currentrate < globalmin2) {
	  globalmin2 = net2->links[i][j]->util.currentrate;
	}
	weightedavg1 += net1->links[i][j]->util.currentrate * net1->links[i][j]->capacity;
	weightedavg2 += net2->links[i][j]->util.currentrate * net2->links[i][j]->capacity;
      }
    }
  }
  weightedavg1 /= getTotalLinkCapacity(net1);
  weightedavg2 /= getTotalLinkCapacity(net2);  
  printf ("  %-25s\t%.3f\t\t%.3f\n", "Avg. link util (all links):",
	  globalavg1.currentrate, globalavg2.currentrate);
  printf ("  %-25s\t%.3f\t\t%.3f\n",  "Link util variance:",
	  globalavg1.currentvar, globalavg2.currentvar);  
  printf ("  %-25s\t%.3f\t\t%.3f\n", "Max avg. link util:", globalmax1, globalmax2);
  printf ("  %-25s\t%.3f\t\t%.3f\n", "Min avg. link util:", globalmin1, globalmin2);
  printf ("  %-25s\t%.3f\t%.3f\n", "Normalized avg. link util:", weightedavg1, weightedavg2);  
}

/********************************************************************************/
#if RS_TRACE_LEVEL >= TRACE_DETAILS
void
printCapacityTable (topology_t *network)
{
  int	i, j, numnodes;

  numnodes = network->numnodes;  

  printf ("Link Capacity Matrix:\n");
  printf ("   ");
  for (i = 0; i < numnodes; i++)
    printf (" %3d ", i);
  printf ("\n  ");		    /* set up table heading */
  printf ("  ");  
  for (i = 0; i < numnodes; i++)
    printf ("-----");
  printf ("\n");
  for (i = 0; i < numnodes; i++) {
    printf ("%3d|", i);
    for (j = 0; j < numnodes; j++) {
      if (network->links[i][j] == (link_t *) 0)
	printf ("  X  ");
      else
	printf (" %.1f ", network->links[i][j]->capacity);
    }
    printf ("\n");
  }
}
#endif
/********************************************************************************/
void
calcLinkCapacityStats (topology_t *network)
{
  int	i, j;

  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("calcLinkCapacityStats: Invalid network\n");
    return;
  }
  initSimpleRatioStat (&(network->linkcapacity), DEFAULT_CONFIDENCE_LEVEL);
  for (i = 0; i < network->numnodes; i++) {
    for (j = 0; j < network->numnodes; j++) {
      if (network->links[i][j] != (link_t *) 0) {
	updateSimpleRatioStat (network->links[i][j]->capacity, &(network->linkcapacity));
      }
    }
  }
}

static void
printLinkCapacityStats (topology_t *network)
{
  simpleratio_t linkcap;

  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("printLinkCapacityStats: Invalid network\n");
    return;
  }
  linkcap = network->linkcapacity;
  
  printf ("Link capacity stats:\n");
  printf ("\tAvg. capacity:\t%.3f\n", linkcap.currentrate);
  printf ("\tVariance:\t%.3f\n", linkcap.currentvar);
  printf ("\tMax capacity:\t%.3f\n", linkcap.max);
  printf ("\tMin capacity:\t%.3f\n", linkcap.min);
}

double
getAvgLinkCapacity (topology_t *network)
{
  return (network->linkcapacity.currentrate);
}

double
getMaxLinkCapacity (topology_t *network)
{
  return (network->linkcapacity.max);
}

double 
getMinLinkCapacity (topology_t *network)
{
  return (network->linkcapacity.min);
}

double
getTotalLinkCapacity (topology_t *network)
{
  return (network->linkcapacity.numersum);
}
/********************************************************************************/
void
printLinkMatrix (topology_t *network)
/*  This routine prints the adjacency matrix.  Note that it depends on the
 *  initMatrix() command being called before the matrix is constructed.
*/ 
{
  int	i, j;

  if (network == (topology_t *) 0) {
    RS_PRINT_ERROR ("printLinkMatrix: Invalid network\n");
    return;
  }
  printf ("Adjacency Matrix:\n");
  printf ("  ");
  for (i = 0; i < network->numnodes; i++)
    printf ("%2d ", i);
  printf ("\n  ");		    /* set up table heading */
  for (i = 0; i < network->numnodes; i++)
    printf ("---");
  printf ("\n");
  for (i = 0; i < network->numnodes; i++) {
    printf ("%d|", i);
    for (j = 0; j < network->numnodes; j++) {
      if (network->links[i][j] == (link_t *) 0)
	printf (" X ");
      else
	printf (" O ");
    }
    printf ("\n");
  }
}


