/* $Id: pathcount.c,v 1.1 1999/06/11 20:06:23 ashaikh Exp $ */


/*  Description:
 *
 *  Counts number of shortest paths in an arbitrary network on a per src-dest pair
 *  basis.  Topology is read on stdin and must be in routesim format.
 *
 */

#include "sptgeneric.h"
#include "linklist.h"
#include "clinklist.h"
#include "rstraffic.h"
#include "rstrace.h"
#include "stat.h"
#include <stdlib.h>

topology_t  *network;

spt_t		**spts;	    /* source-specific shortest-path trees */
int		**pathtable;
simpleratio_t	pathstat;
simpleratio_t singlepathstat;
int printpaths = 0;



static int getOptions (int, char **);
static int computeShortestPaths (void);
static int setPathCounts (void);

static int initPathTable (void);
static void destroyPathTable (void);
static void printPathTable (void);
static void printPathCounts (void);

static int traversePaths (spt_t *, int);
static void traverseVisit (spt_t *spt, int current, int);

/********************************************************************************/
int
main (int argc, char **argv)
{

  if (getOptions(argc, argv)) {
    exit (1);
  }

  if (readTopology (&network)) {
    RS_PRINT_ERROR ("readTopology failed\n");
    exit (1);
  }

  if (initPathTable()) {
    RS_PRINT_ERROR ("constructPathTable returned error\n");
    return (-1);
  }

  if (computeShortestPaths()) {
    RS_PRINT_ERROR ("computeShortestPaths failed\n");
    return (-1);
  }

  if (setPathCounts ()) {
    RS_PRINT_ERROR ("setPathCounts returned error\n");
    return (-1);
  }

  printPathCounts ();
  
  if (printpaths)
    printPathTable ();  

  destroyPathTable();

  return (0);
}

/********************************************************************************/
static int
getOptions (int argc, char **argv)
{
  int c;
  int errflg = 0;
  
   while ((c = getopt(argc, argv, "ph")) != EOF)
   switch (c) {
   case 'p':
     printpaths = 1;
     break;
   case 'h':
     errflg++;
     break;     
   case '?':
     errflg++;
   }
   if (errflg) {
     fprintf (stderr, "usage:\t%s\t[-p]\n", "pathcount");
     return (-1);
   }

   return (0);
     
}
/********************************************************************************/
static int
computeShortestPaths (void)
/* Compute shortest-path trees based on hopcount for all nodes.  Uses the multiple-parent
 * Dijkstra modifications.
 */
{
  int	i, numnodes;


  /* initialize spt data structure */
  numnodes = getNumNodes(network);
  spts = (spt_t **) malloc (sizeof(spt_t **) * numnodes);
  for (i = 0; i < numnodes; i++) {
    if (sptInit (&(spts[i]), network)) {
      RS_PRINT_ERROR ("computeShortestPaths: SPT allocation failed\n");
      return (-1);
    }
  }

  /* compute the shortest-path trees */
  for (i = 0; i < numnodes; i++) {
    sptReset (spts[i]);
    /* compute trees for each source */
    if (sptHopMultiExec(spts[i], i)) {
      RS_PRINT_ERROR ("computeShortestPaths: sptHopMultiExec failed\n");
      return (-1);
    }
    spts[i]->source = i;
    sptMultiPathExtractInit (spts[i]);    /* reset head pointers */
  }

  return (0);
}
/********************************************************************************/
static int
setPathCounts (void)
{
  int	i, j, numnodes;

  numnodes = getNumNodes (network);
  initSimpleRatioStat (&singlepathstat, 99);
  for (i = 0; i < numnodes; i++) {
    for (j = 0; j < numnodes; j++) {
      if (i == j)
	continue;
      traversePaths (spts[i], j);
      updateSimpleRatioStat (pathtable[i][j], &pathstat);
      if (pathtable[i][j] == 1)
	updateSimpleRatioStat (getMinHops(i,j,network), &singlepathstat);      
    }
  }
  /*
   compute stats for paths 
  for (i = 0; i < numnodes; i++) {
    for (j = 0; j < numnodes; j++) {
      if (i == j)
	continue;      

    }
  }
  */
  return (0);

}

/********************************************************************************/
static int
traversePaths (spt_t *spt, int dest)
/* Performs a recursive depth-first traversal of the shortest-path graph.
 */   
{
  int	    source;
  int	    i, numnodes;


  source = spt->source;
  numnodes = getNumNodes(spt->network);
#if RS_TRACE_LEVEL >= TRACE_DETAILS  
  printf ("Traversing from %d -> %d\n", dest, source);
#endif

  /* set walkptr to headptr for all nodes */
  for (i = 0; i < getNumNodes(spt->network); i++) {
    clistIteratorAssign (spt->walkptrs[i], spt->headptrs[i]);
  }

  pathtable[source][dest] = 0;
  traverseVisit (spt, dest, dest);
  
  return (0);
}

static void
traverseVisit (spt_t *spt, int current, int finaldest)
/* Recursive function to follow parents of each node in the shortest-path
 * graph.
 */
{
  int	source, numparents;
  sptparent_t	*nextparent;

  source = spt->source;
#if RS_TRACE_LEVEL >= TRACE_DETAILS    
  printf ("at %d (%.3f)\n", current, inflow);
#endif

  if (current == source) {
    pathtable[source][finaldest]++;
    return;
  }

  /* not yet at the source */
  numparents = clistSize (spt->parents[current]);

  do {	/* visit each parent */
    nextparent = (sptparent_t *) clistIteratorGetCurrent(spt->walkptrs[current]);
    traverseVisit (spt, nextparent->parent, finaldest);
    /* advance to next parent */        
    clistIteratorSetNext(spt->walkptrs[current]);
  }
  /* while walkptr != headptr */
  while (clistIteratorGetCurrent(spt->walkptrs[current]) !=
	 clistIteratorGetCurrent(spt->headptrs[current]));
}
/********************************************************************************/
static void
printPathCounts (void)
{

  printf ("Avg. num paths:\t%.4f\n", pathstat.currentrate);
  printf ("Variance num paths:\t%.4f\n", pathstat.currentvar);
  printf ("Max num paths:\t%.0f\n", pathstat.max);
  printf ("Min num paths:\t%.0f\n", pathstat.min);
  printf ("Single Paths:\n");
  printf ("\tTotal single paths:\t%d\n", singlepathstat.denomsum);
  printf ("\tAvg. single pathlen:\t%.4f\n", singlepathstat.currentrate);
  printf ("\tSingle pathlen var.:\t%.4f\n", singlepathstat.currentvar);
  printf ("\tMax single pathlen:\t%.4f\n", singlepathstat.max);
  printf ("\tMin single pathlen:\t%.4f\n", singlepathstat.min); 
  printf ("\tSingle path proportion:\t%.4f\n", (double)singlepathstat.denomsum/(double)pathstat.denomsum);
  printf ("Total num paths:\t%d\n", pathstat.denomsum);  
  printf ("Avg. hopcount\t%.4f\n", getAvgHopcount(network));
  printf ("Avg. node degree:\t%.4f\n", getAvgDegree(network));
  printf ("Diameter:\t%d\n", getDiameter(network));
  
}
/********************************************************************************/
static int
initPathTable (void)
{
  int	i, j, numnodes;
  
  numnodes = getNumNodes(network);
  /* allocate memory for the load table */
  pathtable = (int **) malloc (sizeof(int) * numnodes);
  if (pathtable == (int **) 0) {
    RS_PRINT_SOFT_ERROR ("constructPathTable:  Table allocation failed\n");
    return (-1);
  }
  for (i = 0; i < numnodes; i++) {
    pathtable[i] = (int *) malloc (sizeof(int) * numnodes);
    if (pathtable[i] == (int *) 0) {
      RS_PRINT_SOFT_ERROR ("initPathTable:  Table row allocation failed\n");
      return (-1);
    }
    /* initialize load table */
    for (j = 0; j < numnodes; j++)
      pathtable[i][j] = 0;
  }

  initSimpleRatioStat (&pathstat, 99);

  return (0);
}

static void
destroyPathTable (void)
{
  int	i, numnodes;

  numnodes = getNumNodes(network);
  for (i = 0; i < numnodes; i++) {
    free (pathtable[i]);
    sptDestroy (spts[i]);
  }

  free (pathtable);
  pathtable = (int **) 0;
  spts = (spt_t **) 0;
}

/********************************************************************************/
static void
printPathTable (void)
{
  int	i, j, numnodes;

  numnodes = network->numnodes;  

  printf ("Number of Src-Dest Paths:\n");
  printf ("   ");
  for (i = 0; i < numnodes; i++)
    printf (" %4d ", i);
  printf ("\n  ");		    /* set up table heading */
  printf ("  ");  
  for (i = 0; i < numnodes; i++)
    printf ("------");
  printf ("\n");
  for (i = 0; i < numnodes; i++) {
    printf ("%3d|", i);
    for (j = 0; j < numnodes; j++) {
      if (pathtable[i][j] == 0) {
	printf (" %4c ", 'X');
      }
      else {
	printf (" %4d ", pathtable[i][j]);
      }
    }
    printf ("\n");
  }
}

