/* $Id: rsprovision.c,v 1.2 1999/06/11 20:12:59 ashaikh Exp $ */

static const char rcsid[] = "$Id: rsprovision.c,v 1.2 1999/06/11 20:12:59 ashaikh Exp $";

/*  Description:
 *
 *	Sets link capacities in an arbitrary network, given a traffic matrix,
 *	to ensure uniform load in the network based on shortest paths.
 *
 */

#include "rsprovision.h"
#include "sptgeneric.h"
#include "linklist.h"
#include "clinklist.h"
#include "rstraffic.h"
#include "rstrace.h"
#include <stdlib.h>

topology_t  *network;
int	    uniqpolicy = 0;

double	    avgrate, avglen, avgbw;

typedef struct {
  double    capacity;
  int	    id;
} pathstacknode_t;

spt_t	**spts;	    /* source-specific shortest-path trees */
double	**loadtable;


static int getOptions (int, char **);

static int computeShortestPaths (void);

static int initLoadTable (void);
static void destroyLoadTable (void);
#if RS_TRACE_LEVEL >= TRACE_DETAILS
static void printLoadTable (void);
#endif

static int setLinkProportions (void);
static int dimensionLinkCapacities (void);

static int traverseShortestPaths (spt_t *, int);
static int traversePaths (spt_t *, int);
static void traverseVisit (spt_t *, int, double);
static int traverseHashPaths (spt_t *, int);
static void printOfferedLoad (void);

/********************************************************************************/
int
main (int argc, char **argv)
{

  if (getOptions(argc, argv)) {
    exit (1);
  }
  
  if (createFileTopology (INPUT_TOPOLOGY_FILE, &network)) {
    RS_PRINT_ERROR ("createFileTopology failed\n");
    exit (1);
  }

  if (initLoadTable()) {
    RS_PRINT_ERROR ("constructPathTable returned error\n");
    return (-1);
  }

  if (computeShortestPaths()) {
    RS_PRINT_ERROR ("computeShortestPaths failed\n");
    return (-1);
  }


  if (setLinkProportions ()) {
    RS_PRINT_ERROR ("setLinkProportions returned error\n");
    return (-1);
  }
  


#if RS_TRACE_LEVEL >= TRACE_DETAILS
  printLoadTable ();  
#endif

  if (dimensionLinkCapacities()) {
    RS_PRINT_ERROR ("dimensionLinkCapacities failed\n");
    return (-1);
  }
  calcLinkCapacityStats(network);

  dumpTopologyFile (OUTPUT_TOPOLOGY_FILE, network);
  printOfferedLoad ();

  destroyLoadTable();

  return (0);
}

/********************************************************************************/
static int
getOptions (int argc, char **argv)
{
  int c;
  extern char *optarg;
  extern int optind;
  int errflg = 0;
  
   while ((c = getopt(argc, argv, "a:l:b:p:")) != EOF)
   switch (c) {
   case 'a':
     avgrate = strtod (optarg, (char **) NULL);
     break;
   case 'l':
     avglen = strtod (optarg, (char **) NULL);
     break;
   case 'b':
     avgbw = strtod (optarg, (char **) NULL);
     break;     
   case 'p':
     uniqpolicy = (int) strtol (optarg, (char **) NULL, 10);
     break;     
   case '?':
     errflg++;
   }
   if (errflg || (optind != 9)) {
     fprintf (stderr, "usage:\t%s\t-a <avg arrival rate>\n\t\t\t-l <avg duration>\n\t\t\t-b <avg bandwidth>\n\t\t\t-p <route policy>\n\troute policies: 0 - RANDOM, 1 - SRCHASH\n\ttopology file: %s\n", "rsprovision", INPUT_TOPOLOGY_FILE);
     return (-1);
   }

   return (0);
     
}
/********************************************************************************/
static int
computeShortestPaths (void)
/* Compute shortest-path trees based on hopcount for all nodes.  Uses the multiple-parent
 * Dijkstra modifications.
 */
{
  int	i, numnodes;


  /* initialize spt data structure */
  numnodes = getNumNodes(network);
  spts = (spt_t **) malloc (sizeof(spt_t **) * numnodes);
  for (i = 0; i < numnodes; i++) {
    if (sptInit (&(spts[i]), network)) {
      RS_PRINT_ERROR ("computeShortestPaths: SPT allocation failed\n");
      return (-1);
    }
  }

  /* compute the shortest-path trees */
  for (i = 0; i < numnodes; i++) {
    sptReset (spts[i]);
    /* compute trees for each source */
    if (sptHopMultiExec(spts[i], i)) {
      RS_PRINT_ERROR ("computeShortestPaths: sptHopMultiExec failed\n");
      return (-1);
    }
    spts[i]->source = i;
    sptMultiPathExtractInit (spts[i]);    /* reset head pointers */
  }

  return (0);
}
/********************************************************************************/
static int
setLinkProportions (void)
{
  int	i, j, numnodes;

  numnodes = getNumNodes (network);
  switch (uniqpolicy) {
  case 0:   /* RANDOM */
    for (i = 0; i < numnodes; i++) {
      for (j = 0; j < numnodes; j++) {
	if (i == j)
	  continue;
	traversePaths (spts[i], j);
      }
    }
    break;
  case 1: /* SRCHASH */
    for (i = 0; i < numnodes; i++) {
      for (j = 0; j < numnodes; j++) {
	if (i == j)
	  continue;
	traverseHashPaths (spts[i], j);
      }
    }
    break;
  default:
    RS_PRINT_ALWAYS ("Unknown uniqpolicy\n");
    return (-1);
  }

  return (0);

}

/********************************************************************************/
static int
dimensionLinkCapacities (void)
/* Basically copies the loadtable entries into link matrix.
 */
{
  int	    i, j, numnodes;

  numnodes = getNumNodes(network);
  for (i = 0; i < numnodes; i++) {
    for (j = 0; j < numnodes; j++) {
      if ( (i == j) || (network->links[i][j] == (link_t *) 0) )
	continue;
      network->links[i][j]->capacity = loadtable[i][j] * avgrate * avglen;
    }
  }

  return (0);
}
/********************************************************************************/
static int
traverseShortestPaths (spt_t *spt, int dest)
/* A non-recursive routine to traverse the shortest-path graph.
 */
{
  int		    source;
  linklist_t	    *pathstack;
  sptparent_t	    *nextparent;
  pathstacknode_t   *node, *popped;
  int		    top, nextid;

  source = spt->source;

  pathstack = listInit();
  node = (pathstacknode_t *) malloc (sizeof(pathstacknode_t));
  node->id = dest;
  node->capacity = 1.0;
  listPush (pathstack, (void *) node);	    /* start at dest node */
  printf ("pushing %d\n", node->id);
  clistIteratorAssign (spt->walkptrs[dest], spt->headptrs[dest]);  
  top = dest;
  
  while (!listEmpty(pathstack)) {
    while ( ((top = ((pathstacknode_t *)listGetTop(pathstack))->id) != source) ) {
      /* explore parents of top node */
      nextparent = (sptparent_t *) clistIteratorGetCurrent(spt->walkptrs[top]);
      nextid = nextparent->parent;
      /* allocate a new node */
      node = (pathstacknode_t *) malloc (sizeof(pathstacknode_t));
      node->id = nextid;
      node->capacity = 1.0 / (clistSize (spt->parents[top]));
      listPush(pathstack, (void *) node);	    /* explore next node in route */
      printf ("pushing %d\n", node->id);
      clistIteratorAssign (spt->walkptrs[nextid], spt->headptrs[nextid]);  /* start at first parent */
    }
    /* reached the source -- pop back to last branch point */
    listPop (pathstack, (void **) &popped);
    printf ("popped %d\n", popped->id);    
    free (popped);
    if (!listEmpty(pathstack)) {
      top = ((pathstacknode_t *)listGetTop(pathstack))->id;  /* reset top */
      while (clistIteratorGetNext(spt->walkptrs[top]) ==
	     clistIteratorGetCurrent(spt->headptrs[top])) {
	listPop (pathstack, (void **) &popped);	/* backtrack from exhausted parent list */
	free (popped);
	if (!listEmpty(pathstack))
	  top = ((pathstacknode_t *)listGetTop(pathstack))->id;  /* reset top */
	else
	  break;	
      }
      /* reached an unexplored branch point -- or stack went empty */
      if (!listEmpty(pathstack)) {
	clistIteratorSetNext(spt->walkptrs[top]);   /* advance to next parent */
      }
      else break;	/* exhausted all parents */
    }
  }

  return (0);
}
/********************************************************************************/
static int
traversePaths (spt_t *spt, int dest)
/* Performs a recursive depth-first traversal of the shortest-path graph.
 */   
{
  int	    source;
  int	    i, numnodes;


  source = spt->source;
  numnodes = getNumNodes(spt->network);
#if RS_TRACE_LEVEL >= TRACE_DETAILS  
  printf ("Traversing from %d -> %d\n", dest, source);
#endif

  /* set walkptr to headptr for all nodes */
  for (i = 0; i < getNumNodes(spt->network); i++) {
    clistIteratorAssign (spt->walkptrs[i], spt->headptrs[i]);
  }

  traverseVisit (spt, dest, 1.0/(float)(numnodes - 1));
  
  return (0);
}

static void
traverseVisit (spt_t *spt, int current, double inflow)
/* Recursive function to follow parents of each node in the shortest-path
 * graph.  Adds to the loadtable entries along the way.
 */
{
  int	source, numparents;
  sptparent_t	*nextparent;

  source = spt->source;
#if RS_TRACE_LEVEL >= TRACE_DETAILS    
  printf ("at %d (%.3f)\n", current, inflow);
#endif

  if (current == source)
      return;

  /* not yet at the source */
  numparents = clistSize (spt->parents[current]);

  do {	/* visit each parent */
    nextparent = (sptparent_t *) clistIteratorGetCurrent(spt->walkptrs[current]);
    loadtable[current][nextparent->parent] += (inflow * 1.0/(double)numparents);
    traverseVisit (spt, nextparent->parent, (inflow * 1.0/(double)numparents));
    /* advance to next parent */        
    clistIteratorSetNext(spt->walkptrs[current]);
  }
  /* while walkptr != headptr */
  while (clistIteratorGetCurrent(spt->walkptrs[current]) !=
	 clistIteratorGetCurrent(spt->headptrs[current]));
}
/********************************************************************************/
static int
traverseHashPaths (spt_t *spt, int dest)
{
  int		source, current;
  int		numparents, chosen, next;
  sptparent_t	*parentptr;

  source = spt->source;
  if (dest == source)
    return (0);

  current = dest;
  while (current != source) {
    /* how many parents does the current node have? */
    numparents = clistSize (spt->parents[current]);
    /* pick the next node to visit */
    if (numparents > 1) {
      chosen = source % numparents; /* a simple hash function based on source id */
      chosen++;   /* because parents are indexed starting at 1, not 0 */
      clistGetElement (spt->parents[current], chosen, (void **) &parentptr);
    }
    else {
      clistGetFirstElement (spt->parents[current], (void **) &parentptr);
    }
    next = parentptr->parent;
    loadtable[current][next] += 1.0;
    current = next;
  }

  return (0);

}
    
/********************************************************************************/
static void
printOfferedLoad (void)
{
  double    offeredload;
  double    avghops, avglinkcap, totalcap;
  double    numnodes, numlinks;

  avghops = getAvgHopcount (network);
  avglinkcap = getAvgLinkCapacity (network);
  totalcap = getTotalLinkCapacity (network);
  numnodes = (double)getNumNodes (network);
  numlinks = (double)getNumLinks (network);

  offeredload = avgrate * avglen * avgbw / avglinkcap;
  offeredload *= (numnodes * avghops / numlinks);

  printf ("Total offered load:\t%.4f\n", offeredload);
  printf ("Total link capacity):\t%.4f\n", totalcap);
  printf ("Avg link capacity):\t%.4f\n", avglinkcap);
  
}
/********************************************************************************/
static int
initLoadTable (void)
{
  int	i, j, numnodes;
  
  numnodes = getNumNodes(network);
  /* allocate memory for the load table */
  loadtable = (double **) malloc (sizeof(double) * numnodes);
  if (loadtable == (double **) 0) {
    RS_PRINT_SOFT_ERROR ("constructLoadTable:  Table allocation failed\n");
    return (-1);
  }
  for (i = 0; i < numnodes; i++) {
    loadtable[i] = (double *) malloc (sizeof(double) * numnodes);
    if (loadtable[i] == (double *) 0) {
      RS_PRINT_SOFT_ERROR ("constructLoadTable:  Table row allocation failed\n");
      return (-1);
    }
    /* initialize load table */
    for (j = 0; j < numnodes; j++)
      loadtable[i][j] = 0.0;
  }

  return (0);
}

static void
destroyLoadTable (void)
{
  int	i, numnodes;

  numnodes = getNumNodes(network);
  for (i = 0; i < numnodes; i++) {
    free (loadtable[i]);
    sptDestroy (spts[i]);
  }

  free (loadtable);
  loadtable = (double **) 0;
  spts = (spt_t **) 0;
}

/********************************************************************************/
#if RS_TRACE_LEVEL >= TRACE_DETAILS
static void
printLoadTable (void)
{
  int	i, j, numnodes;

  numnodes = network->numnodes;  

  printf ("Relative Load Matrix:\n");
  printf ("   ");
  for (i = 0; i < numnodes; i++)
    printf (" %4d ", i);
  printf ("\n  ");		    /* set up table heading */
  printf ("  ");  
  for (i = 0; i < numnodes; i++)
    printf ("------");
  printf ("\n");
  for (i = 0; i < numnodes; i++) {
    printf ("%3d|", i);
    for (j = 0; j < numnodes; j++) {
      if (!areConnected(i,j,network)) {
	printf (" %4c ", 'X');
      }
      else {
	printf (" %.2f ", loadtable[i][j]);
      }
    }
    printf ("\n");
  }
}
#endif
