/* $Id: priorityQ.c,v 1.7 1998/01/05 20:25:19 ashaikh Exp $ */
/*
 *
 *   Description:
 *      
 *	Implementation of a heap-based priority queue.  Elements are stored
 *	in an array and defined in priorityQ.h.  This implementation is designed
 *	for use with elements with floating point keys, sorted as smallest having
 *	highest priority.
 *
 *	Basic queue operations:
 *
 *	    - add item
 *	    - remove item of highest priority
 *
 */

#include "priorityQ.h"
#include <stdio.h>
#include <stdlib.h>

#define left(x)	    ((x)*2)				/* return left child */
#define right(x)    ((x)*2 + 1)				/* return right child */
#define parent(x)   ((x)/2)				/* return parent */

#define isLeaf(x)   (left((x)) > num ? 1 : 0)   /* is this a leaf? */
#define isNull(x)   ((x) > num ? 1 : 0)		/* is this within the queue */

#define val(x)	    ((x)->key)				/* extract key value */

static int fixUp (pq_t *q, int start);
static int fixDown (pq_t *q, int start);
static int getSmallest (pq_t *q, int parent);

/*******************************************************************************/
pq_t
*pqInit (int maxqsize)
{
  pq_t	    *q;
  int	    i;
  pqelement_t	*null;

  q = (pq_t *) malloc (sizeof(pq_t));
  if (q == (pq_t *)0)
    return ( (pq_t *)0 );
  q->numitems = 0;
  q->qsize = maxqsize+1;
  q->Q = (pqelement_t **) malloc (sizeof(pqelement_t *)*(q->qsize));
  if (q->Q == (pqelement_t **) 0)
    return ( (pq_t *) 0 );
  /* initialize element pointers to NULL */
  for (i = 0; i < q->qsize; i++) {
    q->Q[i] = (pqelement_t *) 0;
  }

  /* make a dummy element at the head of the heap */
  null = (pqelement_t *) malloc (sizeof(pqelement_t));
  if (null == (pqelement_t *) 0)
    return ( (pq_t *) 0 );
  null->element = 0;
  q->Q[0] = null;
  return (q);
}
/*******************************************************************************/
int
pqDestroy (pq_t *q)
{
  if (q == (pq_t *) 0)
    return (-1);
  /* caller is responsible for destroying individual elements */
  
  /* free null element */
  free (q->Q[0]);
  free ((q)->Q);
  free (q);
  return (0);
}
/*******************************************************************************/
int
pqAddItem (pq_t *q, pqelement_t *newitem)
{

  if ((newitem == (pqelement_t *) 0) || (q == (pq_t *) 0))
    return (-1);
  /* is this element going to fit */
  if ((q->numitems + 1) > q->qsize)
    return (-1);
  q->Q[++(q->numitems)] = newitem;
  return (fixUp (q, (q->numitems)));
}

/*******************************************************************************/
static int
fixUp (pq_t *q, int start)
{
  pqelement_t	*v;
  int	k = start;

  v = q->Q[k];
  q->Q[0]->key = -1;
  while (val(q->Q[parent(k)]) >= val(v)) {
    q->Q[k] = q->Q[parent(k)];
    k = parent(k);
  }
  q->Q[k] = v;
  return(0);
}

/*******************************************************************************/
int
pqRemoveItem (pq_t *q, pqelement_t **element)
{
  pqelement_t	    *retval;

  if (q == (pq_t *) 0)
    return (-1);

  *element = (pqelement_t *) 0;
  retval = q->Q[1];
  q->Q[1] = q->Q[(q->numitems)--];
  if (fixDown (q, 1))
    return(-1);
  *element = retval;
  return (0);
}
/*******************************************************************************/
double
pqCheckKey (pq_t *q)
{
  if (pqIsQEmpty(q))
    return ((double) -1.0);	/* a very large key */
  else
    return (q->Q[1]->key);
}

/*******************************************************************************/
static int
fixDown (pq_t *q, int start)
{
  int	j;
  int	smallest;
  int	done;
  int	num;
  pqelement_t	*temp;

  j = start;
  num = q->numitems;
  done =  isLeaf(j);
  while (!done) {
    smallest = getSmallest(q, j);
    if ( smallest != j ) {		    /* need to swap */
      temp = q->Q[j];
      q->Q[j] = q->Q[smallest];
      q->Q[smallest] = temp;
      j = smallest;
      done = isLeaf(j);
    }
    else
      done = 1;			    /* we're done early */
  }
  return (0);
}

/*******************************************************************************/
int
pqChangeVal (pq_t *q, int target, double newkey)
{
  int	i;
  int	pos;
  int	found = 0;

  if (q == (pq_t *) 0)
    return (-1);

  i = 1;
  while (!found) {
    if (q->Q[i]->id == target) {
      pos = i;
      found = 1;
    }
    else
      i++;
  }
  if (i > q->numitems)
    return (-1);		    /* make sure we did find target in the queue */
  if (q->Q[pos]->key > newkey) {    /* lowering key value */
    q->Q[pos]->key = newkey;
    if (fixUp (q, pos))
      return (-1);
  }
  else  {			/* raising key value */
    q->Q[pos]->key = newkey;
    if (fixDown (q, pos))
      return (-1);
  }
  return (0);
}
/*******************************************************************************/  
static int
getSmallest (pq_t *q, int parent)
{
  int	smallest;
  int	num;

  
  num = q->numitems;
  if ( (!isNull(left(parent))) && (val(q->Q[left(parent)]) < val(q->Q[parent])) )
    smallest = left(parent);
  else 
    smallest = parent;
  if (!isNull(right(parent)) && val(q->Q[right(parent)]) < val(q->Q[smallest]))
    smallest = right(parent);
  return (smallest);
}

/*******************************************************************************/
int
pqIsQEmpty (pq_t *q)
{
  if (q == (pq_t *) 0)
    return (1);	/* say it's empty */

  return (q->numitems == 0);
}
/*******************************************************************************/
void
pqPrintQ (pq_t *q)
{
  int	i;

  if (q == (pq_t *) 0)
    return;

  printf ("Q: ");
  for (i = 1; i <= q->numitems; i++)
    printf ("[%d] %d-%.1f ", i, q->Q[i]->id, q->Q[i]->key);
  printf ("\n");
}
