/* $Id: sptgeneric.c,v 1.2 1999/05/02 21:17:32 ashaikh Exp $ */

static const char rcsid[] = "$Id: sptgeneric.c,v 1.2 1999/05/02 21:17:32 ashaikh Exp $";

/*
*   Description:
*
*	Generic Dijkstra single-source shortest paths algorithms for
*	the construction of shortest-path trees.
*/

#include "sptgeneric.h"
#include "priorityQ.h"
#include "clinklist.h"
#include "linklist.h"
#include "topology.h"
#include "rstrace.h"
#include <stdlib.h>

/********************************************************************************/
int
sptInit (spt_t **newspt, topology_t *net)
/* Initialize an spt data structure.  This should be called once at startup time to
 * allocate memory for the structure.
 */
{
  int	    i, numnodes;
  spt_t	    *spt;

  if (net == (topology_t *) 0) {
    RS_PRINT_ERROR ("sptInit: Invalid network\n");
    return (-1);
  }
  
  spt = (spt_t *) malloc (sizeof(spt_t));
  if (spt == (spt_t *) 0) {
    RS_PRINT_SOFT_ERROR ("sptInit: SPT allocation failed\n");
    return (-1);
  }

  numnodes = getNumNodes(net);
  /* allocate and initialize parent pointer lists */
  spt->parents = (clinklist_t **) malloc (sizeof(clinklist_t **)*numnodes);
  if (spt->parents == (clinklist_t **) 0) {
    RS_PRINT_SOFT_ERROR ("sptInit: SPT parent list allocation failed\n");
    return (-1);
  }

  spt->headptrs = (clistiterator_t **) malloc (sizeof(clistiterator_t **)*numnodes);
  spt->walkptrs = (clistiterator_t **) malloc (sizeof(clistiterator_t **)*numnodes);
  if ( (spt->headptrs == (clistiterator_t **) 0) || (spt->walkptrs == (clistiterator_t **) 0) ) {
    RS_PRINT_SOFT_ERROR ("sptInit: SPT parent iterator allocation failed\n");
    return (-1);
  }

  for (i = 0; i < numnodes; i++) {
    spt->parents[i] = clistInit();
    if (spt->parents[i] != (clinklist_t *) 0) {
      spt->headptrs[i] = clistIteratorInit(spt->parents[i]);
      spt->walkptrs[i] = clistIteratorInit(spt->parents[i]);
      if ( (spt->headptrs[i] == (clistiterator_t *) 0) ||
	   (spt->walkptrs[i] == (clistiterator_t *) 0) ) {
	RS_PRINT_SOFT_ERROR ("SPT allocation failed\n");
	return (-1);
      }
    }
    else {
      RS_PRINT_SOFT_ERROR ("SPT allocation failed\n");
      return (-1);
    }
  }

  spt->network = net;
  spt->valid = 0;   /* initialize valid flag to zero */
  spt->comptime = -1.0;
  spt->computations = 0;
    
 *newspt = spt;
  return (0);
}
/********************************************************************************/
int
sptHopMultiExec (spt_t *spt, int source)
/* Similar to sptExec() but tree calculation is based only on hopcount.
 */
{
  pqelement_t    *popped, *newel;
  int		current;
  int		neighbor;
  int		dist;
  int		i, numnodes;
  int		*hops;	    /* distance estimate for each node */
  pq_t		*sptQ;	    /* priority Q for network nodes */

  if (spt == (spt_t *) 0) {
    RS_PRINT_ERROR ("sptHopMultiExec:  Invalid spt pointer\n");
    return (-1);
  }

  numnodes = spt->network->numnodes;
  /* initialize the shortest-path tree data structure */
  spt->source = source;
  spt->valid = 0;
  sptQ = pqInit(numnodes);
  if (sptQ == (pq_t *) 0) {
    RS_PRINT_SOFT_ERROR ("sptHopMultiExec: SPT pq allocation failure\n");
    return (-1);
  }

  /* allocate array for distance estimates */
  hops = (int *) malloc (sizeof(int)*numnodes);
  if (hops == (int *) 0) {
    RS_PRINT_SOFT_ERROR ("sptHopMultiExec: SPT key array allocation failure\n");
    return (-1);
  }
  else {
    for (i = 0; i < numnodes; i++)
      hops[i] = SPTINF;
  }

  /* set estimate of source to 0 */
  hops[source] = 0;

  /* put all the nodes in a heap */
  /* this can be optimized to avoid putting nodes in until they are scanned */
  for (i=0; i < numnodes; i++) {
    newel = (pqelement_t *) malloc (sizeof(pqelement_t));
    newel->id = i;
    newel->key = hops[i];
    newel->element = (void *) 0;
    pqAddItem(sptQ, newel);
  }

  while (!pqIsQEmpty(sptQ)) {
    if (pqRemoveItem (sptQ, &popped)) {	    /* examine node with shortest dist. estimate */
      RS_PRINT_SOFT_ERROR ("sptHopMultiExec: pqRemoveItem failed\n");
      sptCleanup (sptQ);
      free (hops);
      return (-1);
    }
    current = popped->id;
    free (popped);

    /* check for connectivity */
    if (hops[current] == SPTINF) {
      /* If tree is disconnected we call it an error */
#if RS_TRACE_LEVEL >= TRACE_SOFT_ERRORS
      printf ("sptHopMultiExec: %d is disconnected from source %d\n", current, source);
#endif
      sptCleanup (sptQ);
      free (hops);
      return(-1);
    }

    neighbor = -1;
    for (i = 0; i < numnodes; i++) {	/* check all neighboring nodes */
      if (spt->network->links[current][i] != (link_t *) 0 &&
	  spt->network->links[current][i]->active != 0) {
	/* it's a neighbor */
	neighbor = i;
	dist = (int) spt->network->links[current][neighbor]->hopcount.current;
	if (hops[neighbor] > (hops[current] + dist)) {	 /* need to update */
	  /* first change the value in the heap */
	  if (pqChangeVal (sptQ, neighbor, (hops[current] + dist))) {
	    RS_PRINT_SOFT_ERROR ("sptHopMultiExec: pqChangeVal failed\n");
	    sptCleanup (sptQ);
	    free (hops);
	    return (-1);
	  }
	  /* change the value in the key array */
	  hops[neighbor] = hops[current] + dist;
	  /* get rid of any existing parents list */
	  if (clearParents (spt, neighbor)) {
	    RS_PRINT_SOFT_ERROR ("sptHopMultiExec: clearParent failed\n");
	    return (-1);
	  }
	  /* add the new parent ptr */
	  if (addParent (spt, neighbor, current)) {
	    RS_PRINT_SOFT_ERROR ("sptHopMultiExec: addParent failed\n");
	    return (-1);
	  }
	}
	else {
	  /* equal cost -- add to parent list */
	  if (hops[neighbor] == (hops[current] + dist)) { 
	    if (addParent (spt, neighbor, current)) {
	      RS_PRINT_SOFT_ERROR ("sptHopMultiExec: addParent failed\n");
	      return (-1);
	    }
	  }
	}
      } /* if (link leads to a neighbor */
    } /* for (each possible neighbor */
  } /* while (!pqIsQEmpty) */
    
  spt->valid = 1;
  spt->computations++;  
  sptCleanup (sptQ);
  free (hops);
  
  return (0);
}
/********************************************************************************/
int
addParent (spt_t *spt, int child, int parent)
{
  sptparent_t	*newparent;

  newparent = (sptparent_t *) malloc (sizeof(sptparent_t));

  if (newparent == (sptparent_t *) 0)
    return (-1);

  newparent->parent = parent;
  /*  newparent->scanned = 0; */

  if (clistAttach (spt->parents[child], (void *) newparent)) {
    free (newparent);
    return (-1);
  }
  return (0);
}

int
clearParents (spt_t *spt, int child)
{
  int	i, numparents;
  sptparent_t  *parentptr;

  numparents = clistSize (spt->parents[child]);

  if (numparents < 1) {
    return (0);	/* no parents in list */
  }
  else {
    for (i = 1; i <= numparents; i++) {
      /* remove each element of the parents list */
      if (clistPop (spt->parents[child], (void **) &parentptr)) {
#if RS_TRACE_LEVEL >= TRACE_SOFT_ERRORS
	fprintf (stderr, "clearParents: underflow in parents[%d] list\n", child);
#endif
	return (-1);
      }
      free (parentptr);
    }
  }
  return (0);
}
/********************************************************************************/
int
sptIsNodeConnected (spt_t *spt, int dest)
{
  return ((clistSize(spt->parents[dest]) > 0));
}
/********************************************************************************/

int
calcSimpleSPT (topology_t *network, int source, int *path)
{
  int		*hops;		/* array for distance keys */
  pq_t		*q;		/* priority queue for nodes */
  pqelement_t	*popped, *newel;
  int		current;
  int		neighbor;
  int		i, numnodes;

  numnodes = network->numnodes;
  /*	Algorithm initialization    */
  q = pqInit(numnodes);		/* initialize priority queue */
  hops = (int *) malloc (sizeof(int) * numnodes);
  for (i = 0; i < numnodes; i++) {
    hops[i] = SPTINF;
    path[i] = -1;
  }
  hops[source] = 0;
  for (i = 0; i < numnodes; i++) {
    /* put all nodes in the priority queue */    
    newel = (pqelement_t *) malloc (sizeof(pqelement_t));
    newel->id = i;
    newel->key = hops[i];
    newel->element = 0;
    if (pqAddItem(q, newel))
      return (-1);
  }    

  /*	Algorithm operation	    */

  while (!pqIsQEmpty(q)) {
    if (pqRemoveItem (q, &popped))
      return (-1);
    current = popped->id;
    for (i = 0; i < numnodes; i++)
      if (network->links[current][i] != (link_t *) 0 &&
	  network->links[current][i]->active != 0) {
	/* it's a neighboring node */
	neighbor = i;
	if (hops[neighbor] > (hops[current] +
			      network->links[current][neighbor]->hopcount.current) ) {
	  /* need to relax this edge */
	  if (pqChangeVal (q, neighbor,
		       (hops[current] + network->links[current][neighbor]->hopcount.current))) {
	    return (-1);
	  }
	  /* reset key and parent pointer for this neighbor */
	  hops[neighbor] = (hops[current] + (int)network->links[current][neighbor]->hopcount.current);
	  path[neighbor] = current;
	}
      }
    free (popped);
  }

  /* done -- free up the memory */
  free(hops);
  if (pqDestroy(q))
    return (-1);

  return (0);
}
/********************************************************************************/
int
sptMultiPathExtractInit (spt_t *spt)
/* The spt data struct must be initialized and an SPT must have been calculated
 * when this function is called.  It's possible that some node has no parents, in
 * which case we just skip setting its head ptr.
 */
{
  int	i;

  if ( (spt == (spt_t *) 0) || (spt->valid == 0) ) {
    RS_PRINT_ERROR ("sptMultiPathExtractInit: spt is uninitialized\n");
    return (-1);
  }

  for (i = 0; i < getNumNodes(spt->network); i++)
    if (clistIteratorSetFirst (spt->headptrs[i]))   /* skip for NULL lists */
      continue;

  return (0);

}
/********************************************************************************/
int
sptExtractFirstPath (spt_t *spt, int source, int dest, int **path)
/* Extracts the first path, disregarding multiple parent pointers and link
 * feasibility
 */
{
  int		*treepath;
  int		pathlen;
  int		current, i;
  sptparent_t	*parentptr;

  pathlen = sptGetPathLen (spt, source, dest);
  if (pathlen <= 0) {
    RS_PRINT_SOFT_ERROR ("sptExtractFirstPath:  sptGetPathLen failed\n");
    return (-1);
  }
  treepath = (int *) malloc (sizeof(int) * pathlen);
  
  current = dest;
  i = pathlen - 1;
  while (current != source) {
    treepath[i--] = current;    /* add current node to the path */
    /* now grab first parent of current node */
    if (clistGetFirstElement (spt->parents[current], (void **) &parentptr))
      return (-1);
    /* go on to parent */
    current = parentptr->parent;
  }
  treepath[i] = current;

  *path = treepath;
  return (0);
}
/********************************************************************************/      
int
sptGetPathLen (spt_t *spt, int source, int dest)
/*  Traces a single path from dest to source and returns the number of nodes
 *  encountered.
 */
{
  int	    current, nodes;
  sptparent_t  *parentptr;

  nodes = 1; /* to include source node */
  current = dest;
  while (current != source) {
    nodes++;
    /* just get the first element in the parentlist */
    if (clistGetFirstElement (spt->parents[current], (void **) &parentptr)) {
#if RS_TRACE_LEVEL >= TRACE_ERRORS
      fprintf (stderr, "sptGetPathLen: %d has no parents\n", current);
#endif
      return (-1);
    }
    current = parentptr->parent;
  }

  return (nodes);
}
/********************************************************************************/
void
sptCleanup  (pq_t *sptQ)
/* Deallocate just the priority queue data structures that is particular to a given
 * SPT computation.  Leaves the parent pointer structure intact.
 */
{
  pqelement_t	*popped;

  if (pqIsQEmpty(sptQ)) {
    pqDestroy(sptQ);
  }
  else {
    while (!pqIsQEmpty(sptQ)) {
      pqRemoveItem (sptQ, &popped);
      free (popped);
    }
    pqDestroy(sptQ);    
  }
}


int
sptDestroy (spt_t *spt)
/*  Completely destroys SPT, including parent pointer structure
 */
{
  int	    i, j, numnodes;
  sptparent_t  *parentptr;

  if (spt == (spt_t *) 0) {
    RS_PRINT_ERROR ("sptDestroy:  Invalid spt pointer\n");
    return (-1);
  }

  numnodes = getNumNodes(spt->network);
  for (i = 0; i < numnodes; i++) {
    clistIteratorDestroy(spt->headptrs[i]);
    clistIteratorDestroy(spt->walkptrs[i]);    
    for (j = 1; j <= clistSize(spt->parents[i]); j++) {
      clistGetElement (spt->parents[i], j, (void **) &parentptr);
      free (parentptr);
    }
    clistDestroy (spt->parents[i]);
  }
  free(spt->parents);
  free (spt->headptrs);
  free (spt->walkptrs);

  free(spt);

  return (0);
}

void
sptReset (spt_t *spt)
/* Clears parent pointer structure and invalidates SPT.  All other data items
 * are left intact.
 */
{
  int	    i, j, numnodes;
  int	    numparents;
  sptparent_t  *parentptr;

  numnodes = getNumNodes (spt->network);
  for (i = 0; i < numnodes; i++) {
    clistIteratorReset(spt->headptrs[i]);   /* reset head and walking ptrs */
    clistIteratorReset(spt->walkptrs[i]);
    if ( (numparents = clistSize(spt->parents[i])) < 1 )
      continue;	/* no parents */
    else {
      for (j = 1; j <= numparents; j++) {	/* empty the parent list */
	clistPop (spt->parents[i], (void **) &parentptr);
	free (parentptr);
      }
    }
  }
  spt->valid = 0;
}
/********************************************************************************/
int
sptGetSpecPathLen (int source, int dest, int *tree)
/* Almost the same as sptGetPathLen except this routine
 * counts *hops* as opposed to nodes, and does it for a
 * specified SPT that's passed in the form of an array
 * of parent pointers (ints)
 */

{
  int current, hops;

  hops = 0;
  current = dest;
  while (current != source) {
    hops++;
    current = tree[current];
  }
  return (hops);
}

/********************************************************************************/
void
sptPrintTree (spt_t *spt)
/*  Prints the shortest path tree generated by the algorithm, including all the parents
 *  at each node.
 */ 
{
  int		i, j, numnodes;
  int		size;
  sptparent_t	*pptr;

  if ( (spt->source < 0) || (spt->source > (numnodes - 1)) )
    return;
  numnodes = getNumNodes (spt->network);
  printf ("\nShortest path tree to src %d\n", spt->source);
  for (i = 0; i < numnodes; i++) {
    size = clistSize(spt->parents[i]);
    printf ("parent[%d] (%d): ", i, size);
    for (j = 1; j <= size; j++) {
      if (clistGetElement (spt->parents[i], j, (void **) &pptr)) {
	RS_PRINT_SOFT_ERROR ("listGetElement failed\n");
	return;
      }
      printf ("%d ", pptr->parent);
    }
    printf ("\n");
  }
}

void
sptPrintPathLens (spt_t *spt, int source)
{
  int	i, len, numnodes;

  numnodes = getNumNodes (spt->network);
  printf ("Path lengths (# nodes) from source %d to :\n", source);
  for (i = 0; i < numnodes; i++) {
    len = sptGetPathLen(spt, source, i);
    printf ("\t%d: %d\n", i, len);
  }
}


