/* $Id: stat.c,v 1.3 1998/12/30 19:58:02 ashaikh Exp $ */

static const char rcsid[] = "$Id: stat.c,v 1.3 1998/12/30 19:58:02 ashaikh Exp $";

#include "stat.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>


static int confInterval (double, double, double, confint_t *);
static double tval(double, int);
static double SampleStatistic (double, double, double);


/********************************************************************************/

int
initSimpleRatioStat (simpleratio_t *stat, double conflevel)
{
  if (stat == (simpleratio_t *) 0)
    return (-1);
  stat->numersum = 0;
  stat->numersumsq = 0;
  stat->denomsum = 0;
  stat->currentrate = 0;
  stat->currentvar = 0;
  stat->min = (DBL_MAX);
  stat->max = -(DBL_MAX);    
  stat->conf.ub = 0;
  stat->conf.lb = 0;
  stat->conf.level = conflevel;

  return (0);
}
/********************************************************************************/
int
copySimpleRatioStat (simpleratio_t *from, simpleratio_t *to)
{
  if ( (from == (simpleratio_t *) 0) ||
       (to == (simpleratio_t *) 0) ) {
    return (-1);
  }
  to->numersum = from->numersum;
  to->numersumsq = from->numersumsq;
  to->denomsum = from->denomsum;
  to->currentrate = from->currentrate;
  to->currentvar = from->currentvar;
  to->min = from->min;
  to->max = from->max;
  to->conf.ub = from->conf.ub;
  to->conf.lb = from->conf.lb;
  to->conf.level = from->conf.level;

  return (0);
}
  
int
updateSimpleRatioStat (double newval, simpleratio_t *stat)
{

  if (stat == (simpleratio_t *) 0)
    return (-1);
  stat->numersum += newval;
  stat->numersumsq += pow(newval,2);
  stat->denomsum++;
  stat->currentrate = stat->numersum / (double)stat->denomsum;
  stat->currentvar = (stat->numersumsq - (pow(stat->numersum,2)/(double)stat->denomsum))/(double)(stat->denomsum - 1);
  if (newval > stat->max)
    stat->max = newval;
  if (newval < stat->min)
    stat->min = newval;
  if (confInterval ((double)stat->denomsum, stat->currentrate, sqrt(stat->currentvar), &stat->conf))
    return (-1);
  

  return (0);
}

/********************************************************************************/
int
checkSimpleRatioStat (simpleratio_t *stat, double range)
{
  double    lower;
  double    upper;

  lower = (1.0 - range) * stat->currentrate;
  upper = (1.0 + range) * stat->currentrate;
  if (lower <= stat->conf.lb && upper >= stat->conf.ub) 
    return (1);
  else
    return (0);
  
}
/********************************************************************************/
void
dumpSimpleRatioStat (simpleratio_t *stat)
{
  printf ("numerator:\t%.2f\n", stat->numersum);
  printf ("denominator:\t%d\n", stat->denomsum);   
  printf ("numerator sq:\t%.2f\n", stat->numersumsq);
  printf ("current rate:\t%.3f\n", stat->currentrate);
  printf ("current var:\t%.3f\n", stat->currentvar);
  printf ("current max:\t%.3f\n", stat->max);
  printf ("current min:\t%.3f\n", stat->min);
  printf ("conf interval:\t[%.3f, %.3f]\n", stat->conf.lb, stat->conf.ub);
  fflush (stdout);
}
/********************************************************************************/
int
initProportionStat (proportionstat_t *stat, double conflevel)
{
  if (stat == (proportionstat_t *) 0)
    return (-1);
  stat->numersum = 0.0;
  stat->totaldenom = 0.0;
  stat->proportion = 0.0;
  stat->variance = 0.0;
  stat->conf.level = conflevel;  
  stat->conf.lb = 0;
  stat->conf.ub = 0;
  return (0);
}

int
updateProportionStat (double newval, double totaldenom, proportionstat_t *stat)
{
  if (stat == (proportionstat_t *) 0)
    return (-1);
  stat->numersum += newval;
  stat->totaldenom = totaldenom;
  stat->proportion = stat->numersum / stat->totaldenom;
  stat->variance = stat->proportion * (1.0 - stat->proportion) / stat->totaldenom;
  if (confInterval (stat->totaldenom, stat->proportion, sqrt(stat->variance), &stat->conf))
    return (-1);

  return (0);
}

int
checkProportionStat (proportionstat_t *stat, double range)
{
  double    lower;
  double    upper;

  lower = (1.0 - range) * stat->proportion;
  upper = (1.0 + range) * stat->proportion;
  if (lower < stat->conf.lb && upper > stat->conf.ub) 
    return (1);
  else
    return (0);
  
}

void
dumpProportionStat (proportionstat_t *stat)
{
  printf ("denominator:\t%.3f\n", stat->totaldenom);
  printf ("numerator:\t%.3f\n", stat->numersum);
  printf ("proportion:\t%.3f\n", stat->proportion);
  printf ("variance:\t%.3f\n", stat->variance);
  printf ("conf interval:\t[%.3f, %.3f]\n", stat->conf.lb, stat->conf.ub);  
  fflush (stdout);
}
/********************************************************************************/
static int
confInterval (double n, double mean, double stddev, confint_t *interval)
{
  double	spread;
  
  if (interval == (confint_t *) 0)
    return (-1);
      
  spread = SampleStatistic (n, interval->level, stddev);
  interval->lb = mean - spread;
  interval->ub = mean + spread;

  return (0);
}

/*******************************************************************************/
static double
tval(double p, int df) 
{
  double    t;
  int	    positive = p >= 0.5;
  double    ddf;
  double    a;
  double    aa;

  p = (positive)? 1.0 - p : p;
  if (p <= 0.0 || df <= 0)
    t = HUGE_VAL;
  else if (p == 0.5)
    t = 0.0;
  else if (df == 1)
    t = 1.0 / tan((p + p) * 1.57079633);
  else if (df == 2)
    t = sqrt(1.0 / ((p + p) * (1.0 - p)) - 2.0);
  else {	
    ddf = df;
    a = sqrt(log(1.0 / (p * p)));
    aa = a * a;
    a = a - ((2.515517 + (0.802853 * a) + (0.010328 * aa)) /
             (1.0 + (1.432788 * a) + (0.189269 * aa) +
              (0.001308 * aa * a)));
    t = ddf - 0.666666667 + 1.0 / (10.0 * ddf);
    t = sqrt(ddf * (exp(a * a * (ddf - 0.833333333) / (t * t)) - 1.0));
  }
  return (positive)? t : -t;
}
/*******************************************************************************/
static double
SampleStatistic (double n, double l, double s)
{
  double t;
  int df;

  df = (int) rint((n - 1));	/* degrees of freedom must be an integer */
  if (df <= 0) return HUGE_VAL;
  t = tval((double)(100 + l) * 0.005, df);
  if (t == HUGE_VAL)
    return t;
  else
    return (t * s) / sqrt(n);
}
