/* $Id: usercdf.c,v 1.5 1999/01/29 04:25:58 ashaikh Exp $ */

/*  Description:
 *
 *	Implements some mechanisms for reading, storing, and using an empirical
 *	inverse CDF function.  CDF is read in as inverse CDF mapping from u (0 < u <= 1)
 *	to x.
 *
 *	CDF file format:
 *
 *	    <number of (u,x) pairs>
 *	    <u> <x>
 *	    <u> <x> 
 *	    ...
 *	    1 <x>
 *
 */
 
#include "usercdf.h"
#include "rstrace.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>


int parseCDFFile (FILE *, cdf_t **);
static int searchCDFArray (cdf_t *, int, int, double );
static double calcCDFAvg (cdf_t *);

/********************************************************************************/
int
readCDFFile (cdf_t **cdf, const char *filename)
{
  FILE	    *cdffile;

  cdffile = fopen (filename, "r");
  if (cdffile == NULL) {
#if RS_TRACE_LEVEL >= TRACE_ERRORS
    fprintf (stderr, "Error opening %s for reading: %s\n",
	     filename, strerror(errno));
#endif
    return (-1);
  }
  return (parseCDFFile (cdffile, cdf));
}

int
readCDFStdin (cdf_t **cdf)
{
  return (parseCDFFile (stdin, cdf));
}
/********************************************************************************/
int
parseCDFFile (FILE *cdffile, cdf_t **cdf)
{
  int	    result, i;
  int	    numentries;
  cdf_t	    *newcdf;
  double    u, x;

 *cdf = (cdf_t *) 0;
  
  result = fscanf (cdffile, "%d\n", &numentries);
  if (result != 1) {
    RS_PRINT_ERROR ("readCDFFile:  Couldn't read numentries... aborting\n");
    return (-1);
  }
  if (numentries < 1) {
    RS_PRINT_SOFT_ERROR ("readCDFFile: inverse CDF file must have at least one entry\n");
    return (-1);
  }
  newcdf = (cdf_t *) malloc (sizeof(cdf_t));
  newcdf->numvalues = numentries;
  newcdf->values = (cdfentry_t *) malloc (sizeof(cdfentry_t) * numentries);
  for (i = 0; i < numentries; i++) {
    result = fscanf (cdffile, "%lf %lf\n", &u, &x);
    if (result != 2) {
      RS_PRINT_SOFT_ERROR ("readCDFFile: Couldn't read CDF entry pair ...aborting\n");
      free (newcdf);
      return (-1);
    }
    newcdf->values[i].u = u;
    newcdf->values[i].x = x;
  }
  fclose (cdffile);
  newcdf->avg = calcCDFAvg (newcdf);

  *cdf = newcdf;
  return (0);
}
/********************************************************************************/
static double
calcCDFAvg (cdf_t *cdf)
{
  double    sum;
  double    lastu = 0.0;
  int	    i;

  if (cdf == (cdf_t *) 0) {
    RS_PRINT_ERROR ("calcCDFAvg:  Invalid cdf structure\n");
    return (-1.0);  /* this can't reliably be interpreted by caller as error */
  }

  sum = 0.0;
  for (i = 0; i < cdf->numvalues; i++) {
    sum += ((cdf->values[i].u - lastu) * cdf->values[i].x);
    lastu = cdf->values[i].u;
  }

  return (sum);
}

double
getCDFAvg (cdf_t *cdf)
{
  if (cdf == (cdf_t *) 0) {
    RS_PRINT_ERROR ("getCDFAvg:  Invalid cdf structure\n");
    return (-1.0);  /* this can't reliably be interpreted by caller as error */
  }

  return (cdf->avg);
}

/********************************************************************************/
double
sampleCDF (cdf_t *cdf)
{
  double    gen;
  int	    pos;

  gen = drand48();

  pos = searchCDFArray (cdf, 0, cdf->numvalues - 1, gen);
  if (pos < 0) {
    RS_PRINT_ERROR ("getCDFVal:  searchCDFArray failed\n");
    return (-1.0);
  }

  return (cdf->values[pos].x);  

}

double
getCDFVal (cdf_t *cdf, double u)
{
  int	pos;

  pos = searchCDFArray (cdf, 0, cdf->numvalues - 1, u);
  if (pos < 0) {
    RS_PRINT_ERROR ("getCDFVal:  searchCDFArray failed\n");
    return (-1.0);
  }
  return (cdf->values[pos].x);
}
/********************************************************************************/
static int
searchCDFArray (cdf_t *cdf, int l, int r, double u)
/* A binary search to find the position of the x-value that corresponds to the
 * given u-value.
 */
{
  int	m;

  m = (l + r) / 2;
  if (l > r) {
    RS_PRINT_SOFT_ERROR ("searchCDFArray:  Invalid array bound parameters\n");
    return (-1);
  }
  /* check if we found the right position */
  if (m > 0) {
    if ((u <= cdf->values[m].u) && (u > cdf->values[m - 1].u))
      return (m);
  }
  else if (m == 0) {
    if (u > cdf->values[m].u)
      return (m+1);
    else
      return (m);
  }

  /* this shouldn't occur */
  /* if (l == r) */
  /*	return (-1); */
  
  if (u < cdf->values[m].u)
    return (searchCDFArray(cdf, l, m - 1, u));
  else
    return (searchCDFArray(cdf, m + 1, r, u));
}
/********************************************************************************/
int
splitCDF (cdf_t *cdf, double split, cdf_t **lower, cdf_t **upper)
{
  int	    pos, numlower, numupper;
  double    borderval, trans;
  cdf_t	    *uppercdf, *lowercdf;
  int	    i, index;

  *lower = *upper = (cdf_t *) 0;

  if (cdf == (cdf_t *) 0) {
    RS_PRINT_SOFT_ERROR ("splitCDF:  Invalid cdf structure\n");
    return (-1);
  }

  /* figure out where in the original array the CDF is to be split */
  pos = searchCDFArray (cdf, 0, cdf->numvalues - 1, split);
  if (pos < 0) {
    RS_PRINT_ERROR ("splitCDF:  searchCDFArray failed\n");
    return (-1);
  }

  numlower = pos + 1;

  /* have to handle the case when the split falls on a discontinuity */
  if (cdf->values[pos].u == split) {
    numupper = cdf->numvalues - numlower;
  }
  else {
    numupper = cdf->numvalues - (numlower - 1);
  }
  
  borderval = cdf->values[pos].x;

  /* allocate new cdf arrays */
  uppercdf = (cdf_t *) malloc (sizeof(cdf_t));
  lowercdf = (cdf_t *) malloc (sizeof(cdf_t));
  if ( (uppercdf == (cdf_t *) 0) || (lowercdf == (cdf_t *) 0) ) {
    RS_PRINT_ERROR ("splitCDF:  Could not allocate new cdf structures\n");
    return (-1);
  }
  uppercdf->numvalues = numupper;
  lowercdf->numvalues = numlower;
  uppercdf->values = (cdfentry_t *) malloc (sizeof(cdfentry_t) * numupper);
  lowercdf->values = (cdfentry_t *) malloc (sizeof(cdfentry_t) * numlower);
  if ( (uppercdf->values == (cdfentry_t *) 0) || (lowercdf->values == (cdfentry_t *) 0) ) {
    RS_PRINT_ERROR ("splitCDF:  Could not allocate new cdf arrays\n");
    return (-1);
  }

  /* write the new lower cdf */
  trans = 1.0/split;    /* expand each interval by split^(-1) */
  for (i = 0; i < numlower - 1; i++) {
    lowercdf->values[i].u = cdf->values[i].u * trans;
    lowercdf->values[i].x = cdf->values[i].x;
  }
  /* handle the border case */
  lowercdf->values[i].u = 1.0;
  lowercdf->values[i].x = borderval;

  /* write the new upper cdf */
  trans = 1.0 / (1.0 - split);	/* expand each interval by (1 - split)^(-1) */
  for (i = 0; i < numupper; i++) {
    index = cdf->numvalues - numupper + i; 
    uppercdf->values[i].u = (cdf->values[index].u - split) * trans;
    uppercdf->values[i].x = cdf->values[index].x;
  }
  uppercdf->avg = calcCDFAvg (uppercdf);
  lowercdf->avg = calcCDFAvg (lowercdf);
  
  *upper = uppercdf;
  *lower = lowercdf;

  return (0);
  
}
/********************************************************************************/
void
printCDF (cdf_t *cdf)
{
  int	i, size;

  if (cdf == (cdf_t *) 0) {
    RS_PRINT_SOFT_ERROR ("printCDF:  Invalid cdf structure\n");
  }

  size =  cdf->numvalues;

  printf ("%.6f\t%.6f < u <= %.6f\n", cdf->values[0].x, 0.0, cdf->values[0].u);
  for (i = 1; i < size; i++) {
    printf ("%.6f\t", cdf->values[i].x);
    printf ("%.6f < u <= %.6f\n", cdf->values[i-1].u, cdf->values[i].u);
  }
}
/********************************************************************************/
void
printCDFGraph (cdf_t *cdf)
/* same as printCDF but in x-y coordinate form suitable for a graphing program
 */
{
  int	i, size;

  if (cdf == (cdf_t *) 0) {
    RS_PRINT_SOFT_ERROR ("printCDFGraph:  Invalid cdf structure\n");
  }

  size = cdf->numvalues;

  printf ("%.6f %.6f\n", 0.0, 0.0);
  for (i = 0; i < size; i++) {
    printf ("%.6f %.6f\n", cdf->values[i].x, cdf->values[i].u);
  }

}
/********************************************************************************/
int
dumpCDFFile (cdf_t *cdf, char *filename)
{
  FILE	*cdffile;
  int	i;
  
  if (cdf == (cdf_t *) 0) {
    RS_PRINT_SOFT_ERROR ("dumpCDFFile:  Invalid cdf structure\n");
  }
  
  cdffile = fopen (filename, "w");
  if (cdffile == NULL) {
#if RS_TRACE_LEVEL >= TRACE_ERRORS
    fprintf (stderr, "Error opening %s for writing: %s\n",
	     filename, strerror(errno));
#endif
    return (-1);
  }

  fprintf (cdffile, "%d\n", cdf->numvalues);
  for (i = 0; i < cdf->numvalues; i++) {
      fprintf (cdffile, "%f %f\n", cdf->values[i].u, cdf->values[i].x);
  }

  fclose (cdffile);

  return (0);
}
/********************************************************************************/
int
destroyCDF (cdf_t *cdf)
{

  if (cdf == (cdf_t *) 0)
    return (-1);

  free (cdf->values);
  free (cdf);

  return (0);
}
